
/**
	Turtle library, using media.gw
	@name   turtle (jC media Library)
    @brief  Allows to use a turtle to draw, as in logo.
    @date   2013-05-2
    @author jbgarcia@uvigo.es
*/

public class turtle {

private static final String _jc_lib_req = "media.gw";

public static final double Up = 0;
public static final double Arriba = Up;
public static final double Down = 180;
public static final double Abajo = Down;
public static final double Right = 90;
public static final double Derecha = Right;
public static final double Left = 270;
public static final double Izquierda = Left;

/**
 * Makes the turtle advance 'distance' pixels in the current angle
 * @param distance The distance to advance, as an int
 */
public static void forward(int distance)
{
    advance( distance, angle );
}

/**
 * Makes the turtle advance 'distance' pixels in the inverse angle
 * @param distance The distance to advance, as an int
 */
public static void backward(int distance)
{
    advance( distance, ( angle + 180 ) % 360 );
}

/**
 * Dibuja un círculo en la posición actual de la tortuga
 * @param radius El radio del círculo, como int
 */
public static void circle(int radius)
{
    gw.circle( posx, posy, radius, false );
}

/**
 * Dibuja un círculo en la posición actual de la tortuga, y lo rellena
 * @param radius El radio del círculo, como int
 */
public static void fillCircle(int radius)
{
    gw.circle( posx, posy, radius, true );
}

/**
 * Dibuja un punto en la pantalla, en la posición actual de la tortuga
 */
public static void dot()
{
    gw.point( posx, posy );
}

/**
 * Gira la tortuga para que encare unos grados absolutos.
 * @param a Los grados que debe encarar la tortuga, como double.
 */
public static void turn(double a)
{
    angle = a;
}

/**
 * Gira la tortuga para que encare los 0º
 */
public static void right()
{
    turn( Right );
}

/**
 * Gira la tortuga para que encare los 180º
 */
public static void left()
{
    turn( Left );
}

/**
 * Gira la tortuga para que encare los 90º
 */
public static void up()
{
    turn( Up );
}

/**
 * Gira la tortuga para que encare los 270º
 */
public static void down()
{
    turn( Down );
}

/**
 * Gira la tortuga en el sentido de las agujas del reloj,
 * añadiendo los grados aportados.
 * @param a Los grados a añadir hacia la derecha, como double
 */
public static void turnLeft(double a)
{
    double newAngle = ( angle - a ) % 360;

    if ( newAngle < 0 ) {
        newAngle = newAngle + 360;
    }

    turn( newAngle );
}

/**
 * Gira la tortuga en el sentido contrario de las agujas del reloj,
 * añadiendo los grados aportados.
 * @param a Los grados a añadir hacia la izquierda, como double
 */
public static void turnRight(double a)
{
    turn( ( angle + a ) % 360 );
}

/**
 * Cambia el color actual con el que dibuja la tortuga
 * @param color El color a escoger.
 */
public static void setColor(int color)
{
    gw.setColor( color );
}

/**
 * Cambia el color actual con el que dibuja la tortuga,
 * mediante el esquema RGB
 * @param r El componente rojo.
 * @param g El componente verde.
 * @param b El componente azul.
 */
public static void setColor(int r, int g, int b)
{
    gw.setColor( r, g, b );
}

/**
 * Recoloca la tortuga en el centro y prepara de nuevo todas las vbles.
 */
public static void reset()
{
    posx = gw.getWidth() / 2;
    posy = gw.getHeight() / 2;
    angle = 90;
    color = gw.Black;
    down = true;
}

/**
 * Comprueba si la tortuga dibuja al moverse o no.
 * @return true si dibuja al moverse, false en otro caso.
 */
public static boolean isPenDown()
{
    return down;
}

/**
 * Mueve la tortuga a la posición dada, utilizando coordenadas cartesianas
 * @param x La coordenada x de la posición, como int
 * @param y La coordenada y de la posición, como int
 */
public static void setPosition(int x, int y)
{
    moveTo( x, y );
}

/**
 * Mueve la tortuga a la posición dada, utilizando coordenadas polares
 * @param distance La distancia a avanzar, como int
 * @param angle El ángulo en el que avanzar, como double
 */
public static void setPosition(int distance, double angle)
{
    advance( distance, angle );
}

/**
 * Devuelve el ángulo que sería necesario para llegar a la posición
 * dada desde la posición actual.
 * @param x La coordenada x de la posición de destino, como int
 * @param y La coordenada y de la posición de destino, como int
 * @return El ángulo necesario, como double
 */
public static double getAngleTo(int x, int y)
{
    return Math.toDegrees( Math.atan( y / x ) );
}

/**
 * Devuelve la distancia necesaria para llegar a la posición dada
 * desde la posición actual.
 * @param x La coordenada x de la posición de destino, como int
 * @param y La coordenada y de la posición de destino, como int
 * @return La distancua al punto de dest¡ino, como int
 */
public static double getDistanceTo(int x, int y)
{
    return Math.sqrt( ( x * x ) + ( y * y ) );
}

/**
 * Devuelve la coordenada x de la posición actual
 * @return La coordenada x, como int
 */
public static int getX()
{
    return posx;
}

/**
 * Devuelve la coordenada y de la posición actual
 * @return La coordenada y, como int
 */
public static int getY()
{
    return posy;
}

/**
 * Hace que la tortuga deje rastro al avanzar
 */
public static void setPenDown()
{
    down = true;
}

/**
 * Hace que la tortuga no deje rastro al avanzar
 */
public static void setPenUp()
{
    down = false;
}

/**
 * Hace que la tortuga deje o no deje rastro al avanzar
 * @param d true para que deje rastro, false para que no lo deje.
 */
public static void setPen(boolean d)
{
    down = d;
}

/**
 * Visualizar un texto en la posición actual de la tortuga
 * @param msg El texto a visualizar, como char[]
 */
public static void print(char[] msg)
{
    gw.print( posx, posy, msg );
}

private static void moveTo(int x, int y)
{
    gw.setColor( color );

    // Draw trace, if needed
    if ( down ) {
        gw.line( posx, posy, x, y );
    }

    posx = x;
    posy = y;
}

private static void advance(int distance, double angle)
{
    double deltaY = Math.cos( Math.toRadians( angle ) ) * distance;
    double deltaX = Math.sin( Math.toRadians( angle ) ) * distance;

    moveTo( (int) ( posx + deltaX ), (int) ( posy + deltaY ) );
}

private static int posx = 160;
private static int posy = 120;
private static int color = gw.Black;
private static double angle = 90;
private static boolean down = true;


} // jc turtle media lib
