
/**
	Turtle library, using media.gw
	@name   turtle (jC media Library)
    @brief  Allows to use a turtle to draw, as in logo.
    @date   2013-05-2
    @author jbgarcia@uvigo.es
*/

var turtle = ( function() {

	var Up = 0;
	var Down = 180;
	var Right = 90;
	var Left = 270;

	/**
	 * Makes the turtle advance 'distance' pixels in the current angle
	 * @param distance The distance to advance, as an int
	 */
	function forward(distance)
	{
		advance( distance, angle );
	}

	/**
	 * Makes the turtle advance 'distance' pixels in the inverse angle
	 * @param distance The distance to advance, as an int
	 */
	function backward(distance)
	{
		advance( distance, ( angle + 180 ) % 360 );
	}

	/**
	 * Dibuja un círculo en la posición actual de la tortuga
	 * @param radius El radio del círculo, como int
	 */
	function circle(radius)
	{
		gw.circle( posx, posy, radius, false );
	}

	/**
	 * Dibuja un punto en la pantalla, en la posición actual de la tortuga
	 */
	function dot()
	{
		gw.point( posx, posy );
	}

	/**
	 * Gira la tortuga para que encare unos grados absolutos.
	 * @param a Los grados que debe encarar la tortuga, como double.
	 */
	function turn(a)
	{
		angle = a;
	}

	/**
	 * Gira la tortuga para que encare los 0º
	 */
	function right()
	{
		turn( Right );
	}

	/**
	 * Gira la tortuga para que encare los 180º
	 */
	function left()
	{
		turn( Left );
	}

	/**
	 * Gira la tortuga para que encare los 90º
	 */
	function up()
	{
		turn( Up );
	}

	/**
	 * Gira la tortuga para que encare los 270º
	 */
	function down()
	{
		turn( Down );
	}

	/**
	 * Gira la tortuga en el sentido de las agujas del reloj,
	 * añadiendo los grados aportados.
	 * @param a Los grados a añadir hacia la derecha, como double
	 */
	function turnLeft(a)
	{
		var newAngle = ( angle - a ) % 360;

		if ( newAngle < 0 ) {
			newAngle = newAngle + 360;
		}

		turn( newAngle );
	}

	/**
	 * Gira la tortuga en el sentido contrario de las agujas del reloj,
	 * añadiendo los grados aportados.
	 * @param a Los grados a añadir hacia la izquierda, como double
	 */
	function turnRight(a)
	{
		turn( ( angle + a ) % 360 );
	}

	/**
	 * Cambia el color actual con el que dibuja la tortuga
	 * opcional: mediante el esquema RGB
	 * @param color El color a escoger.
	 * @param r El componente rojo.
	 * @param g El componente verde.
	 * @param b El componente azul.
	 */
	function setColor(color)
	{
		if ( arguments.length == 1 ) {
			gw.setFgColor( color );
		}
		else
		if ( arguments.length == 3 ) {
			gw.setFgColor( arguments[ 0 ], arguments[ 1 ], arguments[ 2 ] );
		}
		else {
			gw.setFgColor( gw.Black );
		}
		
		return;
	}

	/**
	 * Recoloca la tortuga en el centro y prepara de nuevo todas las vbles.
	 */
	function reset()
	{
		posx = gw.getWidth() / 2;
		posy = gw.getHeight() / 2;
		angle = 90;
		color = gw.Black;
		down = true;
	}

	/**
	 * Comprueba si la tortuga dibuja al moverse o no.
	 * @return true si dibuja al moverse, false en otro caso.
	 */
	function isPenDown()
	{
		return down;
	}

	/**
	 * Mueve la tortuga a la posición dada, utilizando coordenadas cartesianas
	 * Mueve la tortuga a la posición dada, utilizando coordenadas polares
	 * Enviar las coordenadas polares como num. flotante.
	 * @param x La coordenada x de la posición, como int
	 * @param y La coordenada y de la posición, como int
	 * @param distance La distancia a avanzar, como int
	 * @param angle El ángulo en el que avanzar, como double
	 */
	function setPosition(x, y)
	{
		if ( ( typeof x === 'number'
		    && parseFloat( x ) == parseInt( x, 10 ) && !isNaN( x ) )
		  || ( typeof y === 'number'
		    && parseFloat( y ) == parseInt( y, 10 ) && !isNaN( y ) ) )
		{
			advance( x, y );
		} else {
			moveTo( x, y );
		}

		return;
	}

	/**
	 * Hace que la tortuga no deje rastro al avanzar
	 */
	function setPenUp()
	{
		down = false;
	}

	/**
	 * Hace que la tortuga deje o no deje rastro al avanzar
	 * @param d true para que deje rastro, false para que no lo deje.
	 */
	function setPen(d)
	{
		down = d;
	}

	/**
	 * Visualizar un texto en la posición actual de la tortuga
	 * @param msg El texto a visualizar, como char[]
	 */
	function print(msg)
	{
		gw.print( posx, posy, msg );
	}

	function moveTo(x, y)
	{
		gw.setColor( color );

		// Draw trace, if needed
		if ( down ) {
			gw.line( posx, posy, x, y );
		}

		posx = x;
		posy = y;
	}

	function toRadians(value)
	{
		return ( value * Math.PI / 180 );
	}

	function advance(distance, angle)
	{
		var deltaY = Math.cos( toRadians( angle ) ) * distance;
		var deltaX = Math.sin( toRadians( angle ) ) * distance;

		moveTo( posx + deltaX, posy + deltaY );
	}

	var posx = 160;
	var posy = 120;
	var color = gw.Black;
	var angle = 90;
	var down = true;

	return {
		"Up": Up,
		"Arriba": Up,
		"Down": Down,
		"Abajo": Down,
		"Right": Right,
		"Derecha": Right,
		"Left": Left,
		"Izquierda": Left,
		"forward": forward,
		"backward": backward,
		"circle": circle,
		"dot": dot,
		"turn": turn,
		"right": right,
		"left": left,
		"up": up,
		"down": down,
		"turnLeft": turnLeft,
		"turnRight": turnRight,
		"setColor": setColor,
		"reset": reset,
		"isPenDown": isPenDown,
		"setPosition": setPosition,
		"setPenUp": setPenUp,
		"setPen": setPen,
		"print": print,
		"moveTo": moveTo,
		"advance": advance,
	};
})();
