/**
    @name   Char management (jC Standard Library)
    @brief  Common operations over characters
    @date   2013-07-15
    @author jbgarcia@uvigo.es
*/

var charType = ( function() {
	
	var PunctuationChars = [ '.', ',', ';', ':', '-', '(', ')' ];
	var BlankChars = [ ' ', '\t' ];
	
	function isString(str)
	{
		return ( typeof str === "string" || str instanceof String );
	}
	
	/**
		@brief Decides whether the char is letter
		@param ch The char to check
		@return true if it is letter; false if not
	*/
	function isLetter(ch)
	{
		return ( isString( ch )
		      && ( ch.length === 1 )
		      && ( ch.match( /[a-z]/i ) != null ) );
	}
	
	/**
		@brief Decides whether the char is digit or letter
		@param ch The char to check
		@return true if it is letter; false if not

	*/
	function isDigit(ch)
	{
		return ( isString( ch )
		      && ( ch.length === 1 )
		      && ( !isNaN( ch ) ) );
	}
	
	/**
		@brief Decides whether the char is digit or letter
		@param ch The char to check
		@return true if it is letter or digit; false otherwise
	*/
	function isAlphaNumeric(ch)
	{
		return ( isDigit( ch ) || isLetter( ch ) );
	}
	
	/**
		@brief Decides whether the char is a control character
		@param ch The char to check
		@return true if it is a basic letter or digit; false otherwise
	*/
	function isControl(ch)
	{
		return ( isString( ch )
			     && ch.length === 1
			     && ( ch.charCodeAt( 0 ) < " ".charCodeAt( 0 ) ) );
	}
	
	/**
		@brief Decides whether the char is lower case
		@param ch The char to check
		@return true if it is lowercase; false if not
	*/
	function isLowerCase(ch)
	{
		return ( isString( ch )
		      && isNaN( ch )
		      && ( ch.toLowerCase() == ch ) );
	}

	/**
		@brief Decides whether the char is upper case
		@param ch The char to check
		@return true if it is lowercase; false if not
	*/
	function isUpperCase(ch)
	{
		return ( isString( ch )
		      && isNaN( ch )
		      && ( ch.toUpperCase() == ch ) );
	}
	
	/**
		@brief Checks if the character is a delimiter
		@param ch The character to check
		@return true if ch is tab (various forms) or space, false otherwise.
	*/
	function isSpace(ch)
	{
		return ( isString( ch )
		      && ch.length === 1
		      && ( BlankChars.indexOf( ch ) >= 0 ) );
	}
	
	/**
		@brief Checks if the character is punctuation mark
		@param ch The character to check
		@see PunctuationChars
		@return true if ch is punctuation, false otherwise.
	*/
	function isPunctuation(ch)
	{
		return ( isString( ch )
		      && ch.length === 1
		      && ( PunctuationChars.indexOf( ch ) >= 0 ) );
	}
	
	/**
		@brief Changes the char to lower case
		@param ch The char to change
		@return The converted char
	*/
	function toLowerCase(ch)
	{
		var toret = "";
		
		if ( isString( ch ) ) {
			toret = ch.toLowerCase();
		}
		
		return toret;
	}
	
	/**
		@brief Changes the char to upper case
		@param ch The char to change
		@return The converted char
	*/
	function toUpperCase(ch)
	{
		var toret = "";
		
		if ( isString( ch ) ) {
			toret = ch.toUpperCase();
		}
		
		return toret;
	}
	
	return {
		"PunctuationChars": PunctuationChars,
		"BlankChars": BlankChars,
		"isLetter": isLetter,
		"isDigit": isDigit,
		"isAlphaNumeric": isAlphaNumeric,
		"isControl": isControl,
		"isLowerCase": isLowerCase,
		"isUpperCase": isUpperCase,
		"isSpace": isSpace,
		"isPunctuation": isPunctuation,
		"toLowerCase": toLowerCase,
		"toUpperCase": toUpperCase,
	};
})();
