
/**
    jC util standard library package
    @name   Util (jC Standard Library)
    @brief  Common convenience functions
    @date   2010-12-1
    @author jbgarcia@uvigo.es
*/

var util = ( function() {

	/// Program termination code for a failed program
	var ExitFailure = 1;

	/// Program termination code for a successful program
	var ExitSuccess = 0;

	/// Number of seconds in a day 24 * 60 * 60
	var SecondsInDay =  86400;

	/// Number of minutes in a day
	var MinutesInDay = 1440;

	/// The number of months in a year
	var MonthsInYear = 12;
	
	/**
		@brief returns the size of a vector
		@param vector The vector
		@return the size as an int
	*/
	function size(obj)
	{
		var toret = 0;
		
		if ( obj !== null ) {
			if ( typeof obj == 'string'
			  || obj instanceof String
			  || obj instanceof Array )
			{
				toret = obj.length;
			} else {
				toret = 1;
			}
		}
		
		return toret;
	}
	
	/**
		@brief returns whether the size of a vector is zero
		@param vector The vector
		@return true if size equals 0, false otherwise
		@see size
	*/
	function isEmpty(obj)
	{
		return size( obj ) == 0;
	}
	
	/**
		@brief Assert a condition, stop the program if not true. Exits with error code -1
		@param expr The condition to check
		@param msg The message to show
		@return If expr is false, it never returns
	*/
	function verify(expr, msg)
	{
		if ( !expr ) {
			document.write( new String( msg ) );
		}

		return;
	}
	
	/**
		@brief Returns the current date in the system
		@return Date as a vector of three ints: 0=year,1=month,2=day
	*/
	function getDate()
	{
		var toret = [ 1900, 1, 1 ]
		var now = new Date();

		toret[ 0 ] = now.getFullYear();
		toret[ 1 ] = now.getMonth();
		toret[ 2 ] = now.getDate();
		
		return toret;
	}
	
	/**
		@brief Returns the current date in the format "year-month-day",
			   as in ISO 8601 format.
		@return Date as a human readable vector of char.
	*/
	function getStrDate()
	{
		var toret = "";
		var date = getDate();

		toret += date[ 0 ];
		toret += '-';
		toret += date[ 1 ];
		toret += '-';
		toret += date[ 2 ];

		return toret;
	}
	
	/**
		@brief Returns the current time in the system
		@return Date as a vector of three ints: 0=hour,1=minute,2=second
	*/
	function getTime()
	{
		var toret = [ 0, 0, 0 ];
		var now = new Date();

		toret[ 0 ] = now.getHours();
		toret[ 1 ] = now.getMinutes();
		toret[ 2 ] = now.getSeconds();

		return toret;
	}

	/**
		@brief Returns the current time in the format "hour:minute:second"
		@return Date as a string
	*/
	function getStrTime()
	{
		var toret = "";
		var time = getTime();

		toret += time[ 0 ];
		toret += ':';
		toret += time[ 1 ];
		toret += ':';
		toret += time[ 2 ];

		return toret;
	}
	
	/**
		@brief Returns the number of seconds from a given date to the beginning
			   of the year of that date.
		@return The number of seconds, as an int.
	 */
	function getSecondsToDate(year, month, day)
	{
		var org = ( new Date( year, 1, 1, 0, 0, 0 ).getTime() ) / 1000;
		var end = ( new Date( year, month, day, 0, 0, 0 ).getTime() ) / 1000;
		
		return ( end - org );
	}

	/**
		 @brief Returns the number of seconds from the beginning
				of the year of the current date to that date
		 @return The number of seconds, as an int.
	*/
	function getSecondsToNow()
	{
		var date = getDate();
		
		return getSecondsToDate( date[ 0 ], date[ 1 ], date[ 2 ] );
	}
	
	/**
		@brief Convert x from double to int
		@param x The value to be converted.
		@return The value finally converted, as an integer.
	*/
	function dblToInt(x)
	{
		return Math.round( x );
	}

	/**
		@brief Convert x from int to double
		@param x The value to be converted.
		@return The value finally converted, as a double.
	*/
	function intToDbl(x)
	{
		return x;
	}
	
	/**
		@brief Convert x from char to int
		@param x The value to be converted.
		@return The value finally converted.
	*/
	function chrToInt(x)
	{
		var toret = 0;
		
		if ( x != null ) {
			var str = new String( x );
			
			if ( str.length > 0 ) {
				toret = str.charCodeAt( 0 );
			}
		}
		
		return toret;
	}

	/**
		@brief Convert x from int to char
		@param x The value to be converted.
		@return The value finally converted.
	*/
	function intToChr(x)
	{
		var index = new Number( x );
		
		if ( isNaN( index ) ) {
			index = 0;
		}
		
		return String.fromCharCode( index );
	}

	/**
		@brief Sleeps for ms miliseconds
		@param ms The number of ms to stop the program.
	*/
	function sleep(ms)
	{
	  var start = new Date().getTime();
	  var now = start;
	  
	  while ( ms > ( now - start ) ) {		 
		now = new Date().getTime(); 
	  }
	  
	  return;
	}

	return {
		"ExitFailure": ExitFailure,
		"ExitSuccess": ExitSuccess,
		"SecondsInDay": SecondsInDay,
		"MinutesInDay": MinutesInDay,
		"MonthsInYear": MonthsInYear,
		"size": size,
		"isEmpty": isEmpty,
		"verify": verify,
		"getDate": getDate,
		"getStrDate": getStrDate,
		"getTime": getTime,
		"getStrTime": getStrTime,
		"getSecondsToDate": getSecondsToDate,
		"getSecondsToNow": getSecondsToNow,
		"dblToInt": dblToInt,
		"intToDbl": intToDbl,
		"chrToInt": chrToInt,
		"intToChr": intToChr,
		"sleep": sleep,
	};
})();
