// Ast.java

package core;

import core.exceptions.semantic.AmbiguityError;
import core.exceptions.SemanticError;
import core.exceptions.syntax.IncorrectId;
import core.intermediatecode.CodeBlockOwner;
import core.intermediatecode.CodeUnit;
import core.intermediatecode.instrs.Cal;
import core.intermediatecode.Data;
import core.intermediatecode.instrs.Fun;
import core.intermediatecode.Module;
import core.intermediatecode.Prg;
import core.intermediatecode.RefId;
import core.lexical.Position;
import java.util.ArrayList;

/**
 * The abstract syntax tree. It is created with CodeUnits as its immediate
 * leaves.
 * @author baltasarq
 */
public class Ast {
    
    // Public singleton
    protected Ast()
    {
        this.codeUnits = new ArrayList<CodeUnit>();
    }
    
    public static Ast getAst()
    {
        if ( ast == null ) {
            ast = new Ast();
        }

        return ast;
    }

    public static Ast reset()
    {
        ast = null;
        return getAst();
    }
    
    /** 
     * Performs a check of the code in the Ast
     */ 
    public void chk() throws CompilingError
    {
        for(Module module: this.getModules()) {
            module.chk();
        }
        
        this.getPrg().chk();
    }

    /**
     * Adds a new codeunit (module or program) to the Ast
     * @param cu The CodeUnit to add
     * @return The CodeUnit added
     */
    public CodeUnit addCodeUnit(CodeUnit cu) throws SemanticError
    {

        if ( cu instanceof Prg ) {
            if ( this.prg == null ) {
                this.prg = (Prg) cu;
            } else {
                Position pos = new Position( this.prg.getName(), 0, 0 );
                throw new SemanticError( pos, I18n.get( I18n.Word.Duplicated )
                        + " prg."
                );
            }
        } else {
            this.codeUnits.add( cu );
        }

        return cu;
    }
    
    /**
     * Adds a new module to the Ast
     * @param mod The new module to add
     * @return The Module added.
     */
    public Module addModule(Module mod) throws SemanticError
    {
        this.addCodeUnit( mod );
        return mod;
    }

    /**
     * Adds the program to the Ast - there can be only one program.
     * @param prg The Prg to add
     * @return The Prg added.
     */
    public Prg addPrg(Prg prg) throws SemanticError
    {
        this.addCodeUnit( prg );
        return prg;
    }

    /**
     * Returns the modules (not the program) in the Ast,
     * @return The modules in the Ast, as a vector of Module
     */
    public Module[] getModules()
    {
        return this.codeUnits.toArray( new Module[ this.codeUnits.size() ] );
    }

    /**
     * Looks for a module of a given name
     * @param name The name of the module to look for.
     * @return The module, as a Module class.
     */
    public Module lookForModule(String name)
    {
        Module toret = null;
        final boolean isCompleteName = name.contains( "." );

        searchForModule:
        for(Module mod: this.getModules()) {

            if ( !isCompleteName ) {
                if ( mod.getName().equals( name ) ) {
                    toret = mod;
                    break searchForModule;
                }
            } else {
                if ( mod.getCompleteName().equals( name ) ) {
                    toret = mod;
                    break searchForModule;
                }
            }
        }

        return toret;
    }
    
    /**
     * Returns the program of this compilation in the Ast
     * @return The program, as a Prg object.
     */
    public Prg getPrg()
    {
        return this.prg;
    }

    /**
     * Returns all the CodeUnit's in this Ast, i.e., all modules AND the program
     * @return All CodeUnit objects, including the Prg, as a CodeUnit vector.
     */
    public CodeUnit[] getCodeUnits()
    {
        // Add the modules
        CodeUnit[] toret = this.codeUnits.toArray( new CodeUnit[ this.codeUnits.size() +1 ] );

        // Add the program
        toret[ toret.length -1 ] = this.prg;
        return toret;
    }
    
    public Data locateData(Position pos, CodeBlockOwner owner, RefId id)
                                              throws IncorrectId, AmbiguityError
    {
        return AstMemberFinder.locateData( pos, owner, id );
    }
    
    public Fun locateFunction(Position pos, Cal cal, RefId id) throws CompilingError
    {
        return AstMemberFinder.locateFunction( pos, cal, id );
    }

    private ArrayList<CodeUnit> codeUnits;
    private Prg prg = null;
    protected static Ast ast;
}
