// AstMemberFinder.java

package core;

import core.intermediatecode.CodeBlock;
import core.exceptions.semantic.AmbiguityError;
import core.exceptions.syntax.IncorrectId;
import core.intermediatecode.CodeBlockOwner;
import core.intermediatecode.CodeUnit;
import core.intermediatecode.instrs.Cal;
import core.intermediatecode.Data;
import core.intermediatecode.instrs.Fin;
import core.intermediatecode.instrs.Fun;
import core.intermediatecode.Member;
import core.intermediatecode.Module;
import core.intermediatecode.RefId;
import core.intermediatecode.Type;
import core.lexical.Position;
import core.lexical.Reserved;
import core.exceptions.SemanticError;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * Find members, by its id's, in the Ast
 * @author baltasarq
 */
public class AstMemberFinder {
    
    protected static Ast getAst()
    {
        return Ast.getAst();
    }
    
    /**
     * Locate data in the Ast. Returns all data members that have that id.
     * They could be:
     *  - a local variable or const in the CodeBlock the reference was made
     *  - a const in the module where the reference was made
     *  - a const in another module
     * @param id The identifier to look for.
     * @param owner The owner of the codeblock in which the id happened.
     * @return 
     */
    public static Data[] lookUpData(RefId id, CodeBlockOwner owner)
    {
        ArrayList<Data> dataFound = new ArrayList<Data>();
        Data cnst = null;
        
        if ( id.getPrefix() != null ) {
            // Prefix found, look in the specific module
            Module module = getAst().lookForModule( id.getPrefix() );

            if ( module != null ) {
                cnst = module.lookForConst( id.getName() );
                
                if ( cnst != null
                && ( (Fin) cnst ).isVisible() )
                {
                    dataFound.add( cnst );
                }
            }  
        } else {
            CodeBlockOwner lookForOwner = owner;
            
            do {
                // Look in its own codeblock, and owners, if needed
                cnst = lookForOwner.lookForData( id.getName() );
                
                if ( cnst == null ) {
                    lookForOwner = lookForOwner.getOwner();
                }
                
            } while( lookForOwner != null
                  && cnst == null );
            
            if ( cnst != null ) {
                dataFound.add( cnst );
            } else {
                // Locate the module where the reference was done
                CodeUnit homeModule = CodeBlock.getModuleOwner( owner );
                cnst = homeModule.lookForConst( id.getName() );

                if ( cnst != null ) {
                    dataFound.add( cnst );
                } else {
                    // Not found, no prefix, lookup needed for all modules
                    for(Module module: getAst().getModules()) {
                        cnst = module.lookForConst( id.getName() );

                        if ( cnst != null
                        && ( (Fin) cnst ).isVisible() )
                        {
                            dataFound.add( cnst );
                        }
                    }
                 }
            }
        }
        
        return dataFound.toArray( new Data[ dataFound.size() ] );
    }
    
    /**
     * Locate data in the Ast. It must be only one member as answer, otherwise
     * it raises an exception.
     * 
     * It can be:
     *  - a local variable or const in the CodeBlock the reference was made
     *  - a const in the module where the reference was made
     *  - a const in another module
     * 
     * @param pos The pos in the source code in which the id happened.
     * @param owner The owner of the codeblock in which the id happened.
     * @param id The id itself.
     * @return The const or variable found, or an AmbiguityError exception.
     */
    public static Data locateData(Position pos, CodeBlockOwner owner, RefId id)
                                            throws IncorrectId, AmbiguityError
    {
        Data toret = null;
        Data[] dataMembers;
        
        // First try
        owner = CodeBlock.getRootOwner( owner );
        dataMembers = lookUpData( id, owner );
        
        // Analyze whether the data  was found or not
        if ( dataMembers.length == 1 ) {
            toret = dataMembers[ 0 ];
            fixId( id, owner, toret );
        }
        else
        if ( dataMembers.length > 1 ) {
            throw createAmbiguityError( pos, true, id, (Member[]) dataMembers );
        }
        
        return toret;
    }
    
    /**
     * Looks up for functions with a given id, residing in a Codeblock, with
     * a given number of arguments.
     * 
     * It could be:
     *  - The id is complete, the module the function should live in is given
     *  - A function in the same module the reference happened
     *  - A function in any module
     * 
     * @param id
     * @param owner
     * @param reqArgs
     * @return 
     */
    public static Fun[] lookUpFns(RefId id, CodeBlockOwner owner, int reqArgs)
    {
        ArrayList<Fun> fnsFound = new ArrayList<Fun>();
        Fun[] fns = null;
        
        // Find possibilities
       if ( id.getPrefix() != null ) {
            // Has prefix, look up in a specific module
            Module module = Ast.getAst().lookForModule( id.getPrefix() );

            if ( module != null ) {
                fns = module.getFunctionByName( id.getName(), reqArgs );
                fns = filterFunctionsByVisibility( fns );

                if ( fns != null ) {
                    fnsFound.addAll( Arrays.asList( fns ) );
                }
            }
       } else {
           // Look where the reference was done
           fns = CodeBlock.getModuleOwner( owner ).getFunctionByName( id.getName(), reqArgs );
           
           if ( fns == null ) {                      
               // No prefix, not in owner, look in all modules
               for(Module module: getAst().getModules()) {
                    fns = module.getFunctionByName( id.getName(), reqArgs );
                    fns = filterFunctionsByVisibility( fns );

                    if ( fns != null ) {
                        fnsFound.addAll( Arrays.asList( fns ) );
                    }
                }
           } else {
               fnsFound.addAll( Arrays.asList( fns ) );
           }
       }
       
       return fnsFound.toArray( new Fun[ fnsFound.size() ] );
    }
    
    protected static Type[]
    getTypesForRealArguments(Position pos, Cal cal, CodeBlockOwner owner)
                                               throws SemanticError, IncorrectId
    {
        Type[] realArgsTypes = new Type[ cal.getNumberOfArgs() ];
        RefId[] realArgs = cal.getArgs();
        
        // Prepare data about call
        for(int i = 0; i < realArgsTypes.length; ++i) {
           Data data = locateData( pos, owner, realArgs[ i ] );

           if ( data != null ) {
                realArgsTypes[ i ] = data.getType();
           } else {
               throw new SemanticError(
                       pos,
                       I18n.get( I18n.Word.Missing )
                       + ": '" + realArgs[ i ].toString()
                       + "' " + I18n.get( I18n.Sentence.InCallTo  )
                       + ' ' + cal.getFunctionId() + "()"
               );
           }
        }
        
        return realArgsTypes;
    }
    
    protected static Fun[] filterFunctionsByParams(
                        Position pos,
                        Fun[] fns,
                        Cal cal,
                        CodeBlockOwner owner,
                        Type[] realArgsTypes)
                               throws IncorrectId, AmbiguityError, SemanticError
    {
        ArrayList<Fun> toret = new ArrayList<Fun>();

        // Filter by parameters' type
        for(int i = 0; i < fns.length; ++i) {
           final Data[] realParams = fns[ i ].getParams();

           int n = 0;
           for(; n < realParams.length; ++n) {

               if ( !Type.areCompatible( realArgsTypes[ n ], realParams[ n ].getType() ) )
               {
                   break;
               }
           }

           if ( n == realParams.length ) {
               toret.add( fns[ i ] );
           }
        }

        return toret.toArray( new Fun[ toret.size() ] );
    }
    
    public static Fun[] filterFunctionsByExactParams(Fun[] fns, Type[] realArgsTypes)
    {
        ArrayList<Fun> toret = new ArrayList<Fun>();

        // Filter by parameters' type
        for(int i = 0; i < fns.length; ++i) {
           Data[] realParams = fns[ i ].getParams();    

           int n = 0;
           for(; n < realParams.length; ++n) {

               if ( !realArgsTypes[ n ].equals( realParams[ n ].getType() ) )
               {
                   break;
               }
           }

           if ( n == realParams.length ) {
               toret.add( fns[ i ] );
           }
        }

        return toret.toArray( new Fun[ toret.size() ] );
    }
    
    /**
     * Returns a new vector of functions, with only those being public.
     * @param fns The vector of functions to filter, as a primitive vector.
     * @return The new list of functions, as a primitive vector
     */
    protected static Fun[] filterFunctionsByVisibility(Fun[] fns)
    {
        ArrayList<Fun> toret = new ArrayList<Fun>();

        if ( fns != null ) {
            // Filter by visibility
            for(int i = 0; i < fns.length; ++i) {
               if ( fns[ i ].getVisibility().isVisible() ) {
                   toret.add( fns[ i ] );
               }
            }
        }

        return toret.toArray( new Fun[ toret.size() ] );
    }
    
    /**
     * Locates a function in all modules of the Ast, if needed
     * There should be only one, or an exception is raised
     * 
     * It could be:
     *  - A function in the same module the reference happened
     *  - A function in another module
     * 
     * @param pos   The position, in case it is needed to throw an exception
     * @param cal   The Cal instruction
     * @param id    The RefId itself
     * @return      The function references, as a Fun object.
     */
    public static Fun locateFunction(Position pos, Cal cal, RefId id) throws CompilingError
    {
       Fun toret = null;
       int reqArgs = cal.getNumberOfArgs();
       CodeBlockOwner owner = CodeBlock.getFunctionOwner( cal.getOwner() );
       Fun[] fns;
       Type[] realArgsTypes;
       
       // Determine owner
       if ( owner == null ) {
           owner = CodeBlock.getModuleOwner( cal.getOwner() );
       }
       
       // Retrieve possible functions
       fns = lookUpFns( id, owner, reqArgs );
       realArgsTypes = getTypesForRealArguments( pos, cal, owner );
       fns = filterFunctionsByParams( pos, fns, cal, owner, realArgsTypes );
       
       if ( fns.length > 1 ) {
           fns = filterFunctionsByExactParams( fns, realArgsTypes );
       }

       // Analyze whether it has been found or not
       if ( fns.length == 1 ) {
           toret = fns[ 0 ];
           fixId( id, owner, toret );
       }
       else
       if ( fns.length > 1 ) {
           throw createAmbiguityError( pos, true, cal.getFunctionId(), fns );
       }
        
       return toret;
    }
    
    protected static void fixId(RefId id, CodeBlockOwner owner, Member member) throws IncorrectId
    {
        CodeUnit memberHomeModule = CodeBlock.getModuleOwner( member.getOwner() );
        CodeUnit idHomeModule = CodeBlock.getModuleOwner( owner );

        if ( !id.hasPrefix()
          && memberHomeModule != idHomeModule )
        {
            id.setPrefix( memberHomeModule.getName() );
        }
        
        return;
    }
    
    protected static
    AmbiguityError createAmbiguityError(Position pos,
                                       boolean isFunction,
                                       RefId id,
                                       Member[] members)
    {
        StringBuilder message = new StringBuilder();

        if ( isFunction ) {
            message.append( " fn.: " );
        } else {
            message.append( I18n.get( I18n.Word.Constant ) );
            message.append( ", variable " );
        }
        
        message.append( id.toString() );
        message.append( ", among:\n" );
        for(Member item: members) {
            message.append( Reserved.cnvtNativeToType( item.getType() ) );
            message.append( ' ' );
            message.append( item.getDesc() );
            
            message.append( '\n' );
        }

        return new AmbiguityError( pos, message.toString() );
    }
    
}
