// BshBackend.java

package core;

import core.bshbackend.Opcode;
import core.bshbackend.Reserved;
import core.exceptions.SemanticError;
import core.intermediatecode.CodeBlock;
import core.intermediatecode.CodeUnit;
import core.intermediatecode.Instr;
import core.intermediatecode.Prg;
import core.intermediatecode.Data;
import core.intermediatecode.instrs.Fin;
import core.intermediatecode.instrs.Fun;
import core.intermediatecode.literals.ArrayLiteral;

/**
 *
 * @author baltasarq
 */
public class BshBackend extends Backend {

    public BshBackend(CodeUnit cdu)
    {
        super( cdu );
        builder = new StringBuilder();
    }

    @Override
    public byte[] generate() throws SemanticError {

        // Clean
        builder.delete( 0, builder.length() );

        // Add title
        builder.append( Reserved.CommentMark );
        builder.append( ' ' );
        builder.append( this.getCodeUnit().getName() );
        builder.append( "\n\n" );

        // Choose kind of code unit
        ArrayLiteral.setCArrayDelimiters();
        if ( this.getCodeUnit() instanceof Prg ) {
            this.generateProg();
        } else {
            this.generateModule();
        }

        return builder.toString().getBytes();
    }

    /**
     * Generates code for a program
     */
    private void generateProg() throws SemanticError
    {
        generateBody( this.getCodeUnit().getInstrs() );
    }

    /**
     * Generates code for a given module
     */
    private void generateModule() throws SemanticError
    {
        // Generate class
        builder.append( "class " );
        builder.append( this.getCodeUnit().getName() );
        builder.append( "{\n" );

        // Generate all constants
        for(Fin constant: this.getCodeUnit().getConstants()) {
            generateModuleConstant( constant );
        }
        
        // Generate all functions
        for(Fun function: this.getCodeUnit().getFunctions())
        {
            generateFunction( function, true );
        }

        builder.append("}\n" );
    }

    /**
     * Generates code for a given function
     * @param isModule Generate or not visibility attributes.
     * @param fun The function to generate code for.
     */
    public static void generateFunction(Fun fun, boolean isModule) throws SemanticError
    {
        // Generate visibility
        if ( isModule ) {
            if ( fun.getVisibility().isVisible() ) {
                builder.append( Reserved.Public );
            } else {
                builder.append( Reserved.Private );
            }

            builder.append( " static " );
        }

        // Generate type
        builder.append( Reserved.getType( fun.getType() ) );
        builder.append( ' ' );

        // Generate name & params
        builder.append( fun.getId().toString() );
        builder.append( '(' );

        Data[] params = fun.getParams();
        for (int i = 0; i < params.length; ++i ) {
            // Generate type & name
            builder.append( Reserved.getType( params[ i ].getType() ) );
            builder.append( ' ' );
            builder.append( params[ i ].getId() );

            // Generate comma
            if ( i < ( params.length -1 ) ) {
                builder.append( ',' );
            }

        }

        builder.append( "){\n" );
        generateBody( fun.getInstrs() );
        builder.append( "}\n" );
    }

    public static void generateBody(CodeBlock cb) throws SemanticError
    {
        for(int i = 0; i < cb.size(); ++i) {
            Instr instr = cb.get( i );

            if ( instr instanceof Fun ) {
                generateFunction( (Fun) instr, false );
            } else {
                builder.append( Opcode.getOpcode( instr ).emit() );
            }

            builder.append( '\n' );
        }
    }
    
    public static void generateModuleConstant(Fin constant) throws SemanticError
    {
        builder.append( "public static ");
        builder.append( Opcode.getOpcode( constant ).emit() );
        builder.append( '\n' );
    }

    private static StringBuilder builder = null;
}
