// BshCompiler.java

package core;

import core.exceptions.SemanticError;
import core.intermediatecode.CodeUnit;
import core.intermediatecode.Lib;
import core.intermediatecode.Module;
import core.lexical.Position;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * Drives the compilation process into BSH (BeanShell's backend).
 * The compilation process accomplishes a few phases:
 * - First phase: parse all imports and load all libraries
 * - Second phase: trigger parsing in all loaded modules
   - Third phase: Semantic verification for all modules
 * - Fourth phase: Generate backend code for all modules
 * @author baltasarq
 */
public class BshCompiler extends Compiler {

    public static final String ExecutableExt = "jcb";

    /**
     * Creates a new instance. See Compiler.
     * @param f The input file.
     * @param mode Parsing mode (strictness...)
     * @param language Language chosen.
     * @see Compiler
     * @throws SemanticError
     */
    public BshCompiler(File f, Parser.Mode mode, I18n.Language language)
            throws SemanticError
    {
        super( f, mode, language );
    }

    /**
     * Returns a backend adapted for the compilation needs.
     * @param cdu The CodeUnit to create the backend for.
     * @return A new BshBackend object.
     */
    @Override
    public BshBackend createBackend(CodeUnit cdu)
    {
        return new BshBackend( cdu );
    }

    @Override
    public void compile() throws CompilingError, IOException
    {

        super.compile();

        // Code generation
        try (FileOutputStream output = new FileOutputStream(this.getOutFile()))
        {
            output.write( this.buildPreface().getBytes() );
            output.write( this.getGeneratedCode() );
        }

    }

    /**
     * Adds the prefix to the compiled code.
     * This is done after compilation, and after semantic verification
     * The prefix makes it possible to load the required libraries
     * @throws SemanticError in case something goes wrong
     */
    public String buildPreface() throws SemanticError
    {
        StringBuilder preface = new StringBuilder();

        try {
            // Build path to base libraries
            File libDir = LibLoader.getLibDir().getCanonicalFile();

            for(Module mod: Ast.getAst().getModules())
            {
                // Locate all libraries
                if ( mod instanceof Lib ) {
                    Lib lib = (Lib) mod;
                    File libFile = new File( new File( libDir, lib.getPrefix() ), lib.getName() );

                    // Build the addClassPath
                    preface.append( "addClassPath(\"");
                    preface.append( libFile.getCanonicalPath().replace( "\\", "\\\\" ) );
                    preface.append( '.' );
                    preface.append( LibLoader.LibraryFileExt );
                    preface.append( "\");\n" );

                    // Build the second addClassPath, in case program is standalone
                    preface.append( "addClassPath(\"");
                    preface.append( lib.getName() );
                    preface.append( '.' );
                    preface.append( LibLoader.LibraryFileExt );
                    preface.append( "\");\n" );
                }

            }

            preface.append( '\n' );
        }
        catch (IOException ex) {
            throw new SemanticError(
                    new Position( "ast", 0, 0 ),
                    "jcg.BshCompiler.buildPreface(): impossible to create preface, path to library not found"
            );
        }

        return preface.toString();
    }

    /**
     * Returns the extension for files that will be executable
     * @return The extension, as String.
     */
    @Override
    public String getExecutableExtension()
    {
        return ExecutableExt;
    }
}
