// Compiler.java

package core;

import core.I18n.Language;
import core.Parser.Mode;
import core.exceptions.SemanticError;
import core.exceptions.SyntaxError;
import core.intermediatecode.*;
import core.intermediatecode.instrs.Cal;
import core.intermediatecode.instrs.Def;
import core.intermediatecode.instrs.Fun;
import core.intermediatecode.instrs.VarAsg;
import core.intermediatecode.types.IntType;
import core.lexical.Reserved;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

/**
 * Represents the compilation workflow
 * @author baltasarq
 */
public abstract class Compiler {

    public static final String ModuleFileExt = Module.ModuleFileExt;
    public enum OutputPlatform {
        Bsh,
        Js;

        public static OutputPlatform parse(String txt)
        {
            OutputPlatform toret = null;
            txt = txt.trim();

            for(OutputPlatform m: OutputPlatform.values() ) {
                if ( m.toString().equalsIgnoreCase( txt ) ) {
                    toret = m;
                    break;
                }
            }

            if ( toret == null ) {
                throw new IllegalArgumentException( txt );
            }

            return toret;
        }
    }

    public Compiler(File f, Parser.Mode mode, I18n.Language language) throws SemanticError
    {
        Ast.reset();
        LibLoader.reset();

        this.source = f;
        this.mode = mode;
        this.language = language;
        this.parser = new Parser( f, mode, language );
        this.createOutputFile();
    }

    /**
     * Triggers the whole compilation process.
     * This is a very basic implementation. Create your own subclass and
     * overwrite it, calling super if you wish.
     */
    public void compile() throws CompilingError, IOException
    {
        this.parse();
        this.chk();
        this.injectColophon();
        this.generateBackend();
        this.generateDoc();
    }

    /**
     * Sets the path to the library directory
     * Take into account that it first tries to locate the library in the current directory
     * @param libDir The new library directory
     */
    public void setLibPath(File libDir)
    {
        this.libDir = libDir;

        if ( !this.libDir.exists() ) {
            this.libDir = new File( ".", LibLoader.LibraryPathPrefix );
        }

        LibLoader.setLibDir( this.libDir );
    }

    /**
     * Generates the documentation for the project.
     */
    protected void generateDoc() throws IOException
    {
        final Ast ast = Ast.getAst();

        for(CodeUnit cdu: ast.getCodeUnits()) {
            if ( !( cdu instanceof Lib ) ) {
                try (FileWriter outFileWriter
                             = new FileWriter( cdu.buildDocFileName( CodeUnit.IncludePath ) ) )
                {
                    outFileWriter.write( cdu.toHtml() );
                    outFileWriter.close();
                }
            }
        }

        return;
    }

    /**
     * Creates the Ast from the source file
     * @return The global Ast
     */
    public Ast parse() throws CompilingError
    {
        final Ast ast = Ast.getAst();

        // Parse file (this will trigger parser for imports)
        this.parser.parse();

        return ast;
    }

    public void chk() throws CompilingError
    {
        SemanticChecker checker = new SemanticChecker(
                Ast.getAst(),
                this.parser.getLexer().getPos(),
                this.parser.getMode() );

        checker.chk();
        return;
    }

        /**
     * Adds the colophon to the compiled code.
     * This is done after parsing.
     * The colophon recovers the end error code, in case the function main
     * was created.
     * @throws SyntaxError in case something goes wrong
     */
    public void injectColophon() throws SyntaxError
    {
       final Lexer lex = this.parser.getLexer();
       final Prg prg = Ast.getAst().getPrg();
       final Fun[] fnsMain = prg.getFunctionByName( Reserved.EntryPointFn );

       if ( fnsMain != null ) {
           final Fun fnMain = fnsMain[ 0 ];
           Cal cal = new Cal( lex.getPos(), Reserved.EntryPointFn );
           cal.setOwner( prg );
           cal.setLinkedFun( fnMain );

           prg.getInstrs().add( new Def( lex.getPos(), Reserved.ExitCodeVar, IntType.get() ) );

          if ( fnMain.getType() == IntType.get() ) {
               prg.getInstrs().add( new VarAsg( lex.getPos(), Reserved.ExitCodeVar, cal ) );
          } else {
               RefId cero = new RefId( prg.literalCache.get( 0 ) );
               RefId uno = new RefId( prg.literalCache.get( 1 ) );

               prg.getInstrs().add( new VarAsg( lex.getPos(), Reserved.ExitCodeVar, uno ) );
               prg.getInstrs().add( cal );
               prg.getInstrs().add( new VarAsg( lex.getPos(), Reserved.ExitCodeVar, cero ) );

          }
       }

       return;
    }

    /**
     * Returns the generated code for the chosen backend.
     * Run over all code units, and call the BshBackend for them
     * @return The result of the backend code generator.
     */
    public byte[] generateBackend() throws SemanticError
    {
        byte[] bytes;
        int length = 0;
        Backend bshBackend;
        Ast ast = Ast.getAst();
        ArrayList<byte[]> objectCode = new ArrayList<byte[]>();

        // Generate the backend for all modules
        for(Module module: ast.getModules()) {
            if ( !( module instanceof Lib ) ) {
                bshBackend = this.createBackend( module );
                bytes = bshBackend.generate();
                length += bytes.length;
                objectCode.add( bytes );
            }
        }

        // Generate the backend code for the program
        bshBackend = this.createBackend( ast.getPrg() );
        bytes = bshBackend.generate();
        objectCode.add( bytes );
        length += bytes.length;

        // Generate the whole backend
        this.generatedCode = new byte[ length ];
        int pos = 0;
        for(byte[] moduleBytes: objectCode) {
            for(int i = 0; i < moduleBytes.length; ++i) {
                this.generatedCode[ pos ] = moduleBytes[ i ];
                ++pos;
            }
        }

        return this.getGeneratedCode();
    }

    /**
     * Create the adapted backend for this compiler.
     * @param cdu The CodeUnit to create the backend for.
     * @return A new backend object, deriving from Backend.
     */
    public abstract Backend createBackend(CodeUnit cdu);

    /**
     * Creates the output file needed.
     */
    protected final File createOutputFile()
    {
        final String ExecutableExt = this.getExecutableExtension();
        String outFileName = this.getInFile().getAbsolutePath();

        if ( outFileName.length() > 0 ) {
            int dotPos = outFileName.lastIndexOf( '.' );

            if ( dotPos >= 0 ) {
                outFileName = outFileName.substring( 0, dotPos +1 );
                outFileName += ExecutableExt;
            }
        } else {
            outFileName = "out." + ExecutableExt;
        }

        this.out = new File( outFileName );
        return this.out;
    }

    /**
     * Returns the source file
     * @return The input file, as a File object.
     */
    public File getInFile()
    {
        return this.source;
    }

    /**
     * Returns the output file
     * @return The output file, as a File object.
     */
    public File getOutFile()
    {
        return this.out;
    }

    /**
     * Returns the extension needed for output files.
     * @return The extension, as String.
     */
    public abstract String getExecutableExtension();

    /**
     * Returns the code generated by generateBackend()
     * @return the code generated by generateBackend(), as byte[]
     */
    public byte[] getGeneratedCode()
    {
        return this.generatedCode;
    }

    /**
     * Returns the language for error reporting in this compilation
     * @return The language chosen, as an I18n.Language object.
     */
    public Language getLanguage()
    {
        return this.language;
    }

    /**
     * Returns the mode chosen for compilation.
     * @return The mode chosen, as a Parser.Mode object.
     */
    public Mode getMode()
    {
        return this.mode;
    }

    protected File source;
    protected File libDir;
    protected File out;
    protected Parser parser;
    protected Parser.Mode mode;
    protected I18n.Language language;
    protected byte[] generatedCode;
}
