// JsBackend.java

package core;

import core.intermediatecode.literals.ArrayLiteral;
import core.jsbackend.Opcode;
import core.jsbackend.Reserved;
import core.exceptions.SemanticError;
import core.intermediatecode.*;
import core.intermediatecode.instrs.Fin;
import core.intermediatecode.instrs.Fun;

/**
 * A backend for generating JavaScript code.
 * @author baltasarq
 */
public class JsBackend extends Backend {

    public JsBackend(CodeUnit cdu)
    {
        super( cdu );
        builder = new StringBuilder();
    }

    @Override
    public byte[] generate() throws SemanticError
    {
        // Clean
        builder.delete( 0, builder.length() );

        // Add title
        builder.append( Reserved.CommentMark );
        builder.append( ' ' );
        builder.append( this.getCodeUnit().getName() );
        builder.append( "\n\n" );

        // Choose kind of code unit
        ArrayLiteral.setJSArrayDelimiters();
        if ( this.getCodeUnit() instanceof Prg ) {
            this.generateProg();
        } else {
            this.generateModule();
        }

        ArrayLiteral.setCArrayDelimiters();
        return builder.toString().getBytes();
    }

    /**
     * Generates code for a program
     */
    private void generateProg() throws SemanticError
    {
        generateBody( this.getCodeUnit().getInstrs() );
    }

    /**
     * Generates code for a given module
     */
    private void generateModule() throws SemanticError
    {
        final String ModuleName = this.getCodeUnit().getName();

        builder.append( Reserved.CommentMark );
        builder.append( ModuleName );
        builder.append( '\n' );

        // Generate closure
        builder.append( "var " );
        builder.append( ModuleName );
        builder.append( Reserved.AssignOperator );
        builder.append( '(' );
        builder.append( Reserved.FunctionMark );
        builder.append( "(){\n" );

        // Generate all constants
        for(Fin constant: this.getCodeUnit().getConstants()) {
            builder.append( Opcode.getOpcode( constant ).emit() );
            builder.append( '\n' );
        }
        
        // Generate all functions
        for(Fun function: this.getCodeUnit().getFunctions())
        {
            generateFunction( function, true );
        }

        // Generate public members
        builder.append( '\n' );
        builder.append( Reserved.Return );
        builder.append( " {" );

        for(Member m: this.getCodeUnit().getMembers()) {
            final String memberId = m.getId().getId();

            if ( m.isVisible() ) {
                builder.append( '"' );
                builder.append( memberId );
                builder.append( '"' );
                builder.append( Reserved.SecondAssignOperator );
                builder.append( memberId );
                builder.append( ',' );
            }
        }


        builder.append( "};\n})();\n" );
    }

    /**
     * Generates code for a given function
     * @param isModule Generate or not visibility attributes.
     * @param fun The function to generate code for.
     */
    public static void generateFunction(Fun fun, boolean isModule) throws SemanticError
    {
        // Generate function name & params
        builder.append( Reserved.FunctionMark );
        builder.append( ' ' );
        builder.append( fun.getId().toString() );
        builder.append( '(' );

        Data[] params = fun.getParams();
        for (int i = 0; i < params.length; ++i ) {
            // Generate name
            builder.append( params[ i ].getId() );

            // Generate comma
            if ( i < ( params.length -1 ) ) {
                builder.append( ',' );
            }

        }

        builder.append( "){\n" );
        generateBody( fun.getInstrs() );
        builder.append( "}\n" );
    }

    public static void generateBody(CodeBlock cb) throws SemanticError
    {
        for(int i = 0; i < cb.size(); ++i) {
            Instr instr = cb.get( i );

            if ( instr instanceof Fun ) {
                generateFunction( (Fun) instr, false );
            } else {
                builder.append( Opcode.getOpcode( instr ).emit() );
            }

            builder.append( '\n' );
        }
    }
    
    private static StringBuilder builder = null;
}
