// JSCompiler.java

package core;

import core.exceptions.SemanticError;
import core.intermediatecode.CodeUnit;
import core.intermediatecode.Lib;
import core.intermediatecode.Module;
import core.lexical.Position;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * Drives the compilation process into JavaScript (with a small HTML launcher).
 * The compilation process accomplishes a few phases:
 * - First phase: parse all imports and load all libraries
 * - Second phase: trigger parsing in all loaded modules
 * - Third phase: Semantic verification for all modules
 * - Fourth phase: Generate backend code for all modules
 * @author baltasarq
 * @date 9/12/13
 */
public class JsCompiler extends Compiler {
    public static final String ExecutableExt = "html";
    public static final String ExecutableLibExt = "js";

    /**
     * Creates a new instance. See Compiler.
     * @param f The input file.
     * @param mode Parsing mode (strictness...)
     * @param language Language chosen.
     * @see Compiler
     * @throws core.exceptions.SemanticError
     */
    public JsCompiler(File f, Parser.Mode mode, I18n.Language language)
            throws SemanticError
    {
        super( f, mode, language );
    }

    /**
     * Returns a backend adapted for the compilation needs.
     * @param cdu The CodeUnit to create the backend for.
     * @return A new BshBackend object.
     */
    @Override
    public JsBackend createBackend(CodeUnit cdu)
    {
        return new JsBackend( cdu );
    }

    @Override
    public void compile() throws CompilingError, IOException
    {

        super.compile();

        // Stub + code generation
        try (FileOutputStream output = new FileOutputStream( this.getOutFile() ))
        {
            output.write( this.buildStub().getBytes() );
            output.write( this.getGeneratedCode() );
            output.write( "\n</script></div></body></html>\n".getBytes() );
        }

        return;
    }

    /**
     * Adds the prefix to the compiled code.
     * This is done after compilation, and after semantic verification
     * The prefix makes it possible to load the required libraries
     * @throws SemanticError in case something goes wrong
     */
    public String buildStub() throws SemanticError
    {
        String name = Ast.getAst().getPrg().getName();
        StringBuilder stub = new StringBuilder();

        // Prepare name
        name = name.substring( 0, 1 ).toUpperCase() + name.substring( 1 ).toLowerCase();

        // Create Stub
        stub.append( "<html>\n<head><title>" );
        stub.append( name );
        stub.append( "</title>\n" );
        stub.append( "<meta content=\"text/html; charset=utf-8\" http-equiv=\"content-type\">" );
        stub.append( "<meta http-equiv=\"cache-control\" content=\"max-age=0\" />\n" );
        stub.append( "<meta http-equiv=\"cache-control\" content=\"no-cache\" />\n" );
        stub.append( "<meta http-equiv=\"expires\" content=\"0\" />\n" );
        stub.append( "<meta http-equiv=\"expires\" content=\"Tue, 01 Jan 1980 1:00:00 GMT\" />\n" );
        stub.append( "<meta http-equiv=\"pragma\" content=\"no-cache\" />\n" );
        stub.append( "</head>\n<body>\n" );
        stub.append( "<h1>" );
        stub.append( name );
        stub.append( "</h1>\n<div id='app-src'>" );

        try {
            // Build path to base libraries
            File libDir = LibLoader.getLibDir().getCanonicalFile();

            for(Module mod: Ast.getAst().getModules())
            {
                // Locate all libraries
                if ( mod instanceof Lib) {
                    Lib lib = (Lib) mod;
                    File libFile = new File( new File( libDir, lib.getPrefix() ), lib.getName() );

                    // Build the addClassPath
                    stub.append( "<script language=\"JavaScript\" src=\"" );
                    stub.append( libFile.getCanonicalPath().replace("\\", "\\\\" ) );
                    stub.append( '.' );
                    stub.append( ExecutableLibExt );
                    stub.append( "\"></script>\n" );

                    // Build the second addClassPath, in case program is standalone
                    stub.append( "<script language=\"JavaScript\" src=\"" );
                    stub.append( lib.getName() );
                    stub.append( '.' );
                    stub.append( ExecutableLibExt );
                    stub.append( "\"></script>\n" );
                }

            }

            stub.append( "\n<script>\n" );
        }
        catch (IOException ex) {
            throw new SemanticError(
                    new Position( "ast", 0, 0 ),
                    "jcg.JsCompiler.buildStub(): impossible to create preface, path to library not found"
            );
        }

        return stub.toString();
    }

    /**
     * Returns the extension for files that will be executable
     * @return The extension, as String.
     */
    @Override
    public String getExecutableExtension()
    {
        return ExecutableExt;
    }
}
