// Lexer.java

package core;

import core.exceptions.SyntaxError;
import core.exceptions.semantic.FileNotFound;
import core.lexical.Id;
import core.lexical.Position;
import core.lexical.Reserved;
import java.io.*;

/**
 * The Lexer class, that gets the parts of the source code
 * with the first interpretation of their role
 *
 * @author baltasarq
 */
public final class Lexer {
    public final String Blanks = "\t \n\r";
    public final String Delimiters = "(){}";
    public final String StringLiteralMark = "\"";
    public final String CharLiteralMark = "'";

    public enum TokenType {
        Id, Delimiter,
        NumberLiteral, StringLiteral, CharLiteral, BoolLiteral,
        Invalid
    }

    public Lexer(File f) throws FileNotFound, IOException
    {
        this.f = f;
        this.inLineNumber = 1;
        this.inColNumber = 1;
        this.inFileReported = f.getName();

        try {
            this.input = new BufferedReader( new FileReader( f ) );
        }
        catch (FileNotFoundException exc)
        {
            throw new FileNotFound( this.getPos(), f.getName() );
        }
    }

    public char read()
    {
        int toret;

        try {
            if ( this.isEOF() ) {
                throw new EOFException();
            }

            toret = this.input.read();

            // Update line/col count
            ++this.inColNumber;
            if ( ( (char) toret ) == '\n' ) {
                this.incLineNumber();
            }

            // Check for EOF
            if ( toret == -1 ) {
                throw new EOFException();
            }
        }
        catch(Exception exc)
        {
            this.close();
            toret = '\0';
        }

        return ( (char) toret );
    }

    /**
     * Returns the file name that is going to be reported
     * as being scanned. Not necessarily the one being scanned (e.g. modules)
     * @return The file reportedly being scanned, as a String.
     */
    public String getInFileReported()
    {
        return this.inFileReported;
    }

    /**
     * Returns the file being scanned
     * @return The file being scanned, as a File.
     */
    public File getInFile()
    {
        return this.f;
    }

    /**
     * Returns the current pos in the file, useful for reporting errors.
     * @return The current pos in the file, as a Position object.
     */
    public final Position getPos()
    {
        return new Position( this.getInFileReported(),
                             this.inLineNumber,
                             this.inColNumber
               );
    }

    /**
     * Closes the input file.
     */
    public void close()
    {
        try {
            if ( this.input != null ) {
                this.input.close();
            }
        } catch(Exception ignored)
        {
        }
        finally {
            this.input = null;
        }
    }

    /**
     * Determines whether the end of the file has been reached.
     * @return true if EOF has been reached, false otherwise.
     */
    public boolean isEOF()
    {
        return ( this.input == null );
    }


    /**
     * Returns the next id available.
     * @return The next id as a String, or an empty String if a special character
     *         is found.
     */
    public String readToken() throws CompilingError
    {
        char ch;
        StringBuilder toret = new StringBuilder();

        this.skipBlanks();
        while( !this.isEOF() ) {
            ch = this.peekNextChar();

            if ( Blanks.indexOf( ch ) > -1
              || ( !Character.isLetterOrDigit( ch )
              && ch != '_' ) )
            {
                break;
            }
            else {
                toret.append( this.read() );
            }
        }

        this.lastToken = toret.toString();
        return this.lastToken;
    }

    /**
     * Returns the last token read
     * @return The last token read, as a String
     */
    public String getLastToken()
    {
        return this.lastToken;
    }

    /**
     * Skips the next blank chars.
     */
    public void skipBlanks() throws CompilingError
    {
        char ch;

        try {
            while( !this.isEOF() ) {
                ch = this.peekNextChar();

                if ( ch == '/' ) {
                    char ch2;

                    // Read three chars ahead
                    this.mark( 3 );
                    this.read();
                    ch = this.read();
                    ch2 = this.read();
                    this.reset();

                    if ( ch == '/' ) {
                        this.skipSingleLineComment();
                    }
                    else
                    if ( ch == '*'
                      && ch2 != '*' )
                    {
                        this.skipMultipleLineComment();;
                    }
                    else {
                        break;
                    }
                }
                else
                if ( Blanks.indexOf( ch ) > -1 ) {
                    this.read();
                } else {
                    break;
                }
            }
        }
        catch(NullPointerException | IOException exc)
        {
            this.close();
        }

        return;
    }

    /**
     * Reads a whole line from current position until end-
     * @return The whole line beginning in current position, as String.
     */
    public String readLine()
    {
        String toret = null;

        try {
            toret = this.input.readLine();
            this.incLineNumber();
            this.mark( 2 );
            this.read();
            this.reset();
        } catch (IOException ex) {
            this.close();
        }

        return toret;
    }

    /**
     * Skips the current comment until the end of the line
     * @return The comment skipped, as a String.
     */
    public String skipSingleLineComment() throws CompilingError
    {
        String toret = "";
        char ch = this.read();

        if ( ch == '/'
          && this.peekNextChar() == '/' )
        {
            // Pass the second '/'
            this.read();
            toret = this.readLine();
        } else {
            throw new CompilingError( this.getPos(),
                    "jcg.Lexer.skipSingleLineComment(): not a comment" );
        }

        return toret;
    }

    /**
     * Skips the current comment until its end
     * @return The comment skipped, as a String.
     */
    public String skipMultipleLineComment() throws CompilingError
    {
        String toret = "";

        char ch = this.read();

        if ( ch == '/' ) {
            ch = this.read();

            if ( ch == '*' ) {
                if ( this.peekNextChar() != '*' ) {
                    // Skip '*', this is a doc-comment
                    this.read();
                }

                toret = this.skipUntil( Reserved.EndMultiLineCommentMark );
            } else {
                throw new CompilingError( this.getPos(),
                        "jcg.Lexer.skipMultipleLineComment(): not a comment" );
            }
        } else {
            throw new CompilingError( this.getPos(),
                    "jcg.Lexer.skipMultipleLineComment(): not a comment" );
        }

        return toret;
    }

    /**
     * Skips the 'n' next chars.
     */
    public void advance(int delta) throws CompilingError
    {
        if ( delta < 0 ) {
            throw new CompilingError( this.getPos(),
                    "jcg.Lexer.advance(): trying to advance backwards" );
        }

        for(int i = 0; i < delta; ++i) {
            this.read();
        }
    }

    /**
     * Skips all chars until a given mark is found-
     * @param mark The mark to be found
     * @throws CompilingError Will neve be thrown, unless mark.length is < 0
     */
    public String skipUntil(String mark) throws CompilingError
    {
        StringBuilder toret = new StringBuilder();

        // Skip all chars until mark
        while( !this.isEOF()
            && !this.isStringAhead( mark ) )
        {
            toret.append( this.read() );
        }

        // Skip mark
        if ( !this.isEOF() ) {
            this.advance( mark.length() );
        }

        return toret.toString();
    }

    /**
     * Returns the next character, without consuming it.
     * @return The next character, as a char
     */
    public char peekNextChar()
    {
        char toret;

        try {
            if ( this.isEOF() ) {
                throw new EOFException();
            }

            this.mark( 2 );
            toret = this.read();
            this.reset();
        } catch(Exception exc)
        {
            this.close();
            toret = '\0';
        }

        return toret;
    }

    /**
     * Determines whether there is (or not) a literal ahead
     * @return true when there is a literal, false otherwise
     */
    public boolean isLiteralAhead() throws CompilingError
    {
        return ( this.peekNextTokenType() == Lexer.TokenType.CharLiteral
              || this.peekNextTokenType() == Lexer.TokenType.NumberLiteral
              || this.peekNextTokenType() == Lexer.TokenType.StringLiteral
              || this.peekNextTokenType() == Lexer.TokenType.BoolLiteral );
    }

    /**
     * Determines whether a determines mark is ahead
     * @param mark The mark to look for, as a String
     * @return true when it is ahead, false otherwise
     */
    public boolean isStringAhead(String mark)
    {
        boolean toret;

        try {
            if ( this.isEOF() ) {
                throw new EOFException();
            }

            this.mark( mark.length() +1 );
            StringBuilder compare = new StringBuilder();

            // Read chars
            compare.append( this.read() );
            while( compare.length() < mark.length() ) {
                compare.append( this.read() );
            }

            // Compare
            toret = compare.toString().equals( mark );

            // End
            this.reset();
        } catch(Exception exc)
        {
            this.close();
            toret = false;
        }

        return toret;
    }

    /**
     * Determines whether there is or not a call to a function ahead
     * @return true if there is a call to a function, false otherwise
     */
    public boolean isCallAhead()
    {
        char ch;
        boolean toret = false;
        boolean thereIsBracket;
        StringBuilder id1 = new StringBuilder();
        StringBuilder id2 = new StringBuilder();

        try {
            if ( this.isEOF() ) {
                throw new EOFException();
            }

            this.skipBlanks();
            this.mark();

            // Read the first id of the complete RefId
            ch = this.read();
            while( ( Character.isLetterOrDigit( ch )
                  || ch == '_' )
                && !this.isEOF() )
            {
                id1.append( ch );
                ch = this.read();
            }

            // Pass spaces
            while( Blanks.indexOf( ch ) >= 0
                && !this.isEOF() )
            {
                ch = this.read();
            }

            thereIsBracket = ( ch == '(' );

            if ( ch == '.'
              && !thereIsBracket )
            {
                // Pass spaces
                ch = this.read();
                while( Blanks.indexOf( ch ) >= 0
                    && !this.isEOF() )
                {
                    ch = this.read();
                }

                // Read second id of the complete RefId
                while( ( Character.isLetterOrDigit( ch )
                      || ch == '_' )
                    && !this.isEOF() )
                {
                    id2.append( ch );
                    ch = this.read();
                }

                thereIsBracket = ( ch == '(' );
            }

            // Maybe there are some spaces before the bracket
            if ( !thereIsBracket ) {
                while( Blanks.indexOf( ch ) >= 0
                    && !this.isEOF() )
                {
                    ch = this.read();
                }
                thereIsBracket = ( ch == '(' );
            }

            // Everything correct, provided there is a bracket, and valid ids
            if ( thereIsBracket ) {
                toret = true;
                // Check whether they are valid id's
                try {
                    Id.chkId( id1.toString() );

                    if ( id2.length() > 0 ) {
                        Id.chkId( id2.toString() );
                    }
                }
                catch(Exception exc) {
                    toret = false;
                }
            }

            // End
            this.reset();
        } catch(CompilingError | IOException exc)
        {
            this.close();
            toret = false;
        }

        return toret;
    }

    /**
     * Returns the next token type
     * @return The next token type, as a TokenType object.
     */
    public TokenType peekNextTokenType() throws CompilingError
    {
        TokenType toret = TokenType.Invalid;
        char ch;

        this.skipBlanks();
        ch = this.peekNextChar();

        if ( Character.isDigit( ch ) ) {
            toret = TokenType.NumberLiteral;
        }
        else
        if ( Character.isLetter( ch )
          || ch == '_')
        {
            toret = TokenType.Id;

            String token = this.peekNextToken();

            if ( token.equals( Reserved.True )
              || token.equals( Reserved.False ) )
            {
                toret = TokenType.BoolLiteral;
            }
        }
        else
        if ( Delimiters.indexOf( ch ) > -1 ) {
            toret = TokenType.Delimiter;
        }
        else
        if ( this.isStringLiteralMark( ch ) ) {
            toret = TokenType.StringLiteral;
        }
        else
        if ( this.isCharLiteralMark( ch ) ) {
            toret = TokenType.CharLiteral;
        }

        return toret;
    }

    /**
     * Returns the next token
     * @return The next token, as a String.
     */
    public String peekNextToken() throws CompilingError
    {
        char ch;
        StringBuilder toret = new StringBuilder();

        try {

            if ( !this.isEOF() ) {
                this.skipBlanks();
                this.mark();

                while( !this.isEOF() ) {
                    ch = this.read();

                    if ( Blanks.indexOf( ch ) > -1
                      || ( !Character.isLetterOrDigit( ch )
                      && ch != '_' ) )
                    {
                        break;
                    }
                    else {
                        toret.append( ch );
                    }
                }

                this.reset();

            }

        } catch (IOException ex) {
            this.close();
        }

        return toret.toString();
    }

    public boolean isStringLiteralMark(char ch)
    {
        return compareMarks( StringLiteralMark, ch );
    }

    public boolean isCharLiteralMark(char ch)
    {
        return compareMarks( CharLiteralMark, ch );
    }

    public boolean isStringLiteralMark(String mrk)
    {
        return compareMarks( StringLiteralMark, mrk );
    }

    public boolean isCharLiteralMark(String mrk)
    {
        return compareMarks( CharLiteralMark, mrk );
    }

    public static boolean compareMarks(String str, char ch)
    {
        return compareMarks( str, Character.toString( ch ) );
    }

    public static boolean compareMarks(String str, String mark)
    {
        return ( str.equals( mark ) );
    }

    /**
     * Reads the string literal that lies ahead. Expects to find a
     * double quote mark, if not, it throws an exception.
     * @return The string literal, without quote marks.
     */
    public String readStringLiteral() throws SyntaxError
    {
        StringBuilder toret = new StringBuilder();

        if( !this.isEOF() ) {
            // Skip double quotes; if not, then stop
            if ( this.isStringLiteralMark( this.peekNextChar() ) ) {
                this.read();

                // Read the char sequence
                while( !this.isEOF() ) {
                    // ... if a double quote is not present
                    if ( !this.isStringLiteralMark( this.peekNextChar() ) ) {
                        toret.append( this.read() );
                    } else {
                        this.read();
                        break;
                    }
                }
            } else {
                throw new SyntaxError( this.getPos(), I18n.get( I18n.Word.Expected )
                        + ": '\"' literal: \"" + toret.toString() + '"'
                );
            }

        }

        return toret.toString();
    }

    /**
     * Reads the next char literal
     * @return The char literal, as a single char.
     * @throws SyntaxError if the char literal marks are missing.
     */
    public char readCharLiteral() throws SyntaxError
    {
        char toret = '\0';

        if ( !this.isEOF() ) {
            if ( this.isCharLiteralMark( this.peekNextChar() ) ) {
                this.read();                    // pass '
                toret = this.read();            // read char

                if ( !this.isCharLiteralMark( this.peekNextChar() ) ) {
                    throw new SyntaxError( this.getPos(), I18n.get( I18n.Word.Expected )
                            + ": \"'\" literal: '" + toret + '\''
                    );
                } else {
                    this.read();                // pass '
                }
            } else {
                throw new SyntaxError( this.getPos(), I18n.get( I18n.Word.Expected )
                        + ": \"'\" literal: '" + toret + '\''
                );
            }
        }

        return toret;
    }

    /**
     * Reads the next number literal
     * @return The number itself, as a string.
     * @throws SyntaxError if the number literal is not a number.
     */
    public String readNumberLiteral() throws SyntaxError
    {
        int numDots = 0;
        StringBuilder toret = new StringBuilder();

        while ( !this.isEOF()
           && ( Character.isDigit( this.peekNextChar() )
              || this.peekNextChar() == '.' ) )
        {
            if ( this.peekNextChar() == '.' ) {
                ++numDots;
            }

            toret.append( this.read() );
        }

        if ( toret.length() == 0
          || numDots > 1 )
        {
            throw new SyntaxError( this.getPos(), I18n.get( I18n.Word.Expected )
                    + ' ' + I18n.get( I18n.Word.Number )
                    + " literal"
            );
        }

        return toret.toString();
    }

    /**
     * Increments the number of lines read from the input.
     * @return The updated input line number
     */
    public long incLineNumber()
    {
        ++this.inLineNumber;
        this.inColNumber = 1;

        return this.inLineNumber;
    }

    /**
     * Mark current position, so it is possible to go back to it
     * The maximum length of advancing is calculated from the current position
     * and the length of the file.
     */
    public void mark() throws IOException
    {
        final long sizeAhead = this.getInFile().length() - this.getPos().getNumCharacter();
        this.mark( (int) sizeAhead );
    }

    /**
     * Mark current position, so it is possible to go back to it
     * @param length The length so far it is going to be advanced.
     */
    public void mark(int length) throws IOException
    {
        this.input.mark( length );

        this.oldInColNumber = this.inColNumber;
        this.oldInLineNumber = this.inLineNumber;
    }

    /**
     * Return to the previous saved position
     */
    public void reset() throws IOException
    {
        this.input.reset();

        this.inColNumber = this.oldInColNumber;
        this.inLineNumber = this.oldInLineNumber;
    }

    /**
     * Returns the next token, an operator, if it exists
     * @return The operator read, as a String
     */
    public String readOperator() throws CompilingError
    {
        StringBuilder toret = new StringBuilder();

        // Prepare to read
        this.skipBlanks();

        // Read next token
        while ( !this.isEOF()
             && Reserved.OperatorChars.indexOf( this.peekNextChar() ) > -1 )
        {
            toret.append( this.read() );
        }

        return toret.toString();
    }

    /**
     * Returns whether the next char is a binary operator or not
     * @return True if it is a binary operator, false otherwise
     */
    public boolean isBinaryOperatorAhead()
    {
        return ( Reserved.OperatorChars.indexOf( this.peekNextChar() ) > -1 );
    }

    /**
     * Returns whether there is an advanced operator ahead (++,--...)
     * @return True when there is an advanced operator ahead, false otherwise
     */
    public boolean isAdvancedOperatorAhead() throws CompilingError
    {
        StringBuilder op = new StringBuilder();
        char ch;
        int charsRead = 0;

        try {
            // Prepare to read
            this.skipBlanks();
            this.mark();

            // Read characters related to the operator
            ch = this.read();
            while( this.Blanks.indexOf( ch ) == -1
                && Reserved.OperatorChars.indexOf( ch ) >= 0
                && !this.isEOF() )
            {
                ++charsRead;
                op.append( ch );
                ch = this.read();
            }

            // Go back and end
            this.reset();
        }
        catch(IOException exc) {
            this.close();
        }

        return ( charsRead > 0
              && Reserved.AdvancedOperators.contains( ' ' + op.toString() + ' ' ) );
    }

    /**
     * Returns the next operator, without consuming it.
     * @return The next operator, as a String
     */
    public String peekNextOperator() throws CompilingError
    {
        StringBuilder toret = new StringBuilder();

        try {
            this.skipBlanks();
            this.mark();

            char ch = this.read();
            while ( Reserved.OperatorChars.indexOf( ch ) > -1 ) {
                toret.append( ch );
                ch = this.read();
            }

            this.reset();
        } catch(IOException exc)
        {
            this.close();
        }

        return toret.toString();
    }

    /**
     * Checks whether the string passed is a conditional operator or not.
     * @param op The String to check
     * @return true if it is a conditional operator, false otherwise
     */
    public static boolean isConditionalOperator(String op)
    {
        return ( Reserved.ConditionalOperators.indexOf( op.trim() ) > -1 );

    }

    public static boolean isPreferencedConditionalOperator(String op)
    {
        return ( Reserved.PreferencedConditionalOperator.contains( op.trim() ) );
    }

    /**
     * Change the path associated with this lexer
     * Note that the file being read does not change
     * @param f A new object File.
     */
    public void setReportedFilePath(File f)
    {
        this.inFileReported = f.getAbsolutePath();
    }

    /**
     * Change the line reported to be read
     * Note that this does no change the real line number to be read.
     * @param ln The line number, as long.
     */
    public void setReportedLineNumber(long ln)
    {
        this.inLineNumber = ln;
        this.inColNumber = 1;
    }

    private BufferedReader input = null;
    private File f = null;
    private long inLineNumber = 1;
    private int inColNumber = 1;
    private String inFileReported;
    private String lastToken;

    private long oldInLineNumber;
    private int oldInColNumber;
}
