// libloader.jar
// Loading libraries in the form of jar files.

package core;

import core.exceptions.SemanticError;
import core.exceptions.syntax.IncorrectId;
import core.exceptions.SyntaxError;
import core.intermediatecode.CodeUnit;
import core.intermediatecode.instrs.Def;
import core.intermediatecode.instrs.Fin;
import core.intermediatecode.instrs.Fun;
import core.intermediatecode.Lib;
import core.intermediatecode.Literal;
import core.intermediatecode.Type;
import core.intermediatecode.visibilities.PrivateVisibility;
import core.intermediatecode.visibilities.PublicVisibility;
import core.intermediatecode.visibilities.Visibility;
import core.lexical.Position;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashMap;

/**
 * Loads a library in the form of a compiled jar library
 * @author baltasarq
 */
public class LibLoader {
    public static String LibraryPathPrefix =  "lib";
    public static final String LibraryFileExt = "jar";
    public static final String RequisiteStringField = "_jc_lib_req";
    
    /**
     * Creates a library loader for a given File that is a jar file
     * @param f The file that is a jar File
     * @throws SemanticError when the library is not found in the jar file
     */
    protected LibLoader(File f) throws SemanticError
    {
        // Set attributes 
        this.inFile = f;
        this.libName = this.extractNameOfLibFileName( this.getInFile() );
        this.libPrefix = this.extractPrefixOfLibFileName( this.getInFile() );

        // Load the class and register it
        try {
            ClassLoader cl = new URLClassLoader( new URL[] { f.toURI().toURL() } );

            // Load class
            this.classMirror = cl.loadClass( this.getLibName() );

            // Register module
            this.lib = (Lib) Ast.getAst().addCodeUnit( new Lib( this.getLibPrefix(), f ) );

            // Check dependencies
            this.chkDepends( f );
        }
        catch(SemanticError exc) {
            throw exc;
        }
        catch(ClassNotFoundException exc) {
            throw this.createLibNotFound( f, this.getLibName() );
        }
        catch(MalformedURLException exc) {
            throw this.createLibNotFound( f, this.getLibName() );
        }
    }

    /**
     * Checks the prerequisites
     * @throws ClassNotFoundException If there is not a requisite lib loaded.
     */
    private void chkDepends(File f) throws SemanticError
    {
        final Ast ast = Ast.getAst();
        String strReq;
        String[] requisiteMods;

        try {
            Field fldReq = this.classMirror.getDeclaredField( RequisiteStringField );

            if ( fldReq != null ) {
                fldReq.setAccessible( true );
                strReq = (String) fldReq.get( null );
                requisiteMods = strReq.split( ";" );

                for(String reqMod: requisiteMods) {
                    reqMod = reqMod.trim();

                    if ( ast.lookForModule( reqMod ) == null ) {
                        throw this.createLibNotFound( f, reqMod );
                    }
                }
            }
        } catch(NoSuchFieldException | IllegalAccessException ignored) {

        }
    }

    /**
     * Returns the name of the lib's file name
     * i.e., for lib/std/io.jar, it returns "io"
     * @param libFile The path to the lib's jar file
     * @return The prefix name, as a 
     */
    private String extractNameOfLibFileName(File libFile) throws SemanticError
    {
        // Create class name, by erasing the extension
        String toret = CodeUnit.getFileNameWithoutExtension( libFile );
        
        if ( toret.isEmpty() ) {
            throw this.createLibNotFound( this.getInFile(), "?" );
        }
        
        return toret;
    }
    
    /**
     * Returns the prefix just before the lib's file name
     * i.e., for lib/std/io.jar, it returns "std"
     * @param libFile The path to the lib's jar file
     * @return The prefix name, as a 
     */
    private String extractPrefixOfLibFileName(File libFile) throws SemanticError
    {
        String name;
        
        try {
            name = libFile.getCanonicalPath();
            int lastSlashPos = name.lastIndexOf( File.separatorChar );

            if ( lastSlashPos >= 0 ) {
                name = name.substring( 0, lastSlashPos );
                lastSlashPos = name.lastIndexOf( File.separatorChar );

                if ( lastSlashPos >= 0 ) {
                    name = name.substring( lastSlashPos +1 );
                } else {
                    throw new SyntaxError( this.getPos(), I18n.get( I18n.Word.Expected )
                            + ": '/'" );
                }
            } else {
                throw new SyntaxError( this.getPos(), I18n.get( I18n.Word.Expected )
                        + ": '/'" );
            }
        }
        catch(IOException | SyntaxError exc) {
            throw this.createLibNotFound( this.getInFile(), this.getLibName() + "(" + exc.getMessage() + ")" );
        }        
        
        return name;
    }

    /**
     * Creates the syntax exception when the lib is not found in the jar file.
     * @param libName The name of the lib that was supposed to be found here.
     * @return The exception, as a SyntaxError object.
     */
    protected final SemanticError createLibNotFound(File f, String libName)
    {
        return new SemanticError( this.getPos(), I18n.get( I18n.Word.Missing )
                + " lib '" + libName + ": '" + f.getAbsolutePath()
        );
    }

    /**
     * Triggers the collections of fields and methods. This will be, respectively,
     * constants (public static final fields), and functions (public static methods)
     * in the jar file.
     * @throws SyntaxError When an Id was found to be incorrect.
     * @throws SemanticError When any loading error or access error occurs.
     */
    public void collectAllMembers() throws SyntaxError, SemanticError
    {
        Field[] fields = this.classMirror.getDeclaredFields();
        Method[] methods = this.classMirror.getDeclaredMethods();

        // Collect those fields that are static, final and public
        for(Field field: fields) {
            if ( Modifier.isStatic( field.getModifiers() )
              && Modifier.isFinal( field.getModifiers() )
              && this.cnvtVisibility( field ) == PublicVisibility.get() )
            {
                    this.collectConstant( field );
            }
        }

        // Collect those the methods that are static, and public
        for(Method method: methods) {
            if ( Modifier.isStatic( method.getModifiers() )
              && this.cnvtVisibility( method ) == PublicVisibility.get() )
            {
                this.collectFunction( method );
            }
        }
    }

    /**
     * Collects a field to convert it to a constant.
     * @param fld The field found as it was found in the jar file.
     * @throws SyntaxError
     * @throws SemanticError
     */
    protected void collectConstant(Field fld) throws SyntaxError, SemanticError
    {
        // Create the literal and store it
        try {
            // Find the type of the constant
            Type type = Type.cnvtJavaToType( fld.getType() );
            Literal lit = Type.createLiteral( type );
            Visibility visible = PublicVisibility.get();
            
            lit.setLiteral( fld.get( null ) );
            Fin fin = new Fin( this.getPos(), visible, fld.getName(), lit );
            this.lib.getInstrs().add( fin );
        }
        catch(IncorrectId exc) {
            throw new IncorrectId( this.getPos(),
                                    I18n.get( I18n.Sentence.IncorrectId )
                                    + ": " + I18n.get( I18n.Sentence.LoadingLibrary )
                                    + ": " + this.getLibName()
                                    + '.' + fld.getName()
                                    + ": " + exc.toString()
            );
        }
        catch(Throwable exc)
        {
            throw new SemanticError( this.getPos(),
                                    I18n.get( I18n.Sentence.LoadingLibrary )
                                    + ": " + this.getLibName()
                                    + '.' + fld.getName()
                                    + ": " + exc.toString()
             );
        }

    }

    /**
     * Collects a method from the jar file and stores it in the Ast.
     * @param mth The method to collect
     * @throws SyntaxError when the id is not correct
     * @throws SemanticError when any loading or access error occurs.
     */
    protected void collectFunction(Method mth) throws SyntaxError, SemanticError
    {
        // Find the type of the function
        Type type = Type.cnvtJavaToType( mth.getReturnType() );

        // Create the function and store it
        try {
            // Load parameters
            final int numParams = mth.getParameterTypes().length;
            Def[] params = new Def[ numParams ];

            for(int i = 0; i < params.length; ++i) {
                String paramName = "param" + Integer.toString( i +1 );
                params[ i ] =
                        new Def(
                            this.getPos(),
                            paramName,
                            Type.cnvtJavaToType( mth.getParameterTypes()[ i ] )
                        );
            }
            
            Fun fun =
                new Fun(
                    this.getPos(), this.getLib(), type, mth.getName(), params
            );
            this.lib.getInstrs().add( fun );
        }
        catch(IncorrectId exc) {
            throw new IncorrectId( this.getPos(),
                                    I18n.get( I18n.Sentence.IncorrectId )
                                    + ": " + I18n.get( I18n.Sentence.LoadingLibrary )
                                    + ": " + this.getLibName()
                                    + '.' + mth.getName()
                                    + "(): " + exc.toString()
            );
        }
        catch(Exception exc)
        {
            throw new SemanticError( this.getPos(),
                                    I18n.get( I18n.Sentence.LoadingLibrary )
                                    + ": " + this.getLibName()
                                    + '.' + mth.getName()
                                    + "(): " + exc.toString()
             );
        }
    }

    /**
     * Returns an appropriate Position object for the loading of the library.
     * @return The position, as a Position object.
     */
    protected final Position getPos()
    {
        return new Position( this.getLibName(), 0, 0 );
    }

    /**
     * Retrieves the visibility for a given member in the jar file.
     * @param member The member to process, as a Member object.
     * @return The visibility, as a Visibility object.
     */
    protected Visibility cnvtVisibility(Member member)
    {
        Visibility toret = null;

        if ( Modifier.isPublic( member.getModifiers() ) ) {
            toret = PublicVisibility.get();
        }
        else
        if ( Modifier.isPrivate( member.getModifiers() ) ) {
            toret = PrivateVisibility.get();
        }

        return toret;
    }

    /**
     * Returns the lib that this loader is loading
     * @return The library, as a Lib object.
     */
    public final Lib getLib()
    {
        return this.lib;
    }

    /**
     * Returns the File object for the library
     * @return The library, as a File object.
     */
    public final File getInFile()
    {
        return this.inFile;
    }
    
    /**
     * Returns the name of he library being loaded, without any prefix or ext.
     * It should match classLoader.getName(), provided it was possible to load it.
     * @return The name of the library, as a String.
     */
    public final String getLibName()
    {
        return this.libName;
    }
    
    /**
     * Returns the prefix of he library being loaded, typically "std"
     * @return The name of the library, as a String.
     */
    public final String getLibPrefix()
    {
        return this.libPrefix;
    }

    public static LibLoader load(String fileName) throws SemanticError
    {
        int posDot = fileName.indexOf( "." );
        String prefix = fileName.substring( 0, posDot ).trim();
        fileName = fileName.substring( posDot + 1, fileName.length() ).trim();
        fileName += "." + LibLoader.LibraryFileExt;
        fileName = libDir
                + File.separator + prefix
                + File.separator + fileName;

        final File f = new File( fileName );
        final String libName = f.getName();
        LibLoader toret = librariesIncluded.get( libName );

        if ( toret == null ) {
            toret = new LibLoader( f );
            librariesIncluded.put( libName, toret );
        }

        return toret;
    }

    public static void resetIncludedLibraries()
    {
        librariesIncluded.clear();
    }

    public static void setLibDir(File ld)
    {
        libDir = ld;
    }

    public static File getLibDir()
    {
        return libDir;
    }

    public static void reset()
    {
        librariesIncluded.clear();
    }

    private Class classMirror;
    private Lib lib;
    private File inFile;
    private String libName;
    private String libPrefix;
    private static HashMap<String, LibLoader> librariesIncluded = new HashMap<String, LibLoader>();
    private static File libDir = new File( "." + File.separator + LibraryPathPrefix );
}
