// Parser.java

package core;

import core.I18n.Language;
import core.exceptions.SemanticError;
import core.exceptions.SyntaxError;
import core.exceptions.semantic.FileNotFound;
import core.exceptions.syntax.EndOfInstruction;
import core.exceptions.syntax.IncorrectId;
import core.intermediatecode.*;
import core.intermediatecode.instrs.*;
import core.intermediatecode.literals.*;
import core.intermediatecode.types.*;
import core.intermediatecode.visibilities.*;
import core.lexical.Id;
import core.lexical.Position;
import core.lexical.Reserved;
import core.lexical.UsrId;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;

/**
 * Represents the parser.
 * - First phase: Create a lex and build the Ast
 * - Second phase: Apply checks to the Ast
 * @author baltasarq
 */

public class Parser {

    /**
     * Represents a pair of values
     * @param <K> The type of the first value
     * @param <V> The type of the second value
     */
    protected class Pair<K, V> {
        public K first;
        public V second;

        public Pair(K first, V second)
        {
            this.first = first;
            this.second = second;
        }
    }

    public enum Mode {
        Beginner,
        Expert;

        public static Mode parse(String txt)
        {
            Mode toret = null;
            txt = txt.trim();

            for(Mode m: Mode.values() ) {
                if ( m.toString().equalsIgnoreCase( txt ) ) {
                    toret = m;
                    break;
                }
            }

            if ( toret == null ) {
                throw new IllegalArgumentException( txt );
            }

            return toret;
        }
    }
    public enum State { InImports, InConstants, InFunctions, InsideMember }

    public static final String OneLineCommentMark = Reserved.OneLineCommentMark;
    public static final String MultiLineCommentMark = Reserved.MultiLineCommentMark;
    public static final String MultiLineDocCommentMark = Reserved.MultiLineDocCommentMark;
    public static final String EndMultiLineCommentMark = Reserved.EndMultiLineCommentMark;

    /**
     * Creates a Parser suitable for compiling a module
     * @param in The file to parse the module from.
     * @throws SemanticError
     */
    protected Parser(File in) throws SemanticError
    {
        this.targetCodeUnit = new Module( in );
        Ast.getAst().addModule( (Module) this.getCodeUnit() );
        this.init( in, Mode.Expert, I18n.language );
    }

    /**
     * Creates a new Parser object, associated to a file,
     * a compilation mode and a language.
     * This is the way to parse a program.
     * @param in The source file to parse.
     * @param mode Is it beginner or expert mode?
     * @param language The language to use for messages.
     * @throws SemanticError
     */
    public Parser(File in, Mode mode, Language language) throws SemanticError
    {
        this.targetCodeUnit = new Prg( in );
        Ast.getAst().addPrg( (Prg) this.getCodeUnit() );
        this.init( in, mode, language );
    }

    /**
     * Common initialization function.
     * @param in The file to parse the module from.
     * @param mode The mode of strictness.
     * @param language The language used to report error messages.
     * @throws SemanticError
     */
    public final void init(File in, Mode mode, Language language)
            throws SemanticError
    {
        this.status = State.InImports;
        this.mode = mode;
        I18n.language = language;

        this.resetDoc();

        try {
            this.lexer = new Lexer( in );
        }
        catch(IOException exc)
        {
            throw new FileNotFound( new Position( in.getAbsolutePath(), 0, 0 ),
                                    exc.getMessage() );
        }

        return;
    }
    
    /**
     * Returns the source (input) file.
     * @return The input file, as a File object.
     */
    public final File getInFile()
    {
        return this.getLexer().getInFile();
    }

    public final String getInPath()
    {
        return this.getInFile().getAbsoluteFile().getParent();
    }

    /**
     * The status of the parsing task
     * @return The status, as a State value.
     */
    public State getState()
    {
        return this.status;
    }

    /**
     * Changes the status of the parsing task.
     * @param st
     */
    protected void setState(State st)
    {
        this.status = st;
    }

    /**
     * Returns the lex being used.
     * @return The lex, as a Lexer object.
     */
    public Lexer getLexer()
    {
        return this.lexer;
    }

    /**
     * Returns the current block of code
     * @return The current block of code, as a CodeBlock object.
     */
    public CodeBlockOwner getCodeBlock()
    {
        return this.currentCodeBlock;
    }

    /**
     * Returns the target code unit.
     * @return The current code unit, as a CodeUnit object.
     */
    public CodeUnit getCodeUnit()
    {
        return this.targetCodeUnit;
    }

    /**
     * Triggers parsing.
     * @param mode The strictness mode for compilation.
     * @param language The language for message errors.
     * @throws CompilingError If source does not follow syntax/semantic rules.
     */
    public void parse(Mode mode, Language language) throws CompilingError
    {
        this.mode = mode;
        I18n.language = language;
        this.parse();
    }

    /**
     * Triggers parsing.
     * @throws CompilingError If source does not follow syntax/semantic rules.
     */
    public void parse() throws CompilingError
    {
        final Ast ast = Ast.getAst();
        final Lexer lex = this.getLexer();

        this.currentCodeBlock = this.getCodeUnit();

        while( !lex.isEOF() ) {

            lex.skipBlanks();

            if ( this.isDocCommentAhead() ) {
                this.parseDoc();
            }
            else
            if ( this.isCommentAhead() ) {
                Parser.skipComment( lexer );
            }
            else {
                if ( this.getState() != State.InsideMember )
                {
                    if ( lex.peekNextTokenType() == Lexer.TokenType.Id
                      && lex.peekNextToken().equals( Reserved.Import ) )
                    {
                        if ( this.getState() == State.InImports ) {
                            this.parseImport();
                        } else {
                            throw this.createSyntaxError(
                               I18n.get( I18n.Sentence.MisplacedImport )
                            );
                        }
                    }
                    else
                    if ( lex.peekNextTokenType() == Lexer.TokenType.Id
                      && this.isBeginningDefAhead() )
                    {
                        if ( this.getState() == State.InConstants
                          || this.getState() == State.InFunctions
                          || this.getState() == State.InImports )
                        {
                            this.parseMember();
                        }
                        else {
                            throw this.createSyntaxError(
                               I18n.get( I18n.Word.Misplaced )
                                     + ": " + lex.peekNextToken()
                            );
                        }
                    } else {
                        // Probably instructions without function
                        this.status = State.InsideMember;
                        this.parseStatements();
                    }
                } else {
                    this.parseFunctionBody();
                }
            }

        }

        return;
    }

    public static void parseModule(File fMod) throws CompilingError
    {
        Parser parser = new Parser( fMod );
        parser.parse();
    }

    /**
     * Prepares the inclusion of an import. It could be
     *  a library import or an user import.
     * @throws FileNotFound when the file is not found. SyntaxErrors when
     *         the semicolon is not found, and so on.
     */
    protected void parseImport() throws CompilingError
    {
        String fileName;
        final Lexer lex = this.getLexer();

        if ( lex.readToken().equals( Reserved.Import ) ) {
            // Manage documentation
            if ( !wasUnitDocStored ) {
                this.getCodeUnit().setDoc( this.getDoc() );
                this.resetDoc();
                this.wasUnitDocStored = true;
            }

            // Get the file path to import
            fileName = lex.readLine().trim();

            // Check for semicolon
            if ( fileName.endsWith( Reserved.EndOfInstructionMark ) ) {
                // Create library file name
                fileName = fileName.substring(
                    0,
                    fileName.length() - Reserved.EndOfInstructionMark.length()
                );

                // Is it a user module or a library ?
                try {
                    fileName = fileName.trim();
                    int posDot = fileName.indexOf( "." );
                    if ( posDot > -1 ) {
                        // A library
                        LibLoader.load( fileName ).collectAllMembers();
                    } else {
                        // A user module
                        String moduleName = fileName;
                        fileName += "." + Compiler.ModuleFileExt;
                        fileName = this.getInPath() + File.separator + fileName;
                        File fMod = new File( fileName );
                        
                        this.getCodeUnit().registerUserModuleImport(moduleName);

                        // Load it
                        if ( Ast.getAst().lookForModule( moduleName ) == null )
                        {
                            Parser.parseModule( fMod );
                        }
                    }
                }
                catch(FileNotFound exc)
                {
                    throw new FileNotFound( lex.getPos(), fileName );
                }
                catch(SemanticError exc)
                {
                    throw new SemanticError( lex.getPos(), exc.getMessage() );
                }
                catch(Exception exc)
                {
                    throw new SemanticError( lex.getPos(), exc.getMessage() );
                }
            } else {
                throw new EndOfInstruction( lex.getPos(),
                    I18n.get( I18n.Word.Parsing ) + " import" );
            }

        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                        + ' ' + Reserved.Import );
        }

        return;
    }

    /**
     * Parses and end of line
     * @throws CompilingError IF there is not a semicolon at the end.
     */
    protected void parseEOL() throws CompilingError {
        final Lexer lex = this.getLexer();

        lex.skipBlanks();
        if ( lex.peekNextChar() != Reserved.EndOfInstructionMark.charAt( 0 ) ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + " ';'" );
        }

        lex.read();                  // pass the ';'
        lex.skipBlanks();
    }

    /**
     * Parses a member: function or constant, the beginning is the same.
     * @throws CompilingError
     */
    protected void parseMember() throws CompilingError
    {
        Id id;
        Type type;
        String token;
        Visibility visible;
        boolean isFinal = false;

        token = lexer.peekNextToken();

        // Check visibility
        visible = Visibility.createVisibility( token );
        if ( visible != null ) {
            lexer.readToken();
            token = lexer.peekNextToken();
        } else {
            visible = PublicVisibility.get();
        }

        // Check modifier
        if ( token.equals( Reserved.Final ) ) {
            isFinal = true;
            lexer.readToken();
        }

        // Get type
        type = this.parseAnyType();

        // Get id
        id = new UsrId( lexer.readToken(), this.getLexer().getPos() );

        // Get next '=' or '(' ?
        lexer.skipBlanks();
        if ( lexer.peekNextChar() == '=' ) {
            if ( type == null ) {
                throw createSyntaxError( I18n.get( I18n.Word.Expected )
                        + ' ' + I18n.get( I18n.Word.Type )
                        + ", " + I18n.get( I18n.Word.Parsing )
                        + ' ' + I18n.get( I18n.Word.Constant ) );
            }
            else
            if ( !isFinal ) {
                if ( this.mode == Mode.Expert ) {
                    throw createSyntaxError( I18n.get( I18n.Word.Expected )
                            + ' ' + Reserved.Final + ", " + I18n.get( I18n.Word.Parsing )
                            + " '" + id.getId() + '\'' );
                }

                this.parseDef( isFinal, type, id );
                this.parseEOL();
            } else {
                this.parseConstant( visible, isFinal, type, id );
            }
        }
        else
        if ( lexer.peekNextChar() == '(' ) {
            if ( isFinal ) {
                throw createSyntaxError( I18n.get( I18n.Word.Prohibited )
                                            + ' ' + Reserved.Final
                                            + ", " + I18n.get( I18n.Word.Parsing )
                                            + ": " + I18n.get( I18n.Word.Function )
                                            + " '" + id.getId() + '\''
                );
            }

            this.parseFunction( visible, isFinal, type, id );
        }
        else {
            throw createSyntaxError(
                        I18n.get( I18n.Word.Expected )
                        + ' ' + I18n.get( I18n.Word.Variable )
                        + ", " + I18n.get( I18n.Word.Constant )
                        + ", " + I18n.get( I18n.Word.Or )
                        + ' ' + I18n.get( I18n.Word.Function )
                        + ", " + I18n.get( I18n.Word.Parsing )
                        + ", " + id.getId() + '\''
            );
        }

        return;
    }

    /**
     * Parses a variable definition, inside a function.
     * @return The id for the variable.
     * @throws CompilingError
     */
    protected Id parseDef() throws CompilingError
    {
        Id id = null;
        Type type;
        String token;
        boolean isFinal = false;
        final Lexer lex = this.getLexer();

        if ( this.getState() != State.InsideMember ) {
            throw createSyntaxError(
                    I18n.get( I18n.Word.Not )
                    + ' ' + I18n.get( I18n.Word.Expected )
                    + ' ' + I18n.get( I18n.Word.Variable )
                    + ", " + I18n.get( I18n.Word.Outside )
                    + ' ' + I18n.get( I18n.Word.Function )
            );
        }

        token = lex.peekNextToken();

        // Check modifier
        if ( token.equals( Reserved.Final ) ) {
            isFinal = true;
            lex.readToken();
        }

        // Get type
        type = this.parseType();
        if ( type == null ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ' ' + I18n.get( I18n.Word.Type ) );
        }

        // Get id
        id = new UsrId( lex.readToken(), lex.getPos() );

        return parseDef( isFinal, type, id );
    }

    /**
     * Second part of the process of parsing a definition.
     * @param isFinal Has a final as prefix?
     * @param type The type of the variable.
     * @param id The id for the variable name.
     * @return The id for the variable name.
     * @throws CompilingError
     * @throws SemanticError
     * @throws SyntaxError
     */
    protected Id parseDef(boolean isFinal, Type type, Id id) throws CompilingError, SemanticError, SyntaxError
    {
        final Lexer lex = this.getLexer();
        CodeBlock cb = this.getCodeBlock().getInstrs();
        Literal lit = null;
        RValue expr = null;

        // Check initialisation
        lex.skipBlanks();
        if ( lex.peekNextChar() == '=' ) {
            lex.read();

            if ( isFinal ) {
                lit = this.parseLiteral();

                if ( lit == null ) {
                    throw createSyntaxError( I18n.get( I18n.Word.Expected )
                                + ' ' + I18n.get( I18n.Word.Literal ) );
                }

                if ( !Type.areCompatible( lit.getType(), type ) ) {
                    throw createSyntaxError(
                            I18n.get( I18n.Word.Expected )
                            + ' ' + I18n.get( I18n.Word.Type )
                            + ' ' + I18n.get( I18n.Word.Of )
                            + ' ' + id.getId()
                            + ": " + type.toString()
                            + "; " + I18n.Word.Type
                            + ' ' + I18n.get( I18n.Word.Of )
                            + ' ' + I18n.get( I18n.Word.Literal )
                            + ", " + I18n.get( I18n.Word.Instead )
                            + ": " + lit.getType().toString() );
                }
            } else {

                if ( !this.isThereArrayCreation() ) {
                    expr = this.parseExpr();
                } else {
                    expr = this.parseArrayCreation();
                }

                if ( expr == null ) {
                    throw createSyntaxError( I18n.get( I18n.Word.Expected ) + " rvalue" );
                }
            }
        }
        // Create data
        if ( isFinal ) {
            if ( lit == null ) {
                throw createSyntaxError(
                            I18n.get( I18n.Word.Missing )
                            + ' ' + I18n.get( I18n.Word.Literal )
                            + ' ' + I18n.get( I18n.Word.For )
                            + ' ' + I18n.get( I18n.Word.Constant )
                            + ": '" + id.getId() + '\''
                );
            }

            // Create constant
            if ( !cb.add( new Fin( lex.getPos(), id, lit ) ) ) {
                throw createSyntaxError(
                         I18n.get( I18n.Word.Missing )
                         + ' ' +  I18n.get( I18n.Word.Constant )
                         + ": '" + id.getId() + '\''
                );
            }
        } else {
            // Create variable and assign expression
            if ( type == AnyType.get() ) {
                if ( expr == null ) {
                    throw createSyntaxError( I18n.get( I18n.Word.Missing )
                            + " rvalue "
                            + I18n.get( I18n.Word.For )
                            + ' ' + I18n.get( I18n.Sentence.TypeInference )
                    );
                }
            }

            if ( !cb.add( new Def( lex.getPos(), id, type ) ) ) {
                throw createSyntaxError( I18n.get( I18n.Word.Duplicated )
                            + ": " + I18n.get( I18n.Word.Variable )
                            + ' ' + id.getId()
                );
            }

            if ( expr != null ) {
                cb.add( new VarAsg( lex.getPos(), new RefId( id ), expr ) );
            } else {
                // Add a default value
                if ( type != AnyType.get() ) {
                    cb.add( new VarAsg( lex.getPos(), new RefId( id ), Type.createLiteral( type ) ) );
                }
            }
        }
        return id;
    }

    /**
     * Parsers a variable or constant definition, inside a member body
     */
    protected void parseDefStatement() throws CompilingError
    {
        this.parseDef();

        // End
        this.parseEOL();
        return;
    }

    /**
     * Parses return statements
     */
    protected void parseReturn() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        String token = lex.readToken();
        RefId id = null;

        // chk return
        if ( !token.equals( Reserved.Return ) ) {
            throw createSyntaxError(
                    I18n.get( I18n.Word.Expected ) + ' ' + Reserved.Return
            );
        }

        // parse expr
        id = parseExpr();

        // add ret
        this.getCodeBlock().getInstrs().add( new Ret( this.getLexer().getPos(), id ) );

        this.parseEOL();
        return;
    }

    /**
     * Parses if statements
     */
    private void parseIf() throws CompilingError {
        final Lexer lex = this.getLexer();
        String token = lex.readToken();
        RefId id = null;
        If instrIf = null;
        CodeBlockOwner oldCb = this.getCodeBlock();

        // chk return
        if ( !token.equals( Reserved.If ) ) {
            throw createSyntaxError(
                    I18n.get( I18n.Word.Expected ) + ' ' + Reserved.If );
        }

        // chk begin left bracket
        lex.skipBlanks();
        if ( lex.peekNextChar() != '(' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": '('" );
        }
        lex.read();

        // parse expr
        id = parseExpr();

        // chk begin right bracket
        lex.skipBlanks();
        if ( lex.peekNextChar() != ')' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": ')'" );
        }
        lex.read();

        // Create if
        instrIf = new If( lex.getPos(), id );
        oldCb.getInstrs().add( instrIf );

        // Check for the first brace
        lex.skipBlanks();
        if ( lex.peekNextChar() != '{' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": '{'" );
        }
        lex.read();

        // Read first block
        this.currentCodeBlock = instrIf;
        this.parseCodeBlock();

        // Is there an else?
        lex.skipBlanks();
        if ( lex.peekNextToken().equals( Reserved.Else ) ) {
            lex.readToken();

            // Check for the first brace
            lex.skipBlanks();
            if ( lex.peekNextChar() != '{' ) {
                throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": '{'" );
            }
            lex.read();

             // Read second block
            instrIf.setCurrentCodeBlockToSecond();
            this.parseCodeBlock();
        }

        this.currentCodeBlock = oldCb;
        return;
    }

    /**
     * Parses do.. while loopìng structures
     * @throws CompilingError when something goes wrong
     */
    private void parseDoWhile() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        CodeBlockOwner oldCb = this.getCodeBlock();
        RefId conditionId = null;
        Whl instrWhile = null;
        Whl codeBlockWhile = null;
        Whl conditionWhile = null;

        // chk return
        String token = lex.readToken();
        if ( !token.equals( Reserved.Do ) ) {
            throw createSyntaxError(
                    I18n.get( I18n.Word.Expected ) + ' ' + Reserved.Do );
        }

        // chk begin left brace
        lex.skipBlanks();
        if ( lex.peekNextChar() != '{' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": '{'" );
        }
        lex.read();

        // Parse the code block
        codeBlockWhile = new Whl( lex.getPos(), "codeBlockWhile" );
        codeBlockWhile.setOwner( oldCb );
        this.currentCodeBlock = codeBlockWhile;
        this.parseCodeBlock();

        // Parse while
        token = lex.readToken();
        if ( !token.equals( Reserved.While ) ) {
            throw createSyntaxError(
                    I18n.get( I18n.Word.Expected ) + ' ' + Reserved.While );
        }

        // chk begin left bracket
        lex.skipBlanks();
        if ( lex.peekNextChar() != '(' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": '('" );
        }
        lex.read();

        // Parse the condition code
        conditionWhile = new Whl( lex.getPos(), "conditionWhile" );
        conditionWhile.setOwner( oldCb );
        this.currentCodeBlock = conditionWhile;
        conditionId = this.parseExpr();

        // chk begin right bracket
        lex.skipBlanks();
        if ( lex.peekNextChar() != ')' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": ')'" );
        }
        lex.read();

        // Store the code block (first pass always happens)
        oldCb.getInstrs().addAll( codeBlockWhile.getInstrs() );
        oldCb.getInstrs().addAll( conditionWhile.getInstrs() );

        // Store the while
        instrWhile = new Whl( lex.getPos(), conditionId );
        oldCb.getInstrs().add( instrWhile );
        instrWhile.getInstrs().addAll( codeBlockWhile.getInstrs() );
        instrWhile.getInstrs().addAll( conditionWhile.getInstrs() );

        // end
        this.currentCodeBlock = oldCb;
        this.parseEOL();
        return;
    }

    /**
     * Parses the while loop structure
     */
    private void parseWhile() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        String token = lex.readToken();
        CodeBlockOwner oldCb = this.getCodeBlock();
        RefId id = null;
        Whl instrWhile = null;

        // chk return
        if ( !token.equals( Reserved.While ) ) {
            throw createSyntaxError(
                    I18n.get( I18n.Word.Expected ) + ' ' + Reserved.While );
        }

        // chk begin left bracket
        lex.skipBlanks();
        if ( lex.peekNextChar() != '(' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": '('" );
        }
        lex.read();

        // Change the codeblock to a temporal one.
        // Take into account that the condition must be re-evaluated each
        // iteration.
        Whl temp = new Whl( lex.getPos(), RefId.createTemp() );
        temp.setOwner( oldCb );
        this.currentCodeBlock = temp;

        // parse expr
        id = parseExpr();

        // Copy the condition code into the current codeblock
        oldCb.getInstrs().addAll( temp.getInstrs() );

        // chk begin right bracket
        lex.skipBlanks();
        if ( lex.peekNextChar() != ')' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": ')'" );
        }
        lex.read();

        // Create while
        instrWhile = new Whl( lex.getPos(), id );
        oldCb.getInstrs().add( instrWhile );

        // Check for the first brace
        lex.skipBlanks();
        if ( lex.peekNextChar() != '{' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": '{'" );
        }
        lex.read();

        // Read code block
        this.currentCodeBlock = instrWhile;
        this.parseCodeBlock();

        // Copy the condition code at the end of the whl codeblock
        instrWhile.getInstrs().addAll( temp.getInstrs() );

        this.currentCodeBlock = oldCb;
        return;
    }

    /**
     * Parses a for each, when the parsing of for() has already begun
     * This method takes control after ':' has been reached.
     * @param mainCtrlVar The var that will hold each element
     * @param tempCond The block where to put the finishing condition
     * @param tempCond The block where to put the init if the mainCtrlVar
     * @param incAsg The block where to put the increment of the for
     * @return A RefId to the id of the condition
     * @throws IncorrectId If the id's created were not valid.
     */
    protected RefId parseForEach(Id mainCtrlVar, Whl tempCond, Whl preface, Whl incAsg)
            throws CompilingError
    {
        final Lexer lex = this.getLexer();
        Position pos = lex.getPos();
        CodeBlockOwner oldCb = this.getCodeBlock();
        RefId toret = null;

        // Get id's for vector and ctrlVar
        RefId vectorId = this.parseId();
        Id ctrlVar = Id.createTemp();

        // Create & assign ctrlVar, i = 0
        oldCb.getInstrs().add( new Def( pos, ctrlVar, IntType.get() ) );
        oldCb.getInstrs().add( new VarAsg( pos, new RefId( ctrlVar ), new IntLiteral( 0 ) ) );

        // Assign main ctrl var, num = v[ i ]
        Id constZero = this.getCodeUnit().literalCache.get( 0 );
        Acc acc = new Acc( pos, vectorId, new RefId[]{ new RefId( constZero ) } );
        acc.setOwner( oldCb );
        oldCb.getInstrs().add( new VarAsg( pos, new RefId( mainCtrlVar ), acc ) );

        // Create condition
        // len( v )
        Cal calLength = new Cal( pos, new RefId( pos, "size" ), new RefId[] { vectorId } );
        calLength.setOwner( oldCb );

        // Def[int] tempCal
        Id tempCal = Id.createTemp();
        oldCb.getInstrs().add( new Def( pos, tempCal, IntType.get() ) );

        // tempCal = calLength
        tempCond.getInstrs().add( new VarAsg( pos, new RefId( tempCal ), calLength ) );

        // lt is 'i < tempCal'
        BinaryOp lt = new Lt( pos, new RefId( ctrlVar ), new RefId( tempCal ) );
        lt.setOwner( oldCb );

        // cond = lt
        toret = new RefId( Id.createTemp() );
        oldCb.getInstrs().add( new Def( pos, toret.toId(), BoolType.get() ) );
        tempCond.getInstrs().add( new VarAsg( pos, toret, lt ) );

        // Add condition code to function's code
        oldCb.getInstrs().addAll( tempCond.getInstrs() );

        // Create inc, ++i
        Id constOne = this.getCodeUnit().literalCache.get( 1 );
        BinaryOp addCtrlVar = new Add( pos, new RefId( ctrlVar ), new RefId( constOne ) );
        addCtrlVar.setOwner( oldCb );
        incAsg.getInstrs().add( new VarAsg( pos, new RefId( ctrlVar ), addCtrlVar ) );

        // Create preface, num = v[ i ]
        acc = new Acc( pos, vectorId, new RefId[]{ new RefId( ctrlVar ) } );
        acc.setOwner( preface );
        preface.getInstrs().add( new VarAsg( pos, new RefId( mainCtrlVar ), acc ) );

        return toret;
    }

    /**
     * Parses a regular for
     * This method takes place when the first ';' has been reached,
     * and it has been verified it is NOT a foreach
     * @param tempCond The codeblock in which put the condition checker code
     * @return A RefId to the id of the condition
     * @throws CompilingError When something goes wrong.
     */
    protected RefId parseRegFor(Whl tempCond, Whl incAsg) throws CompilingError
    {
        final Lexer lex = this.getLexer();
        CodeBlockOwner oldCb = this.getCodeBlock();
        RefId toret = null;

        if ( lex.peekNextChar() == ';' ) {
            lex.read();
        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ": ';' "
                    + I18n.get( I18n.Word.Parsing ) + ' ' + Reserved.For );

        }

        // Parse condition
        // Change the codeblock to a temporal one.
        // Take into account that the condition must be re-evaluated each
        // iteration.
        this.currentCodeBlock = tempCond;

        // parse expr
        toret = parseExpr();

        // Copy the condition code into the current codeblock
        oldCb.getInstrs().addAll( tempCond.getInstrs() );

        // chk ';'
        if ( lex.peekNextChar() == ';' ) {
            lex.read();
        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ": ';'"
                    + I18n.get( I18n.Word.Parsing ) + ' ' + Reserved.For
            );
        }

        // Parse inc
        lex.skipBlanks();
        if ( lex.peekNextChar() != ')' ) {
            incAsg.getInstrs().add( this.parseAsg() );
        }

        return toret;
    }

    /**
     * Parses a for construction
     * @throws CompilingError If anything is not parsed correctly
     */
    protected void parseFor() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        Position pos = lex.getPos();
        CodeBlockOwner oldCb = this.getCodeBlock();
        String token = "";
        Id mainCtrlVar = null;

        RefId conditionId = null;
        Whl instrWhile = null;
        Whl tempCond = new Whl( lex.getPos(), RefId.createTemp() );
        Whl incAsg = new Whl( lex.getPos(), RefId.createTemp() );
        Whl preface = new Whl( lex.getPos(), RefId.createTemp() );

        // Prepare
        tempCond.setOwner( oldCb );
        incAsg.setOwner( oldCb );
        preface.setOwner( oldCb );

        // chk for
        token = lex.readToken();
        lex.skipBlanks();
        if ( !token.equals( Reserved.For ) ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ' ' + Reserved.For );
        }

        // chk '('
        if ( lex.peekNextChar() == '(' ) {
            lex.read();
        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ": '(' "
                    + I18n.get( I18n.Word.Parsing )
                    + ' ' + Reserved.For
            );
        }

        // check init -- this is to be simply added (if it exists)
        lex.skipBlanks();
        if ( lex.peekNextChar() != ';' ) {
            if ( this.isBeginningDefAhead() ) {
                mainCtrlVar = this.parseDef();
            } else {
                oldCb.getInstrs().add( this.parseAsg() );
            }
        }

        // chk foreach ':' or regular for ';'
        if ( lex.peekNextChar() == ':' ) {
            lex.read();
            conditionId = this.parseForEach( mainCtrlVar, tempCond, preface, incAsg );
        } else {
            conditionId = this.parseRegFor( tempCond, incAsg );
        }

        // chk ')'
        lex.skipBlanks();
        if ( lex.peekNextChar() == ')' ) {
            lex.read();
        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ": ')' "
                    + I18n.get( I18n.Word.Parsing )
                    + ' ' + Reserved.For
            );
        }

         // chk '{'
        lex.skipBlanks();
        if ( lex.peekNextChar() == '{' ) {
            lex.read();
        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ": '{' "
                    + I18n.get( I18n.Word.Parsing )
                    + ' ' + Reserved.For
            );
        }

        // Parse body
        instrWhile = new Whl( pos, conditionId );
        oldCb.getInstrs().add( instrWhile );
        this.currentCodeBlock = instrWhile;
        instrWhile.getInstrs().addAll( preface.getInstrs() );
        this.parseCodeBlock();

        // Add inc & chk condition
        instrWhile.getInstrs().addAll( incAsg.getInstrs() );
        instrWhile.getInstrs().addAll( tempCond.getInstrs() );

        // Clean back & return
        this.currentCodeBlock = oldCb;
        return;
    }

    /**
     * Parses a switch structure, creating the succession of if's
     * @throws CompilingError
     */
    protected void parseSwitch() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        String token = "";
        RefId ctrlVar = null;
        CodeBlockOwner oldCb = this.getCodeBlock();
        ArrayList<Pair<ArrayList<Literal>, Whl>> cases = new ArrayList<Pair<ArrayList<Literal>, Whl>>();

        // chk for
        token = lex.readToken();
        lex.skipBlanks();
        if ( !token.equals( Reserved.Switch ) ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ' ' + Reserved.Switch
            );
        }

        // chk '('
        lex.skipBlanks();
        if ( lex.peekNextChar() == '(' ) {
            lex.read();
        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ": '(' "
                    + I18n.get( I18n.Word.Parsing )
                    + ' ' + Reserved.Switch
            );
        }

        // Parse expression for switch
        ctrlVar = this.parseExpr();

        // chk ')'
        lex.skipBlanks();
        if ( lex.peekNextChar() == ')' ) {
            lex.read();
        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ": ')' "
                    + I18n.get( I18n.Word.Parsing )
                    + ' ' + Reserved.Switch
            );
        }

        // chk '{'
        lex.skipBlanks();
        if ( lex.peekNextChar() == '{' ) {
            lex.read();
        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ": '{' "
                    + I18n.get( I18n.Word.Parsing )
                    + ' ' + I18n.get( I18n.Word.Block )
                    + ' ' + Reserved.Case
                    + ' ' + I18n.get( I18n.Word.Of )
                    + ' ' + Reserved.Switch
            );
        }

        this.parseCases( cases );
        this.currentCodeBlock = oldCb;
        this.generateSwitch( ctrlVar, cases );

        // chk '}'
        lex.skipBlanks();
        if ( lex.peekNextChar() == '}' ) {
            lex.read();
        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ": '}' "
                    + I18n.get( I18n.Word.Parsing )
                    + Reserved.Switch
            );
        }

        return;
    }

    /**
     * Generates the necessary code for a parsed switch
     * @param ctrlVar The refId to the ctrl var of the switch
     * @param cases The cases parsed before
     * @throws CompilingError
     */
    protected void generateSwitch(RefId ctrlVar, ArrayList<Pair<ArrayList<Literal>, Whl>> cases) throws CompilingError
    {
        final Lexer lex = this.getLexer();
        CodeBlockOwner cb = this.getCodeBlock();
        CodeBlockOwner oldCb = cb;
        CodeUnit cdu = this.getCodeUnit();
        RefId caseId = null;

        for(int numCase = 0; numCase < cases.size(); ++numCase) {
            // Equality vector (each literal with the ctrl var)
            Pair<ArrayList<Literal>, Whl> theCase = cases.get( numCase );
            RefId[] idsEq = new RefId[ theCase.first.size() ];

            if ( theCase.first.contains( null ) ) {
                if ( numCase != ( cases.size() -1 ) ) {
                    throw createSyntaxError( I18n.get( I18n.Sentence.DefaultLabelLastOne ) );
                }

                cb.getInstrs().addAll( theCase.second.getInstrs() );
            } else {
                // Prepare the equality of literals
                for(int i = 0; i < theCase.first.size(); ++i) {
                    Literal lit = theCase.first.get( i );

                    // Create the equality: ctrlVar == lit
                    Eq eq = new Eq( lex.getPos(), ctrlVar, new RefId( cdu.literalCache.get( lit ) ) );
                    eq.setOwner( cb );

                    // Create new temporal
                    idsEq[ i ] = new RefId( Id.createTemp() );
                    cb.getInstrs().add( new Def( lex.getPos(), idsEq[ i ].getId(), AnyType.get() ) );

                    // Add the operationg temp = ( ctrlVar == lit )
                    cb.getInstrs().add( new VarAsg( lex.getPos(), idsEq[ i ], eq ) );
                }

                // Prepare the unique comparison result for if's ctrl
                caseId = idsEq[ 0 ];
                if ( idsEq.length > 1 ) {
                    // Prepare the OR'ing needed
                    RefId firstOr = new RefId( Id.createTemp() );
                    cb.getInstrs().add( new Def( lex.getPos(), firstOr.getId(), AnyType.get() ) );
                    Or orInstr = new Or( lex.getPos(), idsEq[ 0 ], idsEq[ 1 ] );
                    orInstr.setOwner( cb );
                    cb.getInstrs().add( new VarAsg( lex.getPos(), firstOr, orInstr ) );
                    for(int i = 2; i < idsEq.length; ++i) {
                        RefId nextOr = new RefId( Id.createTemp() );
                        cb.getInstrs().add( new Def( lex.getPos(), nextOr.getId(), AnyType.get() ) );
                        Or orNextInstr = new Or( lex.getPos(), firstOr, idsEq[ i ] );
                        orNextInstr.setOwner( cb );
                        cb.getInstrs().add( new VarAsg( lex.getPos(), firstOr, orNextInstr ) );

                        firstOr = nextOr;
                    }

                    caseId = firstOr;
                }

                If instrIf = new If( lex.getPos(), caseId );
                instrIf.getFirstCodeBlock().addAll( theCase.second.getInstrs() );
                cb.getInstrs().add( instrIf );
                cb = instrIf;
                instrIf.setCurrentCodeBlockToSecond();
            }
        }

        // End
        this.currentCodeBlock = oldCb;
        return;
    }

    /**
     * Parses the cases labels of a switch
     * Fills a vector of pairs, { [ lit1, lit2, lit3 ], { codeBlock } }
     * Most common case is { [ lit ], { codeBlock } }
     * @throws CompilingError
     */
    protected void parseCases(ArrayList<Pair<ArrayList<Literal>, Whl>> cases) throws CompilingError
    {
        final Lexer lex = this.getLexer();
        String token = "";
        Literal lit = null;
        CodeBlockOwner oldCb = this.getCodeBlock();

        // Get to the case label
        lex.skipBlanks();
        token = lex.peekNextToken();
        while( !lex.isEOF()
            && ( token.equals( Reserved.Case )
              || token.equals( Reserved.Default ) ) )
        {
            boolean isDefault = token.equals( Reserved.Default );

            // Read the case label
            lex.readToken();                    // Pass 'case' or 'default'
            if ( !isDefault ) {
                lit = this.parseLiteral();
            } else {
                lit = null;
            }

            // Match :
            lex.skipBlanks();
            if ( lex.peekNextChar() == ':' ) {
                lex.read();
            } else {
                throw createSyntaxError( I18n.get( I18n.Word.Expected )
                        + ": ':' "
                        + I18n.get( I18n.Word.Parsing )
                        + ' ' + Reserved.Case
                );
            }

            // Create the temporal codeblock
            Whl temp = new Whl( lex.getPos(), RefId.createTemp() );
            temp.setOwner( oldCb );
            this.currentCodeBlock = temp;


            // Read statements in case entry
            lex.skipBlanks();
            while( !lex.isEOF()
                && !( lex.peekNextToken().equals( Reserved.Break ) )
                && !( lex.peekNextToken().equals( Reserved.Case ) ) )
            {
                this.parseStatement();
                lex.skipBlanks();
            }

            // Read break & next case label
            if ( lex.peekNextToken().equals( Reserved.Break ) ) {
                lex.readToken();
                this.parseEOL();

                // Though it may be empty, it has a break inside it.
                temp.getInstrs().add( new Nop( lex.getPos() ) );
            }

            // Store case
            int lastCasePos = cases.size() -1;
            if ( lastCasePos >= 0
              && cases.get( lastCasePos ).second.getInstrs().isEmpty() )
            {
                cases.get( lastCasePos ).first.add( lit );
                cases.get( lastCasePos ).second.getInstrs().addAll( temp.getInstrs() );
            } else {
                ArrayList<Literal> listOfLiteralsForCase = new ArrayList<Literal>();
                listOfLiteralsForCase.add( lit );
                cases.add( new Pair( listOfLiteralsForCase, temp ) );
            }

            lex.skipBlanks();
            token = lex.peekNextToken();
        }

        return;
    }

    /***
     * Parses terminals of expressions
     * @return The RefId which was assigned to the
     * @throws CompilingError
     */
    protected RefId parseExprTerm() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        RefId toret = null;
        CodeBlock cb = this.getCodeBlock().getInstrs();

        lex.skipBlanks();
        if ( !lex.isEOF()
          && lex.peekNextChar() != Reserved.EndOfInstructionMark.charAt( 0 ) )
        {
            // Parse first term
            if ( lex.isCallAhead() ) {
                toret = this.parseCalExpr();
            }
            else
            if ( lex.peekNextTokenType() == Lexer.TokenType.Id ) {
                // Is it a single refid ?
                toret = this.parseId();

                // .. or is it an access operator?
                lex.skipBlanks();
                if ( lex.peekNextChar() == '[' ) {
                    RefId[] position = this.parseArrayPosition();
                    Acc acc = new Acc( lex.getPos(), toret, position );
                    acc.setOwner( cb.getOwner() );

                    toret = new RefId( Id.createTemp() );
                    cb.add( new Def( lex.getPos(), toret.getId(), AnyType.get() ) );
                    cb.add( new VarAsg( lex.getPos(), toret, acc ) );
                }
            }
            else
            if ( lex.isLiteralAhead() ) {
                toret = new RefId(
                        this.getCodeUnit().literalCache.get( this.parseLiteral()
                ) );
            }
            else {
                throw createSyntaxError( I18n.get( I18n.Word.Expected )
                        + ' ' + I18n.get( I18n.Sentence.ExpressionTerm )
                        + ": id, fn., lit." );
            }
        }

        return toret;
    }

    /**
     * Parses a complete expression, in the form (subexpr) && (subexpr)
     * @return The tmp id in which the result of the expression will be stored
     */
    protected RefId parseExpr() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        CodeBlock cb = this.getCodeBlock().getInstrs();
        RefId toret = null;

        // Get the first subexpr
        lex.skipBlanks();
        char ch = lex.peekNextChar();

        if ( ch == Reserved.NotOperator.charAt( 0 ) ) {
            // !(subexpr)
            lex.read();
            RefId term = this.parseSubExpr();
            Not neg = new Not( lex.getPos(), term );
            neg.setOwner( this.getCodeBlock() );
            toret = new RefId( Id.createTemp() );
            cb.add( new Def( lex.getPos(), toret.getId(), AnyType.get() ) );
            cb.add( new VarAsg( lex.getPos(), toret, neg ) );
        } else {
            // Simple subexpr
            toret = this.parseSubExpr();
        }

        // Parse operator (if it exists)
        lex.skipBlanks();
        if ( lex.isBinaryOperatorAhead() ) {
            String op = lex.readOperator();

            // Read the operator
            if ( Lexer.isPreferencedConditionalOperator( op ) )
            {
                // Create operators and operands
                RefId op1 = toret;
                RefId op2 = this.parseExpr();
                BinaryOp opr = BinaryOp.createOp( lex.getPos(), op, op1, op2 );
                opr.setOwner( cb.getOwner() );

                // Store instrs
                toret = new RefId( Id.createTemp() );
                cb.add( new Def( lex.getPos(), toret.toId(), AnyType.get() ) );
                cb.add( new VarAsg( lex.getPos(), toret, opr ) );
            }
            else {
                throw new SemanticError( lex.getPos(),
                                          I18n.get( I18n.Sentence.BadParsing )
                );
            }
        }

        return toret;
    }

    /**
     * Parses an expression enclosed in brackets
     */
    protected RefId parseParExpr() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        CodeBlock cb = this.getCodeBlock().getInstrs();
        RefId toret = null;

        // Chk the first bracket
        lex.skipBlanks();
        if ( lex.peekNextChar() == '(' ) {
            lex.read();
        }

        toret = this.parseExpr();

        // Chk the last bracket
        lex.skipBlanks();
        if ( lex.peekNextChar() == ')' ) {
            lex.read();
        } else {
            throw new SyntaxError( lex.getPos(), I18n.get( I18n.Word.Expected )
                    + ": ')' "
            );
        }

        return toret;
    }

    /**
     * Parses a sub expression
     * A subexpression is the one found on one side of the && or ||
     * preference operators
     * @return The id in which
     */
    protected RefId parseSubExpr() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        CodeBlock cb = this.getCodeBlock().getInstrs();
        RefId toret = null;

        // Is it prefix or infix?
        lex.skipBlanks();
        char ch = lex.peekNextChar();

        // Chk there is an operator
        if ( Reserved.OperatorChars.indexOf( ch ) > -1 ) {
            toret = this.parsePrefixSubExpr();
        } else {
            toret = this.parseRegSubExpr();
        }

        return toret;
    }

     /**
     * Parses a prefix sub expression: -2
     * A subexpression is the one found on one side of the && or ||
     * There must have been found an operator to reach here
     * preference operators
     * @return The id in which
     */
    protected RefId parsePrefixSubExpr() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        CodeBlock cb = this.getCodeBlock().getInstrs();
        RefId toret = null;

        // Pass prefix operator
        lex.skipBlanks();
        char ch = lex.read();
        lex.skipBlanks();

        if ( ch == Reserved.AddOperator.charAt( 0 ) )
        {
            toret = parseRegSubExpr();
        }
        else
        if ( ch == Reserved.SubOperator.charAt( 0 ) )
        {
            RefId term = null;

            // if it is a single symbol or a parenthesized expression
            if ( lex.peekNextChar() == '(' ) {
                term = this.parseParExpr();
            } else {
                term = this.parseExprTerm();
            }

            // build "tmp <- ( 0 - term )"
            toret = new RefId( Id.createTemp() );
            cb.add( new Def( lex.getPos(), toret.toId(), AnyType.get() ) );
            Sub sub = new Sub( lex.getPos(),
                        new RefId( this.getCodeUnit().literalCache.get( 0 ) ),
                        term
            );
            sub.setOwner( this.getCodeBlock() );
            cb.add( new VarAsg( lex.getPos(), toret, sub ) );

            // Start recursion
            toret = this.parseRegSubExprSecOp( toret );

        }
        else {
            throw createSyntaxError(
                I18n.get( I18n.Sentence.IsNotPrefixOperator )
                + ": '" + ch + '\''
            );
        }

        return toret;
    }

    /**
     * Parses a regular sub expression: 2 * 5
     * A subexpression is the one found on one side of the && or ||
     * preference operators
     * @return The id in which
     */
    protected RefId parseRegSubExpr() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        RefId toret = null;

        // Parse first term
        lex.skipBlanks();
        if ( lex.peekNextChar() == '(' ) {
            toret = this.parseParExpr();
        } else {
            toret = this.parseExprTerm();
        }

        toret = parseRegSubExprSecOp( toret );

        return toret;
    }

    /**
     * Parses a regular sub expression: 2 * 5
     * A subexpression is the one found on one side of the && or ||
     * preference operators
     * This is the second part of the function parseRegSubExpr()
     * @return The id in which
     */
    protected RefId parseRegSubExprSecOp(RefId op1) throws CompilingError
    {
        final Lexer lex = this.getLexer();
        CodeBlock cb = this.getCodeBlock().getInstrs();
        RefId toret = op1;

        // Parse operator (if it exists)
        lex.skipBlanks();
        if ( lex.isBinaryOperatorAhead() ) {
            // Read the operator
            if ( !Lexer.isPreferencedConditionalOperator( lex.peekNextOperator() ) )
            {
                String op = lex.readOperator();

                // Create operators and operands
                op1 = toret;
                RefId op2 = this.parseSubExpr();
                BinaryOp opr = BinaryOp.createOp( lex.getPos(), op, op1, op2 );
                opr.setOwner( cb.getOwner() );

                // Store instrs
                toret = new RefId( Id.createTemp() );
                cb.add( new Def( lex.getPos(), toret.toId(), AnyType.get() ) );
                cb.add( new VarAsg( lex.getPos(), toret, opr ) );
            }
        }

        return toret;
    }

    /**
     * Parses an assignment statement
     */
    protected void parseAsgStatement() throws CompilingError
    {
        // Parse asg
        this.getCodeBlock().getInstrs().add( this.parseAsg() );

        // End
        this.parseEOL();
        return;
    }

    /**
     * Parses a regular assignment, of the form x = 1 +...
     * @return The VarAsg Instr built
     * @throws CompilingError when somehting goes wrong
     */
    protected Asg parseRegAsg() throws CompilingError
    {
        RefId id = null;
        Acc acc = null;
        RValue expr = null;
        final Lexer lex = this.getLexer();
        Asg toret = null;

        // Get id
        id = new RefId( lex.getPos(), lex.readToken() );

        // Chk whether it is a collection access
        lex.skipBlanks();
        if ( lex.peekNextChar() == '[' ) {
            // It is a collection's access assign, parse assign
            RefId[] position = this.parseArrayPosition();

            acc = new Acc( lex.getPos(), id, position );
            acc.setOwner( this.getCodeBlock() );
        }

        // Chk initialisation
        lex.skipBlanks();
        if ( lex.peekNextChar() == '=' ) {
            lex.read();

            if ( !this.isThereArrayCreation() ) {
                expr = this.parseExpr();
            } else {
                expr = this.parseArrayCreation();
            }

            if ( expr == null ) {
                throw createSyntaxError( I18n.get( I18n.Word.Expected )
                        + " rvalue"
                );
            }

            if ( acc == null ) {
                toret = new VarAsg( lex.getPos(), id, expr );
            } else {
                toret = new AccAsg( lex.getPos(), acc, expr );
            }
        } else {
            if ( lex.isAdvancedOperatorAhead() ) {
                toret = this.parseAdvPostfixAsg( id );
            } else {
                throw createSyntaxError( I18n.get( I18n.Word.Expected )
                        + ' ' + I18n.get( I18n.Word.Assign )
                );
            }
        }

        return toret;
    }

    /**
     * Parses an adv assign, such as x++, or x += 5
     * @param id The RefId to the var being assigned to.
     * @return The VarAsg instr object.
     */
    protected VarAsg parseAdvPostfixAsg(RefId id) throws CompilingError
    {
        final Lexer lex = this.getLexer();
        String op = "";
        RefId exprId = null;

        // Read operator
        op = lex.readOperator();

        // Decide what to do on operator (args?)
        if ( !op.equals( Reserved.AsgDecByOneOperator )
          && !op.equals( Reserved.AsgIncByOneOperator ) )
        {
            // Create the rvalue for +=, -=,...
            exprId = this.parseExpr();

            if ( exprId == null ) {
                throw createSyntaxError( I18n.get( I18n.Word.Expected )
                        + " rvalue"
                );
            }
        } else {
            // Create the rvalue for -- and ++
            Id idConstOne = this.getCodeUnit().literalCache.get( 1 );
            exprId = new RefId( idConstOne );
        }

        return this.createAdvAsg( id, op, exprId );
    }

    /**
     * Parses an assignment
     */
    protected Asg parseAsg() throws CompilingError
    {
        Asg toret = null;

        if ( this.getState() != State.InsideMember ) {
            throw createSyntaxError( I18n.get( I18n.Sentence.VariableDefOutBody ) );
        }

        if ( this.getLexer().isAdvancedOperatorAhead() ) {
            toret = this.parseAdvPrefixAsg();
        } else {
            toret = this.parseRegAsg();
        }

        return toret;
    }

    /**
     * Parses a RefId, i.e., a reference to an existing id
     * @return The referenced id
     * @see RefId
     */
    protected RefId parseId() throws CompilingError
    {
        String refIdPart;
        final Lexer lex = this.getLexer();
        RefId toret = null;

        // Read id part (probably for the prefix, but it can also be the name)
        refIdPart = lex.readToken();
        lex.skipBlanks();

        // Read complete id, if present
        if ( lex.peekNextChar() == '.' ) {
            // There is a complete identifier 'refIdPart.name'
            lex.read();
            toret = new RefId( lex.getPos(), refIdPart, lex.readToken() );

            lex.skipBlanks();

        } else {
            // There was a single id 'fn(' (the refIdPart was the name)
            toret = new RefId( null, refIdPart );
        }

        return toret;
    }

    protected void parseCalStatement() throws CompilingError
    {
        this.getCodeBlock().getInstrs().add( this.parseCal() );
        this.parseEOL();
    }

    protected Cal parseCal() throws CompilingError
    {
        RefId toret = null;
        RefId fnId = null;
        final Lexer lex = this.getLexer();
        ArrayList<RefId> args = new ArrayList<RefId>();
        Cal cal = null;

        // Get over the RefId part of the fn. call
        fnId = parseId();

        if ( lex.peekNextChar() != '(' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ": '('"
                    + ' ' + I18n.get( I18n.Word.Parsing )
                    + ' ' + I18n.get( I18n.Word.Call )
                    + " fn."
            );
        }

        // Read arguments
        lex.read();                  // pass '('
        lex.skipBlanks();

        while( !lex.isEOF()
            && lex.peekNextChar() != ')' )
        {
            args.add( this.parseExpr() );
            lex.skipBlanks();

            // Pass the comma
            char ch = lex.peekNextChar();
            if ( ch == ',' ) {
                lex.read();
            }
            else
            if ( ch != ')' ) {
                throw createSyntaxError( I18n.get( I18n.Word.Expected )
                        + ": ',' ')' "
                        + I18n.get( I18n.Word.Parsing )
                        + ' ' + I18n.get( I18n.Word.Arguments )
                );
            }
        }

        lex.read();                  // pass the ')'
        lex.skipBlanks();

        // Create cal code
        cal = new Cal( lex.getPos(), fnId, args.toArray( new RefId[ args.size() ] ) );
        cal.setOwner( currentCodeBlock );

        return cal;
    }

    /**
     * Parses a function call inside a expression.
     * Returns the id assigned as the return value.
     */
    protected RefId parseCalExpr() throws CompilingError
    {
        Cal cal = this.parseCal();
        CodeBlock cb = this.getCodeBlock().getInstrs();
        Position pos = this.getLexer().getPos();

        // Create vars
        RefId toret = new RefId( Id.createTemp() );
        VarAsg asg = new VarAsg( pos, toret, cal );

        // Add to the codeblock
        cb.add( new Def( pos, AnyType.get(), toret.toId() ) );
        cb.add( asg );

        return toret;
    }

    protected void parseStatement() throws CompilingError
    {
        final Lexer lex = this.getLexer();

        // Parse function statements
        lex.skipBlanks();

        if ( lex.peekNextTokenType() == Lexer.TokenType.Id ) {
            String token = lex.peekNextToken();

            // It is a local def statement
            if ( this.isBeginningDefAhead() ) {
                this.parseDefStatement();
            }
            else
            // It is a return statement
            if ( token.equals( Reserved.Return ) ) {
                this.parseReturn();
            }
            else
            // It is an if statement
            if ( token.equals( Reserved.If ) ) {
                this.parseIf();
            }
            else
            // It is a while statement
            if ( token.equals( Reserved.While ) ) {
                this.parseWhile();
            }
            else
            // It is a do... while statement
            if ( token.equals( Reserved.Do ) ) {
                this.parseDoWhile();
            }
            else
            // It is a for statement
            if ( token.equals( Reserved.For ) ) {
                this.parseFor();
            }
            else
            // It is a switch statement
            if ( token.equals( Reserved.Switch ) ) {
                this.parseSwitch();
            }
            else
            // It is a break statement
            if ( token.equals( Reserved.Break ) ) {
                this.currentCodeBlock.getInstrs().add( new Brk( lex.getPos() ) );
                lex.readToken();
                this.parseEOL();
            }
            else
            // It is a continue statement
            if ( token.equals( Reserved.Continue ) ) {
                this.currentCodeBlock.getInstrs().add( new Cnt( lex.getPos() ) );
                lex.readToken();
                this.parseEOL();
            }
            else
            // It is an Asg (starts with local vble)
            if ( this.isAsgAhead( token ) ) {
                this.parseAsgStatement();
            }
            // It is a function call
            else {
                this.parseCalStatement();
            }
        } else {
             // It is an VarAsg (starts with local vble)
            if ( this.isAsgAhead( "" ) ) {
                this.parseAsgStatement();
            } else {
                throw createSyntaxError( I18n.get( I18n.Sentence.IllegalStartStatement ) );
            }
        }

        lex.skipBlanks();
        return;
    }

    /**
     * Parses a sequence of instructions
     * @throws CompilingError
     */
    protected void parseStatements() throws CompilingError
    {
        final Lexer lex = this.getLexer();

        // Parse function statements
        lex.skipBlanks();
        while( lex.peekNextChar() != '}'
            && !lex.isEOF() )
        {
            this.parseStatement();
        }
    }

    /**
     * Determines whether there is an VarAsg ahead
     */
    public boolean isAsgAhead(String token) throws CompilingError
    {
        return ( this.getLexer().isAdvancedOperatorAhead()
              || this.getCodeBlock().lookForData( token ) != null )
        ;
    }

    /**
     * Parses a statement beginning with advanced operators
     * @throws CompilingError
     */
    public VarAsg parseAdvPrefixAsg() throws CompilingError
    {
        // Read the operator
        String operator = this.getLexer().readOperator();

        if ( !operator.equals( Reserved.AsgIncByOneOperator )
          && !operator.equals( Reserved.AsgDecByOneOperator ) )
        {
            throw createSyntaxError( I18n.get( I18n.Sentence.IllegalStartStatement )
                    + ". "
                    + I18n.get( I18n.Word.Allowed )
                    + ": '++', '--'"
            );
        }

        // Read the identifier
        RefId var = this.parseId();

        // Create the rvalue
        Id idConstOne = this.getCodeUnit().literalCache.get( 1 );
        return this.createAdvAsg( var, operator, new RefId( idConstOne ) );
    }

    public VarAsg createAdvAsg(RefId var, String operator, RefId exprId) throws IncorrectId
    {
        final Lexer lex = this.getLexer();
        BinaryOp op = null;
        CodeBlockOwner cb = this.getCodeBlock();

        // Determine the operator to use
        if ( operator.equals( Reserved.AsgIncByOneOperator )
          || operator.equals( Reserved.AsgIncByXOperator ) )
        {
            op = new Add( lex.getPos(), var, exprId );
            op.setOwner( cb );
        }
        else
        if ( operator.equals( Reserved.AsgDecByOneOperator )
          || operator.equals( Reserved.AsgDecByXOperator ) )
        {
            op = new Sub( lex.getPos(), var, exprId );
            op.setOwner( cb );
        }
        else
        if ( operator.equals( Reserved.AsgMulByXOperator ) ) {
            op = new Mul( lex.getPos(), var, exprId );
            op.setOwner( cb );
        }
        else
        if ( operator.equals( Reserved.AsgDivByXOperator ) ) {
            op = new Div( lex.getPos(), var, exprId );
            op.setOwner( cb );
        }

        // Create & end
        return new VarAsg( lex.getPos(), var, op );
    }

    /**
     * Parses a codeblock inside a function body
     */
    public void parseCodeBlock() throws CompilingError
    {
        final Lexer lex = this.getLexer();

        // Parse the instructions inside
        this.parseStatements();

        // Reading } or marking error
        if (  !lex.isEOF()
           && lex.peekNextChar() == '}' )
        {
            lex.read();
            lex.skipBlanks();
        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Unexpected )
                    + " eof "
                    + I18n.get( I18n.Word.Parsing )
                    + ": '}'"
            );
        }

        return;
    }

    /**
     * Parses the interior of a function
     */
    public void parseFunctionBody() throws CompilingError
    {
        final Lexer lex = this.getLexer();

        if ( this.getState() != State.InsideMember ) {
            throw createSyntaxError( I18n.get( I18n.Sentence.CodeOutBody ) );
        }

        // Parse the instructions inside
        this.parseStatements();

        // Reading } or marking error
        if (  !lex.isEOF()
           && lex.peekNextChar() == '}' )
        {
            lex.read();
            lex.skipBlanks();
            this.status = State.InFunctions;
            this.currentCodeBlock = this.getCodeUnit();
        } else {
            throw createSyntaxError( I18n.get( I18n.Word.Unexpected )
                    + " eof, "
                    + I18n.get( I18n.Word.Expected )
                    + "; '}'"
            );
        }

        return;
    }

    protected boolean isBeginningDefAhead() throws CompilingError
    {
        String token = lexer.peekNextToken();

        return ( Reserved.Final.equals( token )
          || Reserved.Private.equals( token )
          || Reserved.Public.equals( token )
          || Reserved.isType( token ) );
    }

    public Type parseAnyType() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        Type toret = null;
        String strType = lexer.readToken();

        lex.skipBlanks();
        while( lex.peekNextChar() == '[' ) {
            lex.read();
            lex.skipBlanks();

            if ( lex.read() != ']' ) {
                throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": ']'" );
            }

            strType += "[]";
            lex.skipBlanks();
        }

        toret = Type.getType( Reserved.cnvtTypeToNative( strType ) );

        if ( toret == null
          && !( strType.equals( Reserved.Void ) ) )
        {
            throw createSyntaxError(
                    I18n.get( I18n.Word.Expected )
                    + ' ' + I18n.get( I18n.Word.Type )
                    + ", " + I18n.get( I18n.Word.Not )
                    + ": '" + strType + '\''
            );
        }

        return toret;
    }

    public Type parseType() throws CompilingError
    {
        Type toret = this.parseAnyType();

        if ( toret == null ) {
            throw createSyntaxError(
                    I18n.get( I18n.Word.Expected )
                    + ' ' + I18n.get( I18n.Sentence.NonVoidType )
                    + ", " + I18n.get( I18n.Word.Not )
                    + ": '" + lexer.getLastToken() + '\''
            );
        }

        return toret;
    }

    public void parseConstant(Visibility visible, boolean isFinal, Type type, Id id)
            throws CompilingError
    {
        Literal literal;
        Fin cnst = null;

        if ( this.getState() == State.InImports ) {
            this.status = State.InConstants;
        }
        
        if ( this.getState() != State.InConstants ) {
            throw createSyntaxError( I18n.get( I18n.Sentence.MisplacedConstant ) );
        }

        try {
            // Check '='
            lexer.skipBlanks();
            if ( lexer.read() != '=' ) {
                throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": '='" );
            }

            literal = parseLiteral();

            if ( !Type.areCompatible( type, literal.getType() ) ) {
                throw createSyntaxError( I18n.get( I18n.Word.Type )
                        + ' ' + I18n.get( I18n.Word.Declared )
                        + " (" + type.toString() + ") "
                        + I18n.get( I18n.Sentence.DoesNotMatchWith )
                        + " literal (" + literal.getType() + ")"
                );
            }

        }
        catch(SyntaxError e) {
            throw e;
        }
        catch(Exception e) {
            throw createSyntaxError( e.getMessage() );
        }

         // Prepare Ast info
        cnst = new Fin( lexer.getPos(), id, literal );
        cnst.setDoc( this.getDoc() );
        this.resetDoc();

        if ( !this.getCodeUnit().getInstrs().add( cnst ) ) {
            throw createSyntaxError( I18n.get( I18n.Word.Duplicated )
                        + ' ' + I18n.get( I18n.Word.Constant )
                        + ": "
                                        + Reserved.cnvtNativeToType( type )
                                        + ' '
                                        + id.getId()
            );
        }

        this.parseEOL();
        return;
    }

    /**
     * Checks whether there is an array creation expression "new int[5]" next
     * @return true if there it is, false otherwise
     */
    protected boolean isThereArrayCreation() throws CompilingError
    {
        this.getLexer().skipBlanks();

        return ( this.getLexer().peekNextToken().equals( Reserved.NewOperator ) );
    }

    /**
     * This parses a expression like "new int[5]", used when
     * creating an array.
     * @return A Cra object or ArrayLiteral
     *         holding the information of the array creation
     * @throws CompilingError
     */
    protected RValue parseArrayCreation() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        Type type = null;
        RValue toret = null;

        // Chk 'new'
        lex.skipBlanks();
        if ( !( lex.readToken().equals( Reserved.NewOperator ) ) ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ": '" + Reserved.NewOperator + '\''
            );
        }

        // Read type
        Pair<Type, RefId[]> arrayTypeInfo = this.parseNewArrayType();

        // Read {, whether it exists
        lex.skipBlanks();
        if ( lex.peekNextChar() == '{' ) {

            // Chk
            for(RefId index: arrayTypeInfo.second) {
                if ( index != null ) {
                    throw createSyntaxError(
                            I18n.get( I18n.Sentence.SizeReferencesNull )
                    );
                }
            }

            // Read until end of array
            StringBuilder strArray = new StringBuilder();
            strArray.append( '{' );
            int braceLevel = 1;
            lex.read();

            char ch = lex.read();
            while( braceLevel >= 1
                && !lex.isEOF() )
            {
                strArray.append( ch );

                if ( ch == '{' ) {
                    ++braceLevel;
                    ch = lex.read();
                }
                else
                if ( ch == '}' ) {
                    --braceLevel;
                }
                else ch = lex.read();
            }


            // Prepare type
            String lit = strArray.toString();
            ArrayType arrayType = (ArrayType) arrayTypeInfo.first;
            Type componentType = arrayType.getType();

            if ( arrayType.isString() ) {
                arrayType = StringType.get();
            }

            // Parse & store
            lit = ArrayLiteral.removeExternalBraces( lit );
            ArrayLiteral.StringArrayLiteral literal = new ArrayLiteral.StringArrayLiteral( lit );
            Object array = ArrayLiteral.parseArray( componentType.getJavaType(), literal );

            toret = new ArrayLiteral( arrayType, array );

        } else {
            // Chk
            for(RefId index: arrayTypeInfo.second) {
                if ( index == null ) {
                    throw createSyntaxError( I18n.get( I18n.Sentence.SizeMustBeSpecified ) );
                }
            }

            // Create CRA
            toret = new Cra( lex.getPos(), arrayTypeInfo.first, arrayTypeInfo.second );
            ( (Cra) toret ).setOwner( this.getCodeBlock() );
        }

        return toret;
    }

    /**
     * Parses an array Type with the limits information for each dimension,
     * if present.
     * @return The array type, and a vector holding the limit for each dimension,
     *          in a Pair
     * @throws CompilingError
     */
    public Pair<Type, RefId[]> parseNewArrayType() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        String strType;
        Pair<Type, RefId[]> toret = new Pair<Type, RefId[]>( null, null );

        // Read type
        lex.skipBlanks();
        strType = lex.readToken();

        // Read array limits
        toret.second = this.parseArrayLimits();

        // Fix type, adding [] for each limit
        for(RefId id: toret.second) {
            strType += "[]";
        }

        // Convert type
        toret.first = Type.getType( Reserved.cnvtTypeToNative( strType ) );

        // Chk
        if ( !( toret.first instanceof ArrayType ) ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ' ' + I18n.get( I18n.Word.Type )
                    + ' ' + I18n.get( I18n.Word.Array )
            );
        }

        if ( toret.first == null ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected )
                    + ' ' + I18n.get( I18n.Word.Type )
                    + ", " + I18n.get( I18n.Word.Not )
                    + ": '" + strType + '\''
            );
        }

        // End
        return toret;
    }

    /**
     * Parses an array position access, the [0] in v[0]
     * @return The list of coordinates to the position, as a vector of RefId
     * @throws CompilingError
     */
    protected RefId[] parseArrayPosition() throws CompilingError
    {
        RefId[] toret = this.parseArrayParams();

        for(RefId ref: toret) {
            if ( ref == null ) {
                throw createSyntaxError( I18n.get( I18n.Word.Expected )
                        + ' ' + I18n.get( I18n.Word.Arguments )
                        + ' ' + I18n.get( I18n.Word.For )
                        + ' ' + I18n.get( I18n.Sentence.ArrayAccess )
                );
            }
        }

        return toret;
    }

    /**
     * Parses an array limit definition, the [10] in new int[10]
     * @return The list of limits of the array as a vector of RefId
     * @throws CompilingError
     */
    protected RefId[] parseArrayLimits() throws CompilingError
    {
        return this.parseArrayParams();
    }

    /**
     * This method is able to parse array limits or array positions
     * @return The set of limits/position, as a primitive vector of RefId.
     */
    protected RefId[] parseArrayParams() throws CompilingError
    {
        final Lexer lex = this.getLexer();
        ArrayList<RefId> lims = new ArrayList<RefId>();

        lex.skipBlanks();
        while( lex.peekNextChar() == '[' ) {
            lex.read();
            lex.skipBlanks();

            // Chk there is a limit
            if ( lex.peekNextChar() != ']' ) {
                // Add each limit
                lims.add( this.parseExpr() );
            } else {
                lims.add( null );
            }

            lex.skipBlanks();
            if ( lex.read() != ']' ) {
                throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": ']'" );
            }

            lex.skipBlanks();
        }

        // End
        return lims.toArray( new RefId[ lims.size() ] );
    }

    protected Literal parseLiteral() throws CompilingError
    {
        Literal toret = null;
        final Lexer lex = this.getLexer();
        Lexer.TokenType litType;

        lex.skipBlanks();
        if ( lex.isLiteralAhead() ) {
            litType = lex.peekNextTokenType();

            if ( litType == Lexer.TokenType.CharLiteral ) {
                toret = new CharLiteral( lex.readCharLiteral() );
            }
            else
            if ( litType == Lexer.TokenType.StringLiteral ) {
                toret = new StringLiteral( lex.readStringLiteral() );
            }
            else
            if ( litType == Lexer.TokenType.NumberLiteral ) {
                toret = NumericalLiteral.create( lex.readNumberLiteral() );
            }
            else
            if ( litType == Lexer.TokenType.BoolLiteral ) {
                toret = BoolLiteral.create( lex.readToken() );
            }
        }

        if( toret == null ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + " literal" );
        }

        lex.skipBlanks();
        return toret;
    }

    protected void parseDoc() throws CompilingError
    {
        this.doc = Documentation.parse( this.getLexer().skipMultipleLineComment()  );
    }

    protected boolean isDocCommentAhead()
    {
        return ( this.getLexer().isStringAhead( MultiLineDocCommentMark ) );
    }

    protected boolean isCommentAhead()
    {
        return ( this.getLexer().isStringAhead( OneLineCommentMark )
              || this.getLexer().isStringAhead( MultiLineCommentMark ) );
    }

    public static void skipComment(Lexer lexer) throws CompilingError
    {
        if ( lexer.isStringAhead( OneLineCommentMark ) ) {
            lexer.skipSingleLineComment();
        }
        else
        if ( lexer.isStringAhead( MultiLineCommentMark ) ) {
            lexer.skipMultipleLineComment();
        }

        return;
    }

    public void parseFunction(Visibility visible, boolean isFinal, Type type, Id id) throws CompilingError
    {
        Fun f = null;
        Data[] params = null;

        setState( State.InFunctions );

        // Check '('
        lexer.skipBlanks();
        if ( lexer.read() != '(' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": '('" );
        }

        // Parameters
        lexer.skipBlanks();
        if ( lexer.peekNextChar() != ')' ) {
            params = parseFormalParameters();
        }

        // Check ')'
        lexer.skipBlanks();
        if ( lexer.read() != ')' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": ')'" );
        }

        // Skip '{'
        lexer.skipBlanks();
        if ( lexer.read() != '{' ) {
            throw createSyntaxError( I18n.get( I18n.Word.Expected ) + ": '{'" );
        }

        // Prepare Ast info
        f = new Fun( this.lexer.getPos(), this.getCodeUnit(), visible, type, id, params );
        f.setDoc( this.getDoc() );
        this.resetDoc();
        if ( !this.getCodeUnit().getInstrs().add( f ) ) {
            throw createSyntaxError(  I18n.get( I18n.Word.Duplicated )
                    + " fn.: " + Reserved.cnvtNativeToType( type )
                    + ' ' + id.getId()
            );
        }

        // Done
        this.currentCodeBlock = f;
        setState( State.InsideMember );

        return;
    }

    protected Data[] parseFormalParameters() throws CompilingError
    {
        Type type;
        Id id;
        boolean firstTime = true;
        ArrayList<Def> args = new ArrayList<Def>();

        try {
            do {
                if ( !firstTime
                  && lexer.peekNextChar() == ',' )
                {
                    lexer.read();
                }

                // Parse parameters
                type = parseType();

                // Get id
                id = new Id( lexer.readToken(), this.getLexer().getPos() );

                // Store
                args.add( new Def( this.getLexer().getPos(), type, id ) );
                firstTime = false;
            } while ( lexer.peekNextChar() == ',' );
        } catch(CompilingError e)
        {
            throw createSyntaxError( I18n.get( I18n.Word.Parsing )
                    + ' ' + I18n.get( I18n.Word.Arguments )
                    + ": " + e.getMessage()
            );
        }

        return args.toArray( new Data[ args.size() ] );
    }

    private SyntaxError createSyntaxError(String message)
    {
        return new SyntaxError( this.lexer.getPos(), message );
    }

    /**
     * Get compilation mode; this change from parse to parse
     * @see Mode The compilation mode (advanced, beginner...)
     * @return The mode selected for compilation.
     */
    public Mode getMode()
    {
        return this.mode;
    }

    /**
     * Returns the documentation generated for this module.
     * @return The documentation, as complete as possible.
     */
    private Documentation getDoc()
    {
        return this.doc;
    }

    /**
     * Rests the documentation, in order to attend to another member.
     */
    public void resetDoc()
    {
        this.doc = new Documentation();
    }



    protected Lexer lexer = null;
    private State status;
    private CodeBlockOwner currentCodeBlock = null;
    private CodeUnit targetCodeUnit = null;
    private Mode mode;
    private Documentation doc = null;
    private boolean wasUnitDocStored = false;
}
