// semantic checker

package core;

import core.Parser.Mode;
import core.exceptions.SemanticError;
import core.intermediatecode.Module;
import core.intermediatecode.Type;
import core.intermediatecode.instrs.Fun;
import core.lexical.Position;
import core.lexical.Reserved;

/**
 * Triggers the semantic verification process
 * @author baltasarq
 */
public class SemanticChecker
{
    /**
     * Prepares the semantic checker for a given Ast and position.
     * The language is assumed to have been previously set.
     * @param ast The Ast, as an Ast object.
     * @param pos The position in the source file (usually at the end)
     */
    public SemanticChecker(Ast ast, Position pos, Mode mode)
    {
        this.ast = ast;
        this.pos = pos;
        this.mode = mode;
    }
    
    /**
     * Check semantics
     */
    public void chk() throws CompilingError
    {
        final Ast astTree = this.getAst();
        
        // Initial checks
        for(Module mod: this.getAst().getModules()) {
            mod.chk();
        }
        astTree.getPrg().chk();
        
        // Specific chks
        if ( this.getMode() == Parser.Mode.Expert ) {
            Fun[] fnsMain = astTree.getPrg().lookForFn( Reserved.EntryPointFn, new Type[]{} );
            if ( fnsMain.length != 1 )
            {
                throw new SemanticError( this.getPos(), I18n.get( I18n.Word.Expected )
                        + ' ' + I18n.get( I18n.Word.Function )
                        + " '" + Reserved.EntryPointFn + "()'"
                        + ' ' + I18n.get( I18n.Word.Parsing )
                        + " '" + astTree.getPrg().getName() + '\''
                      );
            }
        }
    }
    
    /**
     * Get the mode for verification (strictness)
     * @return The strictness mode, as a Mode value.
     */
    public Mode getMode()
    {
        return this.mode;
    }
    
    /**
     * Returns the position in the source file when the semantic
     * verification was triggered.
     * @return The position, as a Position object.
     */
    public Position getPos()
    {
        return this.pos;
    }
    
    /**
     * The Ast this verifier is working on.
     * @return The Ast, as an Ast object.
     */
    public Ast getAst()
    {
        return this.ast;
    }
    
    private Ast ast;
    private Position pos;
    private Mode mode;
}
