// CodeBlock.java

package core.intermediatecode;

import core.AstMemberFinder;
import core.intermediatecode.instrs.Def;
import core.intermediatecode.instrs.Fin;
import core.intermediatecode.instrs.Fun;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * Represents a various Instr
 * @author baltasarq
 */
public class CodeBlock implements Collection<Instr> {
/**
     * Creates a new code unit
     */
    public CodeBlock(CodeBlockOwner owner)
    {
        this.instrs = new ArrayList<Instr>();
        this.owner = owner;
    }

    /**
     * Returns the owner of this codeblock: expected a function, or a code unit.
     * @return
     */
    public CodeBlockOwner getOwner()
    {
        return this.owner;
    }
    
    /**
     * Changes the owner of the code block
     * This is normally needed when the owner is not known at construction
     * @param owner The owner of the code block
     */
    public void setOwner(CodeBlockOwner owner)
    {
        this.owner = owner;
    }

    /**
     * Returns the number of instructions in the code unit.
     * @return The number of instrs, as int.
     */
    @Override
    public int size()
    {
        return this.instrs.size();
    }

    /**
     * Returns an instr, given its position.
     * @param x The position of the instr.
     * @return The instr, as an Instr object.
     * @see Instr
     */
    public Instr get(int x)
    {
        return this.instrs.get( x );
    }
    
    /**
     * Appends a new instr to the end of the code unit.
     * @param x The instr to add.
     */
    @Override
    public boolean add(Instr x)
    {
        return this.add( this.size(), x );
    }
    
    /**
     * Appends a collection of instructions
     * @param c The collection of instructions
     * @return true if all of them were added, false otherwise
     */
    @Override
    public boolean addAll(Collection c)
    {
        boolean toret = true;
        
        for(Object instr: c) {
            if ( !( this.instrs.add( (Instr) instr ) ) ) {
                toret = false;
                break;
            }
        }
        
        return toret;
    }

    /**
     * Inserts a new instr at position pos of the code unit.
     * @param x The instr to add.
     * @param pos The position in which to add x
     */
    public boolean add(int pos, Instr x)
    {
        boolean toret = true;
        CodeBlockOwner where = this.getOwner();
        
        // Rectify where, if needed
        if ( x instanceof Data ) {
            where = CodeBlock.getRootOwner( where );

            if ( where == null ) {
                where = this.getOwner();
            }
        }
        
        // Check whether there is already something like this
        if ( x instanceof Fun ) {
            Fun fun = (Fun) x;
            CodeUnit cduOwner = CodeBlock.getModuleOwner( owner );
            Fun[] duplicated = cduOwner.lookForFn( fun.getId().getId(), fun.getParamTypes() );
            duplicated = AstMemberFinder.filterFunctionsByExactParams( duplicated, fun.getParamTypes() );
             
            if ( duplicated.length > 0 ) {
                toret = false;
            }
        }
        else
        if ( x instanceof Data ) {
            if ( owner.lookForData( ( (Data) x ).getId().getId() ) != null ) {
                toret = false;
            }
        }

        if ( toret ) {
            // Add it to the codeblock
            if ( x instanceof Fin
              || x instanceof Def )
            {
                // If its a DEF or FIN, put it at the beginning
                pos = 0;
            }
                
            // Mandatory to differentiate so there is no infinite recursion
            if ( where.getInstrs() == this ) {
                this.instrs.add( pos, x );
            } else {
                where.getInstrs().add( pos, x );
            }
            
            // Set the owner
            x.setOwner( where );
        }

        return toret;
    }

    /**
     * Inserts a new Instr at a given position.
     * @param i The index position
     * @param x The new instruction.
     */
    public void insert(int i, Instr x)
    {
        this.instrs.add( i, x );
    }

    /**
     * Returns the textual version of the code unit.
     * @return The whole code unit, as a String
     */
    @Override
    public String toString()
    {
        StringBuilder toret = new StringBuilder();

        toret.append( "{\n" );
        for(Instr x: this.instrs) {
            toret.append( x.toString() );
            toret.append( '\n' );
        }
        toret.append( "}\n" );

        return toret.toString();
    }

    public Fin lookForConst(String id)
    {
        Fin toret = null;

        lookInInstrs:
        for(int i = 0; i < this.size(); ++i) {
            Instr instr = this.get( i );

            if ( instr instanceof Fin )
            {
                Fin fin = (Fin) instr;

                if ( fin.getId().toString().equals( id ) ) {
                    toret = fin;
                    break lookInInstrs;
                }
            }
        }

        return toret;
    }

    public Def lookForVar(String id)
    {
        Def toret = null;

        lookInInstrs:
        for(int i = 0; i < this.size(); ++i) {
            Instr instr = this.get( i );

            if ( instr instanceof Def )
            {
                Def def = (Def) instr;

                if ( def.getId().toString().equals( id ) ) {
                    toret = def;
                    break lookInInstrs;
                }
            }
        }

        return toret;
    }

    public Data lookForData(String id)
    {
        Data toret = this.lookForConst( id );

        if ( toret == null ) {
            toret = this.lookForVar( id );
        }

        return toret;
    }
    
    /**
     * Reach the owner up to the module
     * @param owner A given CodeBlockOwner can be an Instr (If, Whl), a Fun
     * or a Module object
     * @return The module this owner pertains to, up in the ownership hierarchy
     *          also returns null if parameter owner is null.
     * @see Instr, If, Whl, Fun, Module
     */
    public static CodeUnit getModuleOwner(CodeBlockOwner owner)
    {
        CodeUnit toret = null;
        
        if ( owner instanceof CodeUnit ) {
                toret = (CodeUnit) owner;
        } else {
            while( toret == null
                && owner != null )
            {
                owner = owner.getOwner();
                
                if ( owner instanceof CodeUnit ) {
                    toret = (CodeUnit) owner;
                    break;
                }
                
            }
        }
        
        return toret;
    }
    
    /**
     * Reach the owner up to the function
     * @param owner A given CodeBlockOwner can be an Instr (if, whl), a Fun
     * or a Module object
     * @return The function this owner pertains to as a Fun object,
     * up in the ownership hierarchy; null if the owner is a module or null
     * @see Instr, If, Whl, Fun, Module
     */
    public static Fun getFunctionOwner(CodeBlockOwner owner)
    {
        Fun toret = null;
        
        if ( owner instanceof Fun ) {
            toret = (Fun) owner;
        }
        else {
            while( toret == null
                && owner != null )
            {
                owner = owner.getOwner();
                
                if ( owner instanceof Fun ) {
                    toret = (Fun) owner;
                    break;
                }
            }
        }

        return toret;
    }
    
    /**
     * Returns the top owner for this block
     * @param owner The codeblock to start with
     * @return The module or function that owns this owner
     */
    public static CodeBlockOwner getRootOwner(CodeBlockOwner owner)
    {
        CodeBlockOwner toret = CodeBlock.getFunctionOwner( owner );
        
        if ( toret == null ) {
            toret = CodeBlock.getModuleOwner( owner );
        }
        
        return toret;
    }

    @Override
    public boolean isEmpty() {
        return ( this.instrs.isEmpty() );
    }

    @Override
    public boolean contains(Object o) {
        return this.instrs.contains( o );
    }

    @Override
    public Iterator iterator() {
        return this.instrs.iterator();
    }

    @Override
    public Object[] toArray() {
        return this.instrs.toArray();
    }

    @Override
    public Object[] toArray(Object[] a) {
        return this.instrs.toArray( a );
    }

    public boolean remove(Instr o) {
        return this.instrs.remove( o );
    }

    @Override
    public boolean remove(Object o) {
        return this.instrs.remove( o );
    }

    @Override
    public boolean containsAll(Collection c) {
        return this.instrs.containsAll( c );
    }

    @Override
    public boolean removeAll(Collection c) {
        return this.instrs.removeAll( c );
    }

    @Override
    public boolean retainAll(Collection c) {
        return this.instrs.retainAll( c );
    }

    @Override
    public void clear() {
        this.instrs.clear();
    }

    private final ArrayList<Instr> instrs;
    private CodeBlockOwner owner;
}
