// CodeUnit.java

package core.intermediatecode;

import core.Ast;
import core.CompilingError;
import core.exceptions.SemanticError;
import core.intermediatecode.instrs.Fin;
import core.intermediatecode.instrs.Fun;
import core.lexical.Id;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * A class containing source code.
 * @author baltasarq
 */
public abstract class CodeUnit implements CodeBlockOwner {

    public static final String DocPrefix = "doc_";
    public static final String HtmlExtension = ".html";
    public static final boolean DoNotIncludePath = false;
    public static final boolean IncludePath = true;

    /**
     * Creates a new code unit, associated to a source file
     */
    public CodeUnit(File f)
    {
        this.instrs = new CodeBlock( this );
        this.userModulesImports = new ArrayList<String>();
        this.source = f;
        this.name = getFileNameWithoutExtension( f );
        this.literalCache = new LiteralCache( this );
        this.doc = new Documentation();
    }

    /**
     * Returns the appropriate prefix for the documentation file.
     * @return The prefix, as a String.
     */
    public abstract String getDocFilePrefix();

    /**
     * Returns the appropriate name for the documentation file
     * corresponding to this module.
     * @param includePath determine whether to include the path or not
     * @return The name of the documentation file, as a String.
     */
    public String buildDocFileName(boolean includePath)
    {
        return buildDocFileName( this, includePath );
    }


    /**
     * Returns the appropriate name for the documentation file corresponding to a given CodeUnit
     * @param unit the CodeUnit to create the doc file name for.
     * @param includePath determine whether to include the path or not
     * @return the name of the documentation file, as a String.
     */
    public static String buildDocFileName(CodeUnit unit, boolean includePath)
    {
        return buildDocFileName( unit.getInFile(), unit.getDocFilePrefix(), includePath );
    }

    /**
     * Returns the appropriate name for the documentation file corresponding to a given file.
     * Automatically determines it is a module or a program, and passes the prefix
     * Module.ModulePrefix or Prg.ProgramPrefix to another version of this method.
     * @param f The file to build the doc file name for
     * @param includePath whether to include the complete path or not
     * @return the name of the documentation file, as a String.
     */
    public static String buildDocFileName(File f, boolean includePath)
    {
        String prefix = Prg.ProgramPrefix;

        if ( f.toString().endsWith( Module.ModuleFileExt  ) ) {
            prefix = Module.ModulePrefix;
        }

        return buildDocFileName( f, prefix, includePath );
    }

    /**
     * Returns the appropriate name for the documentation file corresponding to a given file.
     * Instead of automatically determining whether it is a module or a program, just pass the prefix
     * Module.ModulePrefix or Prg.ProgramPrefix
     * @param f The file to build the doc file name for
     * @param prefix  the prefix to append
     * @param includePath whether to include the complete path or not
     * @return the name of the documentation file, as a String.
     */
    public static String buildDocFileName(File f, String prefix, boolean includePath)
    {
        final String path = f.getParent();
        final String fileName = getFileNameWithoutExtension( f );
        String toret = "";

        if ( includePath ) {
            toret = path + File.separator;
        }

        toret += DocPrefix + prefix;
        return ( toret + fileName + HtmlExtension );
    }
    
    public static String getFileNameWithoutExtension(File f)
    {
        String toret = f.getName();
        int dotPos = toret.lastIndexOf( '.' );
        
        if ( dotPos > -1 ) {
            toret = toret.substring( 0, dotPos );
        }
        
        return toret;
    }

    /**
     * Returns the name of the code unit
     * @return The code unit's name, as a String
     */
    public String getName()
    {
        return this.name;
    }

    /**
     * Returns the whole name of the code unit
     * @return The whole code unit's name, as a String
     */
    public String getCompleteName()
    {
        return this.name;
    }

    /**
     * Returns the codeblock inside this code unit.
     * @return The instrs set, as a CodeBlock class.
     */
    @Override
    public CodeBlock getInstrs()
    {
        return this.instrs;
    }

    /**
     * Returns all functions inside this code unit, matching the given name.
     * @return A vector of functions, as a Fun primitive vector, null if not found.
     */
    public Fun[] getFunctionByName(String name)
    {
        ArrayList<Fun> vfns = new ArrayList<Fun>();
        Fun[] toret = null;

        // Lookup for functions
        for(int i = 0; i < this.getInstrs().size(); ++i) {
            Instr instr = this.getInstrs().get( i );

            if ( instr instanceof Fun) {
                Fun fun = (Fun) instr;
                
                if ( fun.getId().getId().equals( name ) ) {
                    vfns.add( fun );
                }
            }
        }
        
        // Prepare the return value, if needed
        if ( vfns.size() > 0 ) {
            toret = vfns.toArray( new Fun[ vfns.size() ] );
        }

        return toret;
    }
    
    /**
     * Returns a number of constants in this module matching the name
     * and number of arguments given.
     * @param name The name of the function to match
     * @param numArgs The number of arguments to match
     * @return A vector of constants, as a Fun primitive vector, null if not found.
     */
    public Fun[] getFunctionByName(String name, int numArgs)
    {
        ArrayList<Fun> vfns = new ArrayList<Fun>();
        Fun[] fns = this.getFunctionByName( name );
        Fun[] toret = null;
        
        // Remove functions not marching args number
        if ( fns != null ) {
            for(Fun fn: fns) {
                if ( fn.getParams().length == numArgs ) {
                    vfns.add( fn );
                }
            }
        }
        
        // Prepare the return value, if needed
        if ( vfns.size() > 0 ) {
            toret = vfns.toArray( new Fun[ vfns.size() ] );
        }
        
        return toret;
    }

    /**
     * Returns the list of functions in the code unit.
     * @return The list of functions, as a vector.
     */
    public Fun[] getFunctions()
    {
        ArrayList<Fun> functions = new ArrayList<Fun>();

        for(int i = 0; i < this.getInstrs().size(); ++i) {
            Instr instr = this.getInstrs().get( i );

            if ( instr instanceof Fun) {
                functions.add( (Fun) instr);
            }
        }

        return functions.toArray( new Fun[ functions.size() ] );
    }
    
    /**
     * Returns the list of constants  in the code unit.
     * @return The list of constants, as a vector.
     */
    public Fin[] getConstants()
    {
        ArrayList<Fin> constants = new ArrayList<Fin>();

        for(int i = 0; i < this.getInstrs().size(); ++i) {
            Instr instr = this.getInstrs().get( i );

            if ( instr instanceof Fin) {
                constants.add( (Fin) instr);
            }
        }

        return constants.toArray( new Fin[ constants.size() ] );
    }

    /**
     * Returns the list of all members  in the code unit.
     * @return The list of members, as a vector.
     */
    public Member[] getMembers()
    {
        ArrayList<Member> members = new ArrayList<>();

        members.addAll( Arrays.asList( this.getConstants() ) );
        members.addAll( Arrays.asList( this.getFunctions() ) );

        return members.toArray( new Member[ members.size() ] );
    }

    /**
     * Returns the textual version of the code unit.
     * @return The whole code unit, as a String
     */
    @Override
    public String toString()
    {
        String toret = "# " + this.getName() + '\n';

        toret += this.getInstrs().toString();

        return toret;
    }

    @Override
    public Data lookForData(String id) {
        return this.getInstrs().lookForData( id );
    }

    @Override
    public Data lookForVar(String id) {
        return this.getInstrs().lookForVar( id );
    }

    @Override
    public Fin lookForConst(String id) {
        return this.getInstrs().lookForConst( id );
    }
    
    /**
     * Checks all instructions in the code unit for sanity
     * @throws SemanticError When an error is found
     */
    public void chk() throws CompilingError
    {
        // Check all intrs
        for(Instr instr: this.getInstrs()) {
            instr.chk();
        }

        return;
    }
    
    /**
     * Locates duplicated for a given fun that is probably about to be inserted
     * @param id The function id that is considered to be inserted.
     * @param paramTypes The types for the parameters of the function.
     * @return The number of constants with exact number & type of params
     */
    @Override
    public Fun[] lookForFn(String id, Type[] paramTypes)
    {
        ArrayList<Fun> toret = new ArrayList<Fun>();
        int reqParams = paramTypes.length;
        
        // Retrieve coincident constants
        Fun[] fns = this.getFunctionByName( id, reqParams );
        
        if ( fns != null ) {
            // Filter
            for (Fun fn : fns) {
                // For each fun
                int n = 0;

                for (; n < reqParams; ++n) {
                    Type duplicatedParamType = fn.getParams()[n].getType();
                    if ( !Type.areCompatible( duplicatedParamType, paramTypes[n] ) )
                    {
                        break;
                    }
                }

                if ( n == reqParams ) {
                    toret.add( fn );
                }
            }
        }
        
        return toret.toArray( new Fun[ toret.size() ] );
    }
    
    /**
     * Returns the HTML representation of this CodeUnit
     * @return The HTML representation, as a String.
     */
    public String toHtml()
    {
        StringBuilder toret = new StringBuilder();
        String[] imports = this.getUserModulesImports();
        Fin[] constants = this.getConstants();
        Fun[] fns = this.getFunctions();
        String brief = this.getDoc().getBrief();
        String author = this.getDoc().getAuthor();
        String date = this.getDoc().getDate();

        // HTML preface
        toret.append( "<!DOCTYPE html>\n" );
        toret.append( "<html>\n" );
        toret.append( "<head>\n<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n"
                    + "<title>" + this.getName() + "</title></head>\n" );
        toret.append( "<body>\n" );

        // Prologue
        toret.append( "<h1 style=\"background-color: rgb(204, 204, 204);\">" );
        toret.append( this.getHtmlHeader() );
        toret.append( "</h1>\n<p>\n" );
        toret.append( "<table width='100%' style=\"background-color: rgb(204, 204, 204);\">\n" );

        if ( !brief.isEmpty() ) {
            toret.append( "<tr>\n<td><big><big><b><i>i</i></b></big></big></td><td><i>" );
            toret.append( brief );
            toret.append( "</i></td></tr>\n" );
        }

        if ( !author.isEmpty() ) {
            toret.append( "<tr>\n<td><big><big><b>&copy;</b></big></big></td><td><i>" );
            toret.append( author );
            toret.append( "</i></td></tr>\n" );
        }

        if ( !date.isEmpty() ) {
            toret.append( "<tr><td><big><big><b>&#964;</b></big></big></td><td><i>" );
            toret.append( date );
            toret.append( "</i></td></tr>\n" );
        }

        toret.append( "</table>\n<p>\n" );

        // Imports
        if ( imports.length > 0 ) {
            toret.append( "<h2 style=\"background-color: rgb(204, 204, 204);\"><i>&#8743;</i></h2>\n" );
            toret.append( "<ul>\n" );
            for(String imp : imports ) {
                final CodeUnit cdu = Ast.getAst().lookForModule( imp );

                if (  cdu != null
                   && !( cdu instanceof Lib ) )
                {
                    toret.append( "<li><a href=\"" );
                    toret.append( cdu.buildDocFileName(DoNotIncludePath) );
                    toret.append( "\">" );
                    toret.append( imp );
                    toret.append( "</a>\n" );
                }
            }
            toret.append( "</ul>\n" );
        }

        // Constants
        if ( constants.length > 0 ) {
            ArrayList<Fin> fins = new ArrayList<>();

            // Filter constants for there are constants created by the compiler.
            for(Fin constant : constants ) {
                if ( !constant.getId().getId().startsWith( Id.TmpId ) ) {
                    fins.add( constant );
                }
            }

            if ( fins.size() > 0 ) {
                toret.append( "\n<h2 style=\"background-color: rgb(204, 204, 204);\">&#945;</h2>" );
                toret.append( "<ul>\n" );

                for(Fin constant : fins ) {
                        toret.append( "\n<li style='padding: 2px 5px 2px 5px;'>\n" );
                        toret.append( constant.toHtml() );
                        toret.append( "</li>\n" );
                }
                toret.append( "</ul>\n" );
            }
        }

        // Functions
        if ( fns.length > 0 ) {
            toret.append( "\n<h2 style=\"background-color: rgb(204, 204, 204);\">&#402;</h2>" );
            toret.append( "\n<ul>\n" );
            for(Fun fn : fns ) {
                if ( fn.getVisibility().isVisible() ) {
                    toret.append( "<li style='padding: 2px 5px 2px 5px;'>\n" );
                    toret.append( fn.toHtml() );
                    toret.append( "</li>\n" );
                }
            }
            toret.append( "</ul>\n" );
        }

        // Footer
        toret.append( "\n<p align=right style=\"background-color: rgb(204, 204, 204);\"><small><i>jC</i></small></p>" );

        // HTML colophon
        toret.append( "\n\n</body>\n</html>\n" );
        
        return toret.toString();
    }
    
    /**
     * Get the HTML header for the container
     * @return The header, as HTML representation in a String
     */
    public String getHtmlHeader()
    {
        return this.getName();
    }
    
    /**
     * Returns the documentation associated to this code unit.
     * @return The documentation, as a Documentation object.
     */
    public Documentation getDoc()
    {
        return this.doc;
    }

    /**
     * Completely changes the current documentation for the unit (program, module, lib)
     * @param doc
     */
    public void setDoc(Documentation doc)
    {
        this.doc = doc;
    }
    
    /**
     * Returns the userModulesImports that have been parsed for this CodeUnit.
     * @return The userModulesImports, as a primitive vector of String.
     */
    public String[] getUserModulesImports()
    {
        return this.userModulesImports.toArray( new String[ this.userModulesImports.size() ] );
    }
    
    public void registerUserModuleImport(String importName)
    {
        this.userModulesImports.add(importName);
    }

    public File getInFile()
    {
        return this.source;
    }

    private final CodeBlock instrs;
    private File source;
    private String name;
    private Documentation doc;
    private ArrayList<String> userModulesImports;
    public LiteralCache literalCache;
}
