// Documentation.java

package core.intermediatecode;

import core.Parser;
import core.lexical.Reserved;

import java.util.ArrayList;
import java.util.HashMap;

/**
 * Documentation stored for any member
 * @author baltasarq
 */
public class Documentation {
    public static String MultiLineCommentMark = Reserved.MultiLineCommentMark;
    public static String MultiLineDocCommentMark = Reserved.MultiLineDocCommentMark;
    public static String EndMultiLineCommentMark = Reserved.EndMultiLineCommentMark;
    
    public enum DocLabel {
        name, brief, author, date, ret, param, hide;

        public static final String EtqName    = "@name";
        public static final String EtqBrief   = "@brief";
        public static final String EtqAuthor  = "@author";
        public static final String EtqDate    = "@date";
        public static final String EtqRet     = "@return";
        public static final String EtqParam   = "@param";
        public static final String EtqHide    = "@hide";
        
        public static final String[] DocLabels = {
            EtqName, EtqBrief, EtqAuthor, EtqDate, EtqRet, EtqParam, EtqHide
        };

        @Override
        public String toString()
        {
            return DocLabels[ this.ordinal() ];
        }
    }

    public Documentation()
    {
        this.doc = new HashMap<String, String>();
        this.params = new HashMap<String, String>();

        this.doc.put( DocLabel.EtqBrief, "" );
        this.doc.put( DocLabel.EtqAuthor, "" );
        this.doc.put( DocLabel.EtqDate, "" );
        this.doc.put( DocLabel.EtqRet, "" );
        this.doc.put( DocLabel.EtqName, "" );
        this.doc.put( DocLabel.EtqHide, "0" );

        this.lastArgKeySet = "";
    }

    /**
     * Returns the last arg that was set in the doc dictionary.
     * @return The last arg, as String.
     */
    public String getLastArgKeySet()
    {
        return this.lastArgKeySet;
    }

    /**
     * @return the name
     */
    public String getName()
    {
        return this.doc.get( DocLabel.EtqName );
    }

    /**
     * @param name the name to set
     */
    public void setName(String name)
    {
        this.doc.put( DocLabel.EtqName, name.trim() );
    }

    /**
     * @return is it hidden or not
     */
    public boolean isHidden()
    {
        return ( this.doc.get( DocLabel.EtqHide ).charAt( 0 ) == '1' );
    }

    /**
     * @param hidden whether it is hidden or not
     */
    public void setHidden(boolean hidden)
    {
        String value = "0";

        if ( hidden ) {
            value = "1";
        }

        this.doc.put( DocLabel.EtqHide, value );
    }

    /**
     * @return the brief
     */
    public String getBrief()
    {
        return this.doc.get( DocLabel.EtqBrief );
    }

    /*
     * @param brief the brief to set
     */
    public void setBrief(String brief)
    {
        this.doc.put( DocLabel.EtqBrief, brief.trim() );
    }

    /**
     * @return the date
     */
    public String getDate()
    {
        return this.doc.get( DocLabel.EtqDate );
    }
    /**
     * @param date the date to set
     */
    public void setDate(String date)
    {
        this.doc.put( DocLabel.EtqDate, date.trim() );
    }

    /**
     * @return the author
     */
    public String getAuthor()
    {
        return this.doc.get( DocLabel.EtqAuthor );
    }

    /**
     * @param author the author to set
     */
    public void setAuthor(String author)
    {
        this.doc.put( DocLabel.EtqAuthor, author.trim() );
    }

    /**
     * @return the ret
     */
    public String getReturnDoc()
    {
        return this.doc.get( DocLabel.EtqRet );
    }

    /**
     * @param returnDoc the documentation for return to set
     */
    public void setReturnDoc(String returnDoc)
    {
        this.doc.put( DocLabel.EtqRet, returnDoc.trim() );
    }

    /**
     * Get all keys
     * @return A vector of all keys of the params documentation.
     */
    public String[] getDocKeys()
    {
        return this.doc.keySet().toArray( new String[ this.doc.size() ] );
    }

    /**
     * Get all keys
     * @return A vector of all keys of the params documentation.
     */
    public String[] getParamDocKeys()
    {
        return this.params.keySet().toArray( new String[ this.params.size() ] );
    }

    /**
     * Get the documentation for a given parameter.
     * @param key The key of the param to retrieve
     * @return The associated documentation, as String-
     */
    public String getParamDoc(String key)
    {
        return this.params.get( key.trim() );
    }

    /**
     * Add a new parameter documentation.
     * @param paramName The name of the parameter.
     * @param paramDoc The documentation associated to this parameter.
     */
    public void setParamDoc(String paramName, String paramDoc)
    {
        this.lastArgKeySet = paramName.trim();
        this.params.put( this.getLastArgKeySet(), paramDoc.trim() );
    }
    
    /**
     * Retrieves a documentation item by its DocLabel label.
     * @param lbl The label of the documentation item to retrieve
     * @return null if the label is not matched, the value otherwise
     */
    public String get(DocLabel lbl)
    {
        return this.doc.get( lbl.toString() );
    }

    /**
     *
     * @param lbl The doc item to set. If this is arg, the lastArgSet is used.
     * @param value A String value to set the argument.
     */
    public void set(DocLabel lbl, String value)
    {
        if ( lbl != DocLabel.hide ) {
            if ( lbl == DocLabel.param ) {
                this.setParamDoc( this.getLastArgKeySet(), value );
            } else {
                this.doc.put( lbl.toString(), value );
            }
        }

        return;
    }

    public static void convertHtmlSpecialCharacters(ArrayList<String> v)
    {
        for(int i = 0; i < v.size(); ++i) {
            String line = v.get( i );

            // Replace spanish vowels
            line = line.replace( "á", "&aacute;" );
            line = line.replace( "é", "&eacute;" );
            line = line.replace( "í", "&iacute;" );
            line = line.replace( "ó", "&oacute;" );
            line = line.replace( "ú", "&uacute;" );

            // Replace spanish ñ
            line = line.replace( "ñ", "&ntilde;" );

            // Replace '<' and '>'
            line = line.replace( "<", "&lt;" );
            line = line.replace( ">", "&gt;" );

            // Store changes
            v.set( i, line );
        }

        return;
    }

    /**
     * Takes a documentation text as a single line, and lexes it.
     * @param text The documentation, as as single String line.
     * @return A vector of Strings.
     */
    public static String[] splitDocInLines(String text)
    {
        String[] doc = text.trim().split( "\n" );
        ArrayList<String> toret = new ArrayList<>();

        // Clean lines of comment marks
        for(String docLine : doc) {
            docLine = docLine.trim();

            // Eliminate comment mark beginning
            if ( docLine.startsWith( MultiLineDocCommentMark ) ) {
                docLine = sub(docLine, 0, MultiLineDocCommentMark.length() ).trim();
            }

            if ( docLine.startsWith( MultiLineCommentMark ) ) {
                docLine = sub( docLine, 0, MultiLineCommentMark.length() ).trim();
            }

            // Eliminate beginning asterisks
            if ( !docLine.isEmpty()
                    && docLine.charAt( 0 ) == '*' )
            {
                docLine = sub( docLine, 1, docLine.length() ).trim();
            }

            // Eliminate ending comment marks
            if ( docLine.endsWith( EndMultiLineCommentMark ) ) {
                docLine = sub( docLine,
                        0,
                        docLine.length()
                                - EndMultiLineCommentMark.length() )
                        .trim();
            }

            // Eliminate ending asterisks
            if ( !docLine.isEmpty()
                    && docLine.charAt( docLine.length() -1 ) == '*' )
            {
                docLine = sub( docLine, 0, docLine.length() -1 ).trim();
            }

            // Is there any info remaining?
            if ( !docLine.isEmpty() ) {
                toret.add( docLine );
            }
        }

        convertHtmlSpecialCharacters( toret );
        return toret.toArray( new String[ toret.size() ] );
    }

    /**
     * Creates a Documentation object from the documentation marks in comment
     * @param text The string with the comment with document marks inside it.
     * @return A new Documentation document.
     */
    public static Documentation parse(String text)
    {
        // Init
        DocLabel status = DocLabel.brief;
        Documentation toret = new Documentation();
        String[] docLines = splitDocInLines( text );

        // Parse contents
        for(String docLine : docLines) {
            // Parse name
            if ( docLine.toLowerCase().startsWith( DocLabel.EtqName ) ) {
                status = DocLabel.name;
                toret.setName( sub( docLine, DocLabel.EtqName.length() ) );
            }
            else
            // Parse brief
            if ( docLine.toLowerCase().startsWith( DocLabel.EtqBrief ) ) {
                status = DocLabel.brief;
                toret.setBrief( sub( docLine, DocLabel.EtqBrief.length() ) );
            }
            else
            // Parse author
            if ( docLine.toLowerCase().startsWith( DocLabel.EtqAuthor ) ) {
                status = DocLabel.author;
                toret.setAuthor( sub( docLine, DocLabel.EtqAuthor.length() ) );
            }
            else
            // Parse date
            if ( docLine.toLowerCase().startsWith( DocLabel.EtqDate ) ) {
                status = DocLabel.date;
                toret.setDate( sub( docLine, DocLabel.EtqDate.length() ) );
            }
            else
            // Parse return info
            if ( docLine.toLowerCase().startsWith( DocLabel.EtqRet ) ) {
                status = DocLabel.ret;
                toret.setReturnDoc( sub( docLine, DocLabel.EtqRet.length() ) );
            }
            else
            // Parse params info
            if ( docLine.toLowerCase().startsWith( DocLabel.EtqParam ) ) {
                status = DocLabel.param;
                docLine = sub( docLine, DocLabel.EtqParam.length() ).trim();
                String id = extractParamName( docLine );
                toret.setParamDoc( id, docLine.substring( id.length() ).trim() );
            }
            else
            // It is the hide label
            if ( docLine.toLowerCase().startsWith( DocLabel.EtqHide ) ) {
                toret.setHidden( true );
                status = DocLabel.hide;
            }
            // It is something related to the last doc item
            else {
                String docText = toret.get( status );

                if ( docText != null ) {
                    docText += " " + docLine;
                } else {
                    docText = docLine;
                }
                
                toret.set( status, docText );
            }
        }

        return toret;
    }
    
    /**
     * Substring: Avoids exceptions thrown when cutting text from one point to end.
     * @param txt The String
     * @param beg The beginning index
     * @return a new string with that substring
     */
    public static String sub(String txt, int beg)
    {
        return sub( txt, beg, txt.length() );
    }
    
    /**
     * Substring: Avoids exceptions thrown when cutting text.
     * @param txt The String
     * @param beg The beginning index
     * @param end The finishing index
     * @return a new string with that substring
     */
    public static String sub(String txt, int beg, int end)
    {
        String toret = "";
        
        // Adjust beginning
        if ( beg < 0 ) {
            beg = 0;
        }

        // Get substring, if possible
        if ( txt.length() > beg ) {
            if ( txt.length() < end ) {
                end = txt.length();
            }

            toret = txt.substring( beg, end );
        }

        return toret;
    }
    
    /**
     * Retrieves the param name, at the beginning of the input string.
     * @param txt The input string.
     * @return The param name, as a String.
     */
    public static String extractParamName(String txt)
    {
        int pos = 0;
        StringBuilder str = new StringBuilder();
        
        txt = txt.trim();
        while ( pos < txt.length() ) {
            char ch = txt.charAt( pos );
            if ( !Character.isWhitespace( ch ) ) {
                str.append( ch );
            } else {
                break;
            }

            ++pos;
        }
        
        return str.toString();
    }

    private HashMap<String, String> doc = null;
    private HashMap<String, String> params = null;
    private String lastArgKeySet;
}
