// LiteralCache.java

package core.intermediatecode;

import core.exceptions.SemanticError;
import core.exceptions.syntax.IncorrectId;
import core.intermediatecode.instrs.Fin;
import core.intermediatecode.literals.BoolLiteral;
import core.intermediatecode.literals.CharLiteral;
import core.intermediatecode.literals.DoubleLiteral;
import core.intermediatecode.literals.IntLiteral;
import core.intermediatecode.literals.StringLiteral;
import core.intermediatecode.visibilities.PrivateVisibility;
import core.lexical.Id;
import core.lexical.Position;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * Literal cache is a cache to avoid creating multiple copies of
 * a given literal. Typical examples are 0's and 1's
 * @author baltasarq
 */
public class LiteralCache {
    /**
     * Prepares a new instance of a cache of literals
     */
    public LiteralCache(CodeUnit myCodeUnit)
    {
        this.owner = myCodeUnit;
        this.charLiterals = new HashMap<Character, Fin>();
        this.intLiterals = new HashMap<Integer, Fin>();
        this.doubleLiterals = new HashMap<Double, Fin>();
        this.stringLiterals = new HashMap<String, Fin>();
        this.boolLiterals = new HashMap<Boolean, Fin>();
    }
    
    /**
     * Returns the code unit this cache literal belongs to
     * @return The code unit, as a CodeUnit object.
     * @see CodeUnit
     */
    public CodeUnit getCodeUnit()
    {
        return this.owner;
    }
    
    /**
     * Returns the id of a new IntLiteral, if it does not exist
     * @param x The new int to create as a new literal
     * @return The id of the constant for that literal
     */
    public Id get(int x) throws IncorrectId
    {
        Fin toret = this.intLiterals.get( x );
        
        if ( toret == null ) {
            toret = this.register( new IntLiteral( x ) );
            this.intLiterals.put( x, toret );
        }
        
        return toret.getId();
    }
    
    /**
     * Returns the id of a new CharLiteral, if it does not exist
     * @param x The new char to create as a new literal
     * @return The id of the constant for that literal
     */
    public Id get(char x) throws IncorrectId
    {
        Fin toret = this.charLiterals.get( x );
        
        if ( toret == null ) {
            toret = this.register( new CharLiteral( x ) );
            this.charLiterals.put( x, toret );
        }
        
        return toret.getId();
    }
    
    /**
     * Returns the id of a new Literal, if it does not exist
     * @param x The new char to create as a new literal
     * @return The id of the constant for that literal
     */
    public Id get(Literal lit) throws IncorrectId, SemanticError
    {
        Id toret = null;
                
        if ( lit instanceof CharLiteral ) {
            toret = this.get( ( (CharLiteral) lit ).getLiteral() );
        }
        else
        if ( lit instanceof IntLiteral ) {
            toret = this.get( ( (IntLiteral) lit ).getLiteral() );
        }
        else
        if ( lit instanceof DoubleLiteral ) {
            toret = this.get( ( (DoubleLiteral) lit ).getLiteral() );
        }
        else
        if ( lit instanceof StringLiteral ) {
            toret = this.get( ( (StringLiteral) lit ).getLiteral() );
        }
        else {
            toret = this.register( lit ).getId();
        }

        
        return toret;
    }
    
    /**
     * Returns the id of a new DoubleLiteral, if it does not exist
     * @param x The new double to create as a new literal
     * @return The id of the constant for that literal
     */
    public Id get(double x) throws IncorrectId
    {
        Fin toret = this.doubleLiterals.get( x );
        
        if ( toret == null ) {
            toret = this.register( new DoubleLiteral( x ) );
            this.doubleLiterals.put( x, toret );
        }
        
        return toret.getId();
    }
    
    /**
     * Returns the id of a new StringLiteral, if it does not exist
     * @param x The new String to create as a new literal
     * @return The id of the constant for that literal
     */
    public Id get(String x) throws IncorrectId, SemanticError
    {
        Fin toret = this.stringLiterals.get( x );
        
        if ( toret == null ) {
            toret = this.register( new StringLiteral( x ) );
            this.stringLiterals.put( x, toret );
        }
        
        return toret.getId();
    }
    
    /**
     * Returns the id of a new BoolLiteral, if it does not exist
     * @param x The new boolean to create as a new literal
     * @return The id of the constant for that literal
     */
    public Id get(boolean x) throws IncorrectId
    {
        Fin toret = this.boolLiterals.get( x );
        
        if ( toret == null ) {
            toret = this.register( new BoolLiteral( x ) );
            this.boolLiterals.put( x, toret );
        }
        
        return toret.getId();
    }
    
    /**
     * Returns all the literals in the cache
     * @return The literals cached, as a single vector Fin[]
     */
    public Fin[] getAll()
    {
        ArrayList<Fin> toret = new ArrayList<Fin>();
        
        // Add all literals cached
        toret.addAll( this.charLiterals.values() );
        toret.addAll( this.intLiterals.values() );
        toret.addAll( this.doubleLiterals.values() );
        toret.addAll( this.stringLiterals.values() );
        toret.addAll( this.boolLiterals.values() );
        
        return toret.toArray( new Fin[ toret.size()] );
        
    }
    
    /**
     * Returns a fake position. Hopefully, this will neve be used.
     * @return A Position object to be used in new constants.
     */
    protected Position getPos()
    {
        return new Position( this.getCodeUnit().getName() + ".LiteralCache", 0, 0 );
    }
    
    /**
     * Given a literal, creates a constant and inserts it in the CodeUnit
     * @param lit
     * @return 
     */
    protected Fin register(Literal lit) throws IncorrectId
    {
        Fin toret = new Fin( this.getPos(), PrivateVisibility.get(),
                             Id.createTemp(), lit );
        this.getCodeUnit().getInstrs().add( toret );
        
        return toret;
    }
    
    protected HashMap<Character, Fin> charLiterals;
    protected HashMap<Integer, Fin> intLiterals;
    protected HashMap<Double, Fin> doubleLiterals;
    protected HashMap<String, Fin> stringLiterals;
    protected HashMap<Boolean, Fin> boolLiterals;
    protected CodeUnit owner = null;
}
