/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package core.intermediatecode;

import core.exceptions.SemanticError;
import core.exceptions.syntax.IncorrectId;
import core.lexical.Id;
import core.lexical.Position;
import core.I18n;

/**
 * A RefId should be normally given complete, which means that it has a prefix,
 * i.e. the name of the module (first part), a dot, and the name
 * of the member itself (second part).
 * @author baltasarq
 */
public class RefId implements RValue, LValue {
    
    /**
     * Creates temporal refIds, to be used internally
     * @return A new RefId for internal use.
     * @throws IncorrectId 
     */
    public static RefId createTemp() throws IncorrectId
    {
        return new RefId(
                new Position( "RefId.tmp", 0, 0 ),
                Id.TmpId + Integer.toString( count++ )
        );
    }

    /**
     * A typical id, normally given when the referenced member is inside the
     * same CodeUnit
     * @param pos The position in which this reference happens.
     * @param name The id for this reference.
     */
    public RefId(Position pos, String name) throws IncorrectId
    {
        this( pos, null, name );
    }
        
    /**
     * A convenience constructor, together with toId()
     * @param name The id for this reference.
     */
    public RefId(Id name) throws IncorrectId
    {
        this( name.getPos(), null, name.getId() );
    }
    
    /**
     * A complete id, with its prefix and name
     * @param prefix The name of the module in which the reference member exists
     * @param name  The name of the referenced member itself
     */
    public RefId(Position pos, String prefix, String name) throws IncorrectId
    {
        this.prefix = prefix;
        this.name = name;
        this.pos = pos;
        
        this.chk();
    }
    
    /**
     * Convenience function to easily convert a RefId in an Id
     * @return This RefId (only the 'name' or second part) as an Id object.
     * @throws IncorrectId 
     */
    public Id toId() throws IncorrectId
    {
        return new Id( this.getName(), this.getPos() );
    }
    
    /**
     * Returns the complete identifier
     * @return The complete identifier, as a String
     */
    public String getId()
    {
        return this.toString();
    }
    
    /**
     * Returns the name or second part of the RefId
     * @return The name, as a String
     */
    public String getName()
    {
        return this.name;
    }
    
    /**
     * Returns the position in which this RefId was defined
     * @return The position, as Position object.
     */
    public Position getPos()
    {
        return this.pos;
    }
    
    /**
     * Returns the prefix or first part of the RefID
     * @return The prefix, as a String
     */
    public String getPrefix()
    {
        return this.prefix;
    }
    
    /**
     * Raises an exception, id's have no type (it can be data or function).
     * @return It does not never return anything.
     * @throws SemanticError always.
     */
    @Override
    public Type getType() throws SemanticError
    {
        throw new SemanticError( Instr.getPos(), "jcg.id.getType(); "
                + I18n.get( I18n.Word.Type ) + '?' );
    }
    
    /**
     * Checks whether the RefId is well-formed.
     * @throws IncorrectId 
     */
    @Override
    public final void chk() throws IncorrectId
    {
        Id.chkId( this.getId() );
        
        if ( this.hasPrefix() ) {
            Id.chkId( this.getPrefix() );
        }
    }
    
    /**
     * Determines whether this RefId has a prefix.
     * @return true if it has a prefix, false otherwise.
     */
    public boolean hasPrefix()
    {
        boolean toret = ( this.getPrefix() != null);
        
        if ( toret ) {
            toret = ( this.getPrefix().trim().length() > 0 );
        }
        
        return toret;
    }
    
    /**
     * Sets the prefix for this RefId.
     * Probably it was not given, and now there is the possibility
     * of completion. The prefix won't be changed if there is found not to be
     * correct.
     * @param prefix The new prefix for this RefId
     */
    public void setPrefix(String prefix) throws IncorrectId
    {
        this.prefix = prefix;        
        this.chk();
    }
    
    /**
     * Returns the complete identifier, in the form 'prefix.name'
     * @return The complete identifier, as String
     */
    @Override
    public String toString()
    {
        String toret = "";
        
        if ( this.hasPrefix() ) {
            toret = this.getPrefix() + '.';
        }
        
        return ( toret + this.getName() );
    }
    
    private String prefix;
    private String name;
    private Position pos;
    private static int count = 1;

}
