// Type.java

package core.intermediatecode;

import core.CompilingError;
import core.I18n;
import core.exceptions.SemanticError;
import core.intermediatecode.literals.ArrayLiteral;
import core.intermediatecode.literals.BoolLiteral;
import core.intermediatecode.literals.CharLiteral;
import core.intermediatecode.literals.DoubleLiteral;
import core.intermediatecode.literals.IntLiteral;
import core.intermediatecode.literals.StringLiteral;
import core.intermediatecode.types.AnyType;
import core.intermediatecode.types.ArrayType;
import core.intermediatecode.types.BoolType;
import core.intermediatecode.types.CharType;
import core.intermediatecode.types.DoubleType;
import core.intermediatecode.types.IntType;
import core.intermediatecode.types.NumericalType;
import core.intermediatecode.types.StringType;
import core.lexical.Position;
import java.lang.reflect.Array;

/**
 * Represents types for literals and variables
 * @author baltasarq
 */
abstract public class Type {
    @Override
    public abstract String toString();
    public abstract Class getJavaType();

    public static Type getType(String strType)
    {
        Type toret = null;

        strType = strType.trim();

        if ( strType.equals( IntType.Name ) ) {
            toret = IntType.get();
        }
        else
        if ( strType.equals( DoubleType.Name ) ) {
            toret = DoubleType.get();
        }
        else
        if ( strType.equals( CharType.Name ) ) {
            toret = CharType.get();
        }
        else
        if ( strType.equals( BoolType.Name ) ) {
            toret = BoolType.get();
        }
        else
        if ( strType.equals( AnyType.Name ) ) {
            toret = AnyType.get();
        }
        else
        if ( strType.endsWith( ArrayType.Name ) ) {
            int dimensions = 1;

            strType = strType.substring( 0, strType.length() - ArrayType.Name.length() );
            while( strType.endsWith( ArrayType.Name ) ) {
                ++dimensions;
                strType = strType.substring( 0, strType.length() - ArrayType.Name.length() );
            }

            Type type = Type.getType( strType );

            if ( type != null ) {
                toret = new ArrayType( type, dimensions );
            }
        }

        return toret;
    }

    public static boolean areCompatible(Type type1, Type type2)
    {
        boolean toret = ( type1 == type2 );

        if ( !toret ) {
            // Check whether one of them is "any"
            toret = ( type1 instanceof AnyType
                   || type2 instanceof AnyType );

            // Check whether both are numerical
            if ( !toret ) {
                toret = ( type1 instanceof NumericalType
                       && type2 instanceof NumericalType );
                
                // Check whether both are arrays
                if ( !toret 
                  && type1 instanceof ArrayType )
                {
                    toret = ( (ArrayType) type1 ).isCompatible( type2 );
                }
            }
        }

        return toret;
    }

    /**
     * Creates an appropriate, empty literal object for this type.
     * @param type The type, as a Type object.
     * @return The new literal object, as a Literal object.
     * @throws SemanticError
     */
    public static Literal createLiteral(Type type) throws SemanticError
    {
        Literal toret = null;

        if ( type == StringType.get() ) {
            toret = new StringLiteral( "" );
        }
        else
        if ( type instanceof ArrayType ) {
            toret = new ArrayLiteral( (ArrayType) type);
        }
        else
        if ( type == IntType.get() ) {
            toret = new IntLiteral( 0 );
        }
        else
        if ( type == DoubleType.get() ) {
            toret = new DoubleLiteral( 0 );
        }
        else
        if ( type == BoolType.get() ) {
            toret = new BoolLiteral( false );
        }
        else
        if ( type == CharType.get() ) {
            toret = new CharLiteral( '0' );
        }

        if ( toret == null ) {
            throw new SemanticError(
                    new Position( "jcg.Type.cnvtType()", 0, 0 ),
                    I18n.get( I18n.Word.Unexpected )
                    + " literal " + I18n.get( I18n.Word.Of )
                    + ' ' + I18n.get( I18n.Word.Type )
                    + ": " + type.toString()
            );
        }

        return toret;
    }
    
    public static Object createNewPrimitiveJavaObject(Class cl, String value) throws CompilingError
    {
        Object toret = null;

        value = value.trim();
        if ( !value.isEmpty() ) {
            try {

                if ( cl == Boolean.class
                  || cl == boolean.class )
                {
                    toret = Boolean.parseBoolean( value );
                }
                else
                if ( cl == Character.class
                  || cl == char.class )
                {
                    toret = CharLiteral.parse( value ).getLiteral();
                }
                else
                if ( cl == Double.class
                  || cl == double.class )
                {
                    toret = Double.parseDouble( value );
                }
                else
                if ( cl == Integer.class
                  || cl == int.class )
                {
                    toret = Integer.parseInt( value );
                }
                else
                if ( cl == String.class
                  || cl == ( Array.newInstance( char.class, 0 ) ).getClass() )
                {
                    toret = value.toCharArray();
                }

            } catch(CompilingError exc) {
                throw exc;
            } catch(Exception exc) {
                toret = null;
            }
        }

        if ( toret == null ) {
            throw new SemanticError(
                    new Position( "jcg.Type.createNewPrimitiveJavaObject()", 0, 0 ),
                    I18n.get( I18n.Word.Type )
                    + ' ' + I18n.get( I18n.Word.Unexpected )
                    + ": " + cl.getName()
            );
        }

        return toret;
    }

    /**
     * Converts the type as loaded from the jar file, and creates an appropriate type.
     * @param type The type, as a java.lang.reflect.Type object.
     * @return The type for the compiler, as a Type object.
     * @throws SemanticError when the type is not supported.
     */
    public static Type cnvtJavaToType(Class type) throws SemanticError
    {
        int dim = 0;
        Type toret = null;

        // Extract type
        if ( type == String.class ) {
            toret = StringType.get();
        }
        else
        if ( type.isArray() ) {
            dim = 1;
            type = type.getComponentType();
            while ( type.isArray() ) {
                type = type.getComponentType();
                ++dim;
            }

            toret = new ArrayType( cnvtJavaToType( type ), dim );
        }
        else
        if ( type == Object.class ) {
            toret = AnyType.get();
        }
        else
        if ( type == int.class
          || type == Integer.class )
        {
            toret = IntType.get();
        }
        else
        if ( type == double.class
          || type == Double.class )
        {
            toret = DoubleType.get();
        }
        else
        if ( type == boolean.class
          || type == Boolean.class )
        {
            toret = BoolType.get();
        }
        else
        if ( type == char.class
          || type == Character.class )
        {
            toret = CharType.get();
        }

        if ( type == null ) {
            throw new SemanticError(
                    new Position( "Type.cnvtJavaToType", 0, 0 ),
                    I18n.get( I18n.Word.Type )
                    + ' ' + I18n.get( I18n.Word.Unexpected )
                    + ": " + type.toString()
            );
        }

        return toret;
    }
    
    /**
     * Returns the HTML representation.
     * @return HTML in a string.
     */
    public String toHtml()
    {
        return "<b><font color=\"blue\">" + this.toString() + "</font></b>";
    }

}
