// Acc.java

package core.intermediatecode.instrs;

import core.intermediatecode.Data;
import core.Ast;
import core.CompilingError;
import core.I18n;
import core.exceptions.SemanticError;
import core.exceptions.syntax.IncorrectId;
import core.intermediatecode.LValue;
import core.intermediatecode.RefId;
import core.intermediatecode.Type;
import core.intermediatecode.types.ArrayType;
import core.lexical.Position;

/**
 * Represents access to vector positions, so data can be retrieved
 * @author baltasarq
 */
public class Acc extends Op implements LValue {

    public Acc(Position position, String id, RefId[] pos) throws IncorrectId
    {
        this( position, new RefId( position, id ), pos );
    }

    public Acc(Position position, RefId id, RefId[] pos)
    {
        super( position, 1 );

        this.ops[ 0 ] = id;
        this.position = pos;
    }
    
    /**
     * Returns the id of the collection this access belongs to.
     * @return The RefId belonging to the "v" in the "v[i]" total.
     */
    public RefId getId()
    {
        return this.getOps()[ 0 ];
    }

    /**
     * Returns a vector that enumerated row, column, depth... of the position
     * @return The position, as a vector of RefIds.
     */
    public RefId[] getElementPosition()
    {
        return this.position;
    }

    /**
     * Returns the number of dimensions needed.
     * @return The number of dimensions, as an int.
     */
    public int getDimensionsNeeded()
    {
        return this.getElementPosition().length;
    }

    @Override
    public String toString() {
        StringBuilder toret = new StringBuilder();

        toret.append( this.getOps()[ 0 ] );

        for(int i = 0; i < this.getDimensionsNeeded(); ++i) {
            toret.append( '[' );
            toret.append( this.getElementPosition()[ i ].toString() );
            toret.append( ']' );
        }

        return toret.toString();
    }

    @Override
    public void chk() throws CompilingError
    {
        Data data = Ast.getAst().locateData( this.getPosition(), this.getOwner(), this.getOps()[ 0 ] );

        if ( data == null ) {
            throw new SemanticError( this.getPosition(), I18n.get( I18n.Word.Array )
                    + ' ' + I18n.get( I18n.Word.Expected )
            );
        }
        
        Type type = data.getType();

        if ( !( type instanceof ArrayType ) ) {
            throw new SemanticError( this.getPosition(), I18n.get( I18n.Word.Expected )
                    + ' ' + I18n.get( I18n.Word.Array )
            );
        }

        ArrayType arrayType = (ArrayType) type;

        if ( arrayType.getDimensions() != this.getDimensionsNeeded() ) {
            throw new SemanticError( this.getPosition(), "# "
                    + I18n.get( I18n.Word.Of )
                    + ' ' + I18n.get( I18n.Word.Dimensions )
                    + ' ' + I18n.get( I18n.Sentence.DoesNotMatchWith )
            );
        }
    }

    @Override
    public Type getType() throws CompilingError
    {
        Data data = Ast.getAst().locateData(
                this.getPosition(),
                this.getOwner(),
                this.getId()
        );
        
        Type type = data.getType();
        ArrayType arrayType = null;
        
        if ( type instanceof ArrayType ) {
            arrayType = (ArrayType) type;
        } else {
            throw new SemanticError( this.getPosition(), I18n.get( I18n.Word.Expected )
                    + ' ' + I18n.get( I18n.Word.Array )
            );
        }

        return arrayType.getType();
    }

    private RefId[] position;

}
