// BinaryOp.java

package core.intermediatecode.instrs;

import core.bshbackend.Reserved;
import core.intermediatecode.Data;
import core.intermediatecode.Type;
import core.CompilingError;
import core.exceptions.SemanticError;
import core.exceptions.SyntaxError;
import core.Ast;
import core.intermediatecode.Instr;
import core.intermediatecode.RefId;
import core.intermediatecode.types.BoolType;
import core.intermediatecode.types.NumericalType;
import core.intermediatecode.types.StringType;
import core.lexical.Position;
import core.I18n;

/**
 * Represents binary operators
 * @author baltasarq
 */
public abstract class BinaryOp extends Op {

    /** A binary operator happens between two refIds
     * @param pos The position in the lexer
     * @param op1 The first operand (RefId to it)
     * @param op2 The second operand (RefId to it)
     */
    public BinaryOp(Position pos, RefId op1, RefId op2)
    {
        super( pos, 2 );

        this.ops[ 0 ] = op1;
        this.ops[ 1 ] = op2;
    }

    /**
     * Return the first operator
     * @return The id of the variable, as an Id object.
     */
    public RefId getOp1()
    {
        return this.getOps()[ 0 ];
    }

    /**
     * Return the second operator
     * @return The id of the variable, as an Id object.
     */
    public RefId getOp2()
    {
        return this.getOps()[ 1 ];
    }

    @Override
    public void chk() throws CompilingError
    {
        Data vop1 = Ast.getAst().locateData( this.getPosition(), this.getOwner(), this.getOp1() );
        Data vop2 = Ast.getAst().locateData( this.getPosition(), this.getOwner(), this.getOp2() );

        if ( vop1 == null ) {
            throw new SemanticError( this.getPosition(), I18n.get( I18n.Word.Missing ) + " '" + this.getOp1().getId() + '\'' );
        }

        if ( vop2 == null ) {
            throw new SemanticError( this.getPosition(), I18n.get( I18n.Word.Missing ) + " '" + this.getOp2().getId() + '\'' );
        }

        Type top1 = vop1.getType();
        Type top2 = vop2.getType();

        if ( !Type.areCompatible( top1, top2 ) ) {
            throw new SemanticError( this.getPosition(),
                    I18n.get( I18n.Sentence.DoesNotMatch )
                    + ": " + top1.toString()
                    + " & " + top2.toString()
            );
        }

        if ( !( this instanceof LogicBinaryOp ) ) {
            // Is it a string type?
            if ( top1 instanceof StringType
              || top2 instanceof StringType )
            {
                throw new SemanticError( this.getPosition(), "opr. "
                    + I18n.get( I18n.Word.Unexpected )
                    + ' ' + I18n.get( I18n.Word.For )
                    + ' ' + I18n.get( I18n.Word.Type )
                    + ' ' + StringType.Name
                );
            }
        } else {
            // Special junctor
            if ( this instanceof And
              || this instanceof Or )
            {
                if ( top1 != BoolType.get()
                  || top2 != BoolType.get() )
                {
                    throw new SemanticError( this.getPosition(),
                            I18n.get( I18n.Word.Type )
                            + ' ' + I18n.get( I18n.Word.Of )
                            + " opr. (bool) "
                            + I18n.get( I18n.Sentence.DoesNotMatchWith )
                            + ' ' + I18n.get( I18n.Word.Type )
                            + ' ' + I18n.get( I18n.Word.Of )
                            + " expr."
                     );
                }
            }
        }
        
        return;
    }

    @Override
    public Type getType() throws CompilingError
    {
        Data vop1 = Ast.getAst().locateData( this.getPosition(), this.getOwner(), this.getOp1() );
        Data vop2 = Ast.getAst().locateData( this.getPosition(), this.getOwner(), this.getOp2() );
        Type toret = null;

        // Look for data
        if ( vop1 == null ) {
            throw new SemanticError( this.getPosition(), I18n.get( I18n.Word.Missing ) + ": '" + this.getOp1().getId() + '\'' );
        }
        
        if ( vop2 == null ) {
            throw new SemanticError( this.getPosition(), I18n.get( I18n.Word.Missing ) + ": '" + this.getOp2().getId() + '\'' );
        }
        
        // Get type of first arg
        toret = vop1.getType();
        
        // Adjust in case of being numerical
        if ( toret instanceof NumericalType ) {
            toret = NumericalType.getWidestOfTypes( toret, vop2.getType() );
        }
        
        return toret;
    }
    
    /**
     * Creates the appropriate operator, following the opr char given.
     * The remaining parameters are needed for constructing the operator.
     * @param pos The position in the Lexer
     * @param opr The char to be honored, as in the operator to build
     * @param op1 The RefId to the first operand
     * @param op2 The RefId to the second operand
     * @return A new created BinaryOp
     * @throws SyntaxError When the opr char does not identify any operator
     */
    public static BinaryOp createOp(Position pos, String opr, RefId op1, RefId op2) throws SyntaxError
    {
        BinaryOp toret = null;
        opr = opr.trim();

        // Maths operators
        if ( opr.equals( Reserved.AddOperator ) ) {
            toret = new Add( pos, op1, op2 );
        }
        else
        if ( opr.equals( Reserved.SubOperator ) ) {
            toret = new Sub( pos, op1, op2 );
        }
        else
        if ( opr.equals( Reserved.MulOperator ) ) {
            toret = new Mul( pos, op1, op2 );
        }
        else
        if ( opr.equals( Reserved.DivOperator ) ) {
            toret = new Div( pos, op1, op2 );
        }
        else
        if ( opr.equals( Reserved.ModOperator ) ) {
            toret = new Mod( pos, op1, op2 );
        }
        else
        // Conditional operators
        if ( opr.equals( Reserved.LtOperator ) ) {
            toret = new Lt( pos, op1, op2 );
        }
        else
        if ( opr.equals( Reserved.GtOperator ) ) {
            toret = new Gt( pos, op1, op2 );
        }
        else
        if ( opr.equals( Reserved.EqOperator ) ) {
            toret = new Eq( pos, op1, op2 );
        }
        else
        if ( opr.equals( Reserved.NeqOperator ) ) {
            toret = new Neq( pos, op1, op2 );
        }
        else
        if ( opr.equals( Reserved.LteOperator ) ) {
            toret = new Lte( pos, op1, op2 );
        }
        else
        if ( opr.equals( Reserved.GteOperator ) ) {
            toret = new Gte( pos, op1, op2 );
        }
        else
        if ( opr.equals( Reserved.AndOperator ) ) {
            toret = new And( pos, op1, op2 );
        }
        else
        if ( opr.equals( Reserved.OrOperator ) ) {
            toret = new Or( pos, op1, op2 );
        }

        // Chk & end
        if ( toret == null ) {
            throw new SyntaxError( Instr.getPos(), "jcg.BinaryOp.createOp(): " + opr + '?' );
        }
        
        return toret;
    }

}
