// Cal.java

package core.intermediatecode.instrs;

import core.Ast;
import core.CompilingError;
import core.I18n;
import core.exceptions.SemanticError;
import core.exceptions.syntax.IncorrectId;
import core.intermediatecode.*;
import core.lexical.Position;

import java.util.ArrayList;
import java.util.Arrays;

/**
 * Represents function calls
 * @author baltasarq
 */
public class Cal extends Instr implements RValue {
    public static final String Name = "CAL";

   /**
     * Builds a call to a function with arguments.
     * @param functionId The id of the function to be called.
     */
    public Cal(Position pos, String functionId) throws IncorrectId
    {
        this( pos, new RefId( pos, functionId ) );
        this.setLinkedFun( null );
    }

    /**
     * Builds a call to a function with arguments.
     * @param functionId The id of the function to be called.
     */
    public Cal(Position pos, String moduleId, String functionId) throws IncorrectId
    {
        this( pos, new RefId( pos, moduleId, functionId ) );
    }

    /**
     * Builds a call to a function with arguments.
     * @param functionId The id of the function to be called.
     */
    public Cal(Position pos, RefId functionId)
    {
        this( pos, functionId, new RefId[0] );
    }

    /**
     * Builds a call to a function with arguments.
     * @param pos The position of the call in the source code.
     * @param moduleId The id of the module the call resides in
     * @param functionId The id of the function to be called.
     * @param args A vector of RValue's as the argument list.
     */
    public Cal(Position pos, String moduleId, String functionId, RefId[] args) throws IncorrectId
    {
        this( pos, new RefId( pos, moduleId, functionId ), args );
    }

    /**
     * Builds a call to a function with arguments.
     * @param pos The position of the call in the source code.
     * @param functionId The id of the function to be called.
     * @param args A vector of RValue's as the argument list.
     */
    public Cal(Position pos, RefId functionId, RefId[] args)
    {
        super( pos );

        this.functionId = functionId;
        this.args = new ArrayList<RefId>();
        this.args.addAll( Arrays.asList( args ) );
    }

    /**
     * Returns the name of the function to be called.
     * @return The name as an Id.
     */
    public RefId getFunctionId()
    {
        return this.functionId;
    }

    /**
     * Adds an rvalue to the args list.
     * @param rvalue An RValue to add.
     */
    public void addArg(RefId rvalue)
    {
        this.args.add( rvalue );
    }

    /**
     * Returns the numbers of arguments.
     * @return
     */
    public int getNumberOfArgs()
    {
        return this.args.size();
    }

    /**
     * Returns the argument at position i.
     * @param i The position of the param to get.
     * @return The argument, as an RefId object.
     */
    public RefId getArg(int i)
    {
        return this.args.get( i );
    }
    
    /**
     * Returns all arguments as a single vector of RefId's.
     * @return The arguments, as a primitive RefId vector.
     */
    public RefId[] getArgs()
    {
        return ( this.args.toArray( new RefId[ this.args.size() ] ) );
    }

    @Override
    public String toString() {
        StringBuilder toret = new StringBuilder();
        
        
        toret.append( Name );
        toret.append( ' ' );

        toret.append( this.getFunctionId().toString() );
        toret.append( '(' );

        for(RValue rvalue: this.args) {
            toret.append( rvalue.toString() );
            toret.append( ' ' );
        }

        toret.append( ')' );
        return toret.toString();
    }

    @Override
    public void chk() throws CompilingError
    {
        CodeUnit functionOwner = CodeBlock.getModuleOwner( this.getOwner() );

        // The owner of the cal must be a function or a module itself
        // The owner of any of the previous must be the module.
        if ( functionOwner == null ) {
            throw new SemanticError(
                     this.getPosition(),
                    "error fn.: " + this.getFunctionHeader()
            );
        }

        this.getType();
    }

    @Override
    public Type getType() throws CompilingError
    {
        Type toret = null;

        if ( this.getLinkedFun() == null ) {
            final Fun fun = Ast.getAst().locateFunction( this.getPosition(), this, this.getFunctionId() );

            if ( fun != null ) {
                this.setLinkedFun( fun  );
            } else {
                throw new SemanticError(
                        this.getPosition(),
                        I18n.get( I18n.Word.Missing )
                                + " fn.: " + this.getFunctionHeader()
                );
            }
        }

        toret = this.getLinkedFun().getType();
        return toret;
    }
    
    public String getFunctionHeader() throws SemanticError
    {
        Type type = null;
        StringBuilder toret = new StringBuilder();
        Data data;
        
        toret.append( this.getFunctionId().toString() );
        toret.append( '(' );
        
        for(int i = 0; i < this.getArgs().length; ++i) {
            try {
                data = Ast.getAst().locateData(
                    this.getPosition(), this.getOwner(), this.getArg( i )
                );
            } catch(Exception e) {
                data = null;
            }
            
            if ( data == null ) {
                toret.append( "<unknown>" );
            } else {
                try {
                    toret.append( data.getType().toString() );
                } catch(Exception exc) {
                    toret.append( "<unknown>" );
                }
            }
            
            if ( i < ( this.getArgs().length -1 ) ) {
                toret.append( ',' );
                toret.append( ' ' );
            } 
        }
        
        toret.append( ')' );        
        return toret.toString();
    }

    public Fun getLinkedFun() {
        return linkedFun;
    }

    public void setLinkedFun(Fun linkedFun) {
        this.linkedFun = linkedFun;
    }

    private RefId functionId;
    private ArrayList<RefId> args;
    private Fun linkedFun;
}
