// Fun.java

package core.intermediatecode.instrs;

import core.CompilingError;
import core.I18n;
import core.bshbackend.Reserved;
import core.exceptions.SemanticError;
import core.exceptions.syntax.IncorrectId;
import core.intermediatecode.CodeBlock;
import core.intermediatecode.CodeBlockOwner;
import core.intermediatecode.CodeUnit;
import core.intermediatecode.Data;
import core.intermediatecode.Documentation;
import core.intermediatecode.Instr;
import core.intermediatecode.Member;
import core.intermediatecode.Prg;
import core.intermediatecode.Type;
import core.intermediatecode.visibilities.PublicVisibility;
import core.intermediatecode.visibilities.Visibility;
import core.lexical.Id;
import core.lexical.Position;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * Definition of a function
 * @author baltasarq
 */
public class Fun extends Instr implements CodeBlockOwner, Member {
    public static String Name = "FUN";

    public Fun(Position position, CodeUnit cdu, Type returnType, String name) throws IncorrectId
    {
        this( position, cdu, PublicVisibility.get(), returnType, new Id( name, position ), new Data[0] );
    }

    public Fun(Position position, CodeUnit cdu, Visibility visible, Type returnType, String name) throws IncorrectId
    {
        this( position, cdu, visible, returnType, new Id( name, position ), new Data[0]);
    }

    public Fun(Position position, CodeUnit cdu, Type returnType, String name, Data[] params) throws IncorrectId
    {
        this( position, cdu, PublicVisibility.get(), returnType, new Id( name, position ), params );
    }

    public Fun(Position position, CodeUnit cdu, Type returnType, Id name)
    {
        this( position, cdu, PublicVisibility.get(), returnType, name, new Data[0] );
    }

    public Fun(Position position, CodeUnit cdu, Visibility visible, Type returnType, Id name)
    {
        this( position, cdu, visible, returnType, name, new Data[0]);
    }

    public Fun(Position position, CodeUnit cdu, Type returnType, Id name, Data[] params)
    {
        this( position, cdu, PublicVisibility.get(), returnType, name, params );
    }

    public Fun(Position position, CodeUnit cdu, Visibility visible, Type returnType, Id name, Data[] params)
    {
        super( position );

        this.cdu = cdu;
        this.visible = visible;
        this.name = name;
        this.type = returnType;
        this.instrs = new CodeBlock( this );
        this.params = new ArrayList<Data>();
        this.doc = new Documentation();
        
        // Add formalParams
        if ( params != null ) {
            this.params.addAll( Arrays.asList( params ) );
            
            for(Data x: this.params) {
                x.setOwner( this );
            }
        }
    }

    /**
     * Returns the unit this function belongs to
     * @return The code unit, as a CodeUnit class.
     */
    public CodeUnit getCodeUnit()
    {
        return this.cdu;
    }
    
    /**
     * Returns the type of the function
     * @return the type of the function, as a Type object
     */
    @Override
    public Type getType()
    {
        return this.type;
    }
    
    /**
     * Returns the name of the function
     * @return The name of the function, as an Id object.
     */
    @Override
    public Id getId()
    {
        return this.name;
    }

    /**
     * Returns the list of formalParams for this function
     * @return
     */
    public Data[] getParams()
    {
        return this.params.toArray( new Data[ this.params.size() ]);
    }
    
    /**
     * Returns the types of the formal parameters
     * @return The types of the formal parameteres, as a vector of Type objects
     */
    public Type[] getParamTypes()
    {
        Data[] formalParams = this.getParams();
        Type[] toret = new Type[ formalParams.length ];
        
        for(int i = 0; i < formalParams.length; ++i) {
            toret[ i ] = formalParams[ i ].getType();
        }
        
        return toret;
    }

    public Data lookForParam(String id)
    {
        Data toret = null;

        lookInParams:
        for(Data var: this.getParams()) {
            if ( var.getId().toString().equals( id ) ) {
                toret = var;
                break lookInParams;
            }
        }

        return toret;
    }

    @Override
    public Fin lookForConst(String id)
    {
        return this.getInstrs().lookForConst( id );
    }

    @Override
    public Def lookForVar(String id)
    {
        return this.getInstrs().lookForVar( id );
    }

    @Override
    public Data lookForData(String id)
    {
        Data toret = this.lookForParam( id );

        if ( toret == null ) {
            toret = this.getInstrs().lookForData( id );
        }
        
        return toret;
    }
    
    @Override
    public Fun[] lookForFn(String id, Type[] paramTypes)
    {
        CodeUnit module = CodeBlock.getModuleOwner( this.getOwner() );
        
        return module.lookForFn( id, paramTypes );
    }

    /**
     * Returns the block of instructions inside this function.
     * @return
     */
    @Override
    public CodeBlock getInstrs()
    {
        return this.instrs;
    }

    /**
     * Returns the visibility of the function.
     * @return The visibility, as a Visibility object.
     */
    public Visibility getVisibility()
    {
        return this.visible;
    }

    /**
     * Decides whether the constant is visible or not.
     * @return The visibility, as a boolean value: true if public, false otherwise.
     */
    public boolean isVisible()
    {
        return this.visible.isVisible();
    }
    
    @Override
    public String toString()
    {
        StringBuilder toret = new StringBuilder();

        // Build the header of the function
        toret.append( Name );
        toret.append( " " );
        toret.append( this.getVisibility().isVisible() ? '+' : '-' );
        toret.append( " " );
        toret.append( this.getId().toString() );
        toret.append( "( " );

        for(int i = 0; i < this.getParams().length; ++i)
        {
            toret.append( this.getParams()[ i ].toString() );

            if ( i < ( this.getParams().length -1 ) ) {
                toret.append( ", " );
            }
        }

        toret.append( " ): " );
        toret.append( this.getType() == null ? "void" : this.getType().toString() );
        toret.append( '\n' );

        // Build the body of the function
        toret.append( this.getInstrs().toString() );

        return toret.toString();

    }

    @Override
    public void chk() throws CompilingError
    {
        Ret ret = null;
        
        // Check all instrs. inside the function
        for(Instr instr: this.getInstrs()) {
            instr.chk();
            
            if ( instr instanceof Ret ) {
                ret = (Ret) instr;
            }
        }
        
        // If a return type is declared, then there should also be a RET
        if ( this.getType() != null ) {
            if ( ret == null ) {
                String module = "";
                
                if ( !( this.getCodeUnit() instanceof Prg ) ) {
                    module = this.getCodeUnit().getName() + ".";
                }
                
                throw new SemanticError(
                        this.getPosition(), I18n.get( I18n.Word.Missing )
                        + ": " + Reserved.Return
                        + ". Fn.: " + this.getDesc()
                );
            }
        }
    }
    
    /**
     * Returns a printable description (as complete as needed) of the function
     * @return A description of this function, as a String object.
     */
    @Override
    public String getDesc()
    {
        String toret = "";
        String module = "";
                
        if ( !( this.getCodeUnit() instanceof Prg ) ) {
            module = this.getCodeUnit().getName() + ".";
        }
        
        toret = module + this.getId().getId();
        
        return ( toret + "()" );
    }
    
    @Override
    public Documentation getDoc()
    {
        return this.doc;
    }

    @Override
    public void setDoc(Documentation doc)
    {
        this.doc = doc;
    }
    
    @Override
    public String toHtml()
    {
        Type type = this.getType();
        StringBuilder toret = new StringBuilder();

        String returnComment = this.getDoc().getReturnDoc();
        String briefComment = this.getDoc().getBrief();

        // Header for the function
        toret.append( this.getVisibility().toHtml() );
        toret.append( " " );

        if ( type == null ) {
            toret.append( "<b><font color=\"blue\">void</font></b>" );
        } else {
            toret.append( this.getType().toHtml() );
        }

        toret.append( " <b>" );
        toret.append( this.getId() );
        toret.append( "</b>(" );
        toret.append( this.getParamsAsHtml() );
        toret.append( ")" );
        toret.append( "<br>\n<ul>" );

        // Brief comment
        if ( briefComment.length() > 0 ) {
            toret.append( "<li>" );
            toret.append( briefComment );
            toret.append( "</li>" );
            toret.append( "\n" );
        }

        // Params. of the function
        Data[] params = this.getParams();
        if ( params.length > 0 ) {
            toret.append( "<li><b>&raquo;</b><ul>\n" );

            // Comment for each parameter
            for(Data arg : params ) {
                String textComment = this.getDoc().getParamDoc( arg.getId().getId() );

                if ( textComment == null ) {
                    textComment = "";
                }

                toret.append( "<li><b>" );
                toret.append( arg.getType().toHtml() );
                toret.append( ' ' );
                toret.append( arg.getId() );
                toret.append( "</b>" );

                if ( textComment.length() > 0 ) {
                    toret.append( ": " );
                    toret.append( textComment );
                }

                toret.append( "</li>\n" );
            }

            toret.append( "</ul></li>\n" );
        }

        // Return comment
        if ( !returnComment.isEmpty() ) {
            toret.append( "<li><b>&laquo;</b><ul><li>" );
            toret.append( returnComment );
            toret.append( "</li></ul></il>\n" );
        }

        toret.append( "</ul>\n" );
        return toret.toString();
    }
    
    /**
     *  Returns the parameter list as HTML.
     *  @return The param list as HTML, in a String.
     */
    protected String getParamsAsHtml()
    {
        Data[] params = this.getParams();
        StringBuilder argsLine = new StringBuilder( "<i>" );

        // Build a comma separated list
        if ( params.length > 0 ) {
            for(int i = 0; i < params.length; ++i) {
                argsLine.append( params[ i ].toHtml() );

                if ( i < ( params.length -1 ) ) {
                    argsLine.append( ", " );
                }
            }
        }

        // End
        argsLine.append( "</i>");
        return argsLine.toString();
    }
    
    private CodeBlock instrs;
    private Visibility visible;
    private ArrayList<Data> params;
    private CodeUnit cdu;
    private Type type;
    private Id name;
    private Documentation doc;
}
