// ArrayLiteral.java

package core.intermediatecode.literals;

import core.CompilingError;
import core.I18n;
import core.exceptions.SemanticError;
import core.exceptions.SyntaxError;
import core.intermediatecode.Literal;
import core.intermediatecode.Type;
import core.intermediatecode.types.ArrayType;
import core.intermediatecode.types.CharType;
import core.intermediatecode.types.StringType;
import java.lang.reflect.Array;
import java.util.ArrayList;

/**
 * Represents array literals
 * @author baltasarq
 */
public class ArrayLiteral extends Literal {

    public static final char BeginCArrayDelimiter = '{';
    public static final char EndCArrayDelimiter = '}';

    public static final char BeginJSArrayDelimiter = '[';
    public static final char EndJSArrayDelimiter = ']';

    public static char BeginArrayDelimiter = BeginCArrayDelimiter;
    public static char EndArrayDelimiter = EndCArrayDelimiter;

    public static void setCArrayDelimiters()
    {
        setArrayDelimiters( BeginCArrayDelimiter, EndCArrayDelimiter );
    }

    public static void setJSArrayDelimiters()
    {
        setArrayDelimiters( BeginJSArrayDelimiter, EndJSArrayDelimiter );
    }

    public static void setArrayDelimiters(char begin, char end)
    {
        BeginArrayDelimiter = begin;
        EndArrayDelimiter = end;
    }

                public static class StringArrayLiteral {
                    public StringArrayLiteral(String lit)
                    {
                        this.lit = lit;
                        this.pos = 0;
                    }

                    public int getPos()
                    {
                        return this.pos;
                    }

                    public String getLit()
                    {
                        return this.lit;
                    }

                    public char getChar()
                    {
                        if ( !this.isEnd() ) {
                            return this.getLit().charAt( this.getPos() );
                        } else {
                            return '\0';
                        }

                    }

                    public void setPos(int pos)
                    {
                        this.pos = pos;
                    }

                    public void incPos() {
                        this.incPos( 1 );
                    }

                    public int size()
                    {
                        return this.getLit().length();
                    }

                    public void incPos(int x) {
                        this.pos += x;
                    }

                    public boolean isEnd()
                    {
                        return ( this.getPos() >= this.getLit().length() );
                    }

                    private int pos;
                    private String lit;
                }

    public ArrayLiteral(ArrayType type)
    {
        this.type = type;
        this.lit = Array.newInstance( type.getType().getJavaType(), type.getDimensions() );
    }

    public ArrayLiteral(ArrayType type, Object lit) throws SemanticError
    {
        this.type = type;
        this.lit = lit;
        
        this.chkArray();
    }

    @Override
    public String toString()
    {
        return ArrayToString( this.lit );
    }

    public static String ArrayToString(Object array)
    {
        final int length = Array.getLength( array );
        StringBuilder toret = new StringBuilder();
        Object element = Array.get( array, 0 );

        if ( array.getClass().isArray()
          && element instanceof Character )
        {
            toret.append( '"' );
            toret.append( String.valueOf( (char[]) array ) );
            toret.append( '"' );
        } else {

            toret.append( BeginArrayDelimiter );

            for(int i = 0; i < length; ++i) {
                element = Array.get( array, i );

                if ( element.getClass().isArray() ) {
                    toret.append( ArrayLiteral.ArrayToString( element) );
                } else {
                    if ( element instanceof Character ) {
                        toret.append( '\'' );
                        toret.append( element.toString() );
                        toret.append( '\'' );
                    } else {
                        toret.append( element.toString() );
                    }
                }

                if ( i < ( length -1 ) ) {
                    toret.append( ',' );
                }
            }

            toret.append(EndArrayDelimiter);
        }

        return toret.toString();
    }

    protected int calculateArrayDims()
    {
        int toret = 0;
        int len = 0;
        Object element = null;
        
        do {
            len = Array.getLength( this.lit );

            ++toret;
            
            if ( len > 0 ) {
                element = Array.get( this.lit, 0 );
            } else {
                break;
            }
        } while( element.getClass().isArray() ) ;

        return toret;
    }

    protected final void chkArray() throws SemanticError
    {
        Type type1 = Type.cnvtJavaToType( this.getLiteral().getClass() );
        Type type2 = this.getType();

        if ( this.getType().getDimensions() != this.calculateArrayDims() ) {
            throw new SemanticError( getPos(), I18n.get( I18n.Word.Dimensions )
                    + ' ' + I18n.get( I18n.Word.Of )
                    + ' ' + I18n.get( I18n.Word.Array )
                    + ' ' + I18n.get( I18n.Sentence.DoesNotMatch )
            );
        }

        if ( !( Type.areCompatible( type1, type2 ) ) ) {
            throw new SemanticError( getPos(), I18n.get( I18n.Word.Type )
                    + ' ' + I18n.get( I18n.Word.Of )
                    + ' ' + I18n.get( I18n.Word.Array )
                    + ' ' + I18n.get( I18n.Sentence.DoesNotMatchWith )
                    + ' ' + I18n.get( I18n.Word.Type )
                    + ' ' + I18n.get( I18n.Word.Of )
                    + " literal"
            );
        }

        return;
    }

    @Override
    public void setLiteral(String lit) throws CompilingError
    {
        Type componentType = this.getType().getType();

        if ( this.getType() instanceof StringType ) {
            componentType = StringType.get();
        }

        lit = removeExternalBraces( lit );
        StringArrayLiteral literal = new StringArrayLiteral( lit );
        this.lit = parseArray( componentType.getJavaType(), literal );

        chkArray();
    }

    public static Object parseArray(Class type, StringArrayLiteral literal) throws SemanticError
    {
        ArrayList elements = new ArrayList();
        Object toret = null;

        do {
            // Go past comma if needed
            if ( literal.getChar() == ',' ) {
                literal.incPos();
            }

            // Take element and store it
            skipSpaces( literal );
            if ( literal.getChar() == BeginCArrayDelimiter) {
                // Take sub array as element
                literal.incPos();
                elements.add( parseArray( type, literal ) );
            } else {
                try {
                    elements.add(
                        Type.createNewPrimitiveJavaObject( type, readToken( literal ) )
                    );
                } catch (Exception ex) {
                    System.out.println( ex.toString() );
                    throw new SemanticError( getPos(), I18n.get( I18n.Word.Type )
                        + ' ' + I18n.get( I18n.Word.Of )
                        + ' ' + I18n.get( I18n.Word.Array )
                        + ' ' + I18n.get( I18n.Sentence.DoesNotMatchWith )
                        + ' ' + I18n.get( I18n.Word.Type )
                        + ' ' + I18n.get( I18n.Word.Of )
                        + ' ' + I18n.get( I18n.Word.Element )
                    );
                }

            }
            
            skipSpaces( literal );
        } while( literal.getChar() == ','
              && !literal.isEnd() );

        // Go past final brace, if needed
        if ( !literal.isEnd() ) {
            if ( literal.getChar() == EndCArrayDelimiter) {
                literal.incPos();
            }
        }

        // Create the array
        Class realType = elements.get( 0 ).getClass();

        if ( !realType.isArray()
          && Type.cnvtJavaToType( type ) != Type.cnvtJavaToType( realType ) )
        {
            throw new SemanticError( getPos(), "jcg.parseArray(): "
                    + ' ' + I18n.get( I18n.Word.Type )
                    + ' ' + I18n.get( I18n.Word.Of )
                    + ' ' + I18n.get( I18n.Word.Array )
                    + ' ' + I18n.get( I18n.Sentence.DoesNotMatchWith )
                    + ' ' + I18n.get( I18n.Word.Type )
                    + ' ' + I18n.get( I18n.Word.Of )
                    + ' ' + I18n.get( I18n.Word.Element )
            );
        }

        // Adjust type for strings
        if ( !realType.isArray()
          && Type.cnvtJavaToType( realType ) == CharType.get() )
        {
            toret = Array.newInstance( char.class, elements.size() );
        } else {
            toret = Array.newInstance( realType, elements.size() );
        }

        for(int i = 0; i < elements.size(); ++i) {
            Array.set( toret, i, elements.get( i ) );
        }

        // Update and return
        return toret;
    }

    protected static String readToken(StringArrayLiteral lit) throws CompilingError
    {
        char ch;
        StringBuilder token = new StringBuilder();
        
        skipSpaces( lit );

        try {
            if ( !lit.isEnd() ) {
                ch = lit.getChar();

                if ( ch == '\'' ) {

                    lit.incPos();
                    token.append( lit.getChar() );
                    lit.incPos();
                    if ( lit.getChar() != '\'' ) {
                        throw new SyntaxError( getPos(), I18n.get( I18n.Word.Expected )
                                + ": '" + I18n.get( I18n.Word.For )
                                + CharType.Name
                        );
                    } else {
                        lit.incPos();
                    }
                }
                else
                if ( ch == '"' ) {
                    lit.incPos();
                    while( !lit.isEnd()
                        && lit.getChar() != '"' )
                    {
                        token.append( CharLiteral.parseChar( lit.getChar() ) );
                        lit.incPos();
                    }
                    if ( lit.getChar() != '"' ) {
                        throw new SyntaxError( getPos(), I18n.get( I18n.Word.Expected )
                                + ": \"" + I18n.get( I18n.Word.For )
                                + StringType.Name
                        );
                    } else {
                        lit.incPos();
                    }
                }
                else {
                    while( ch != ' '
                        && ch != '\t'
                        && ch != ','
                        && !lit.isEnd() )
                    {
                        token.append( ch );

                        lit.incPos();
                        ch = lit.getChar();
                    }
                }
            }
        } catch(CompilingError e)
        {
            throw e;
        } catch(Exception exc)
        {
            throw new SyntaxError( getPos(), exc.getMessage() );
        }

        return token.toString();
    }

    protected static void skipSpaces(StringArrayLiteral lit)
    {
        if ( !lit.isEnd() ) {
            char ch = lit.getChar();
            while( !lit.isEnd()
                && ( ch == ' '
                  || ch == '\t' ) )
            {
                lit.incPos();
                ch = lit.getChar();
            }
        }
    }

    public static String removeExternalBraces(String lit) throws SyntaxError
    {
        lit = lit.trim();
        
        if ( !lit.isEmpty() ) {

            if ( lit.charAt( 0 ) == BeginArrayDelimiter) {
                lit = lit.substring( 1 );
            } else {
                throw new SyntaxError( getPos(), I18n.get( I18n.Word.Expected )
                        + ": '" + BeginArrayDelimiter + "'"
                );
            }

            if ( lit.charAt( lit.length() -1 ) == EndArrayDelimiter) {
                lit = lit.substring( 0, lit.length() -1 );
            } else {
                throw new SyntaxError( getPos(), I18n.get( I18n.Word.Expected )
                        + ": '" + EndArrayDelimiter + "'"
                );
            }
        }

        return lit.trim();
    }

    @Override
    public void setLiteral(Object lit) throws SemanticError
    {
        if ( !( lit.getClass().isArray() ) ) {
            throw new SemanticError( getPos(), "jcg.Array.setLiteral(); "
                    + I18n.get( I18n.Word.Expected )
                    + ' ' + I18n.get( I18n.Word.Array )
            );
        }

        this.lit = lit;
    }

    @Override
    public Object getLiteral()
    {
        return this.lit;
    }

    @Override
    public ArrayType getType()
    {
        return this.type;
    }

    public boolean isString()
    {
        return ( this.getType().isString() );
    }

    private ArrayType type;
    private Object lit;
}
