package ui;

// Main.java

import core.*;
import core.Compiler;
import core.I18n.Language;
import core.Parser.Mode;
import core.exceptions.SyntaxError;
import core.lexical.Position;
import java.io.File;

/**
 * Triggers compilation.
 * @author baltasarq
 */
public class Main {

    public static final String CmdLanguage = "lang";
    public static final String CmdMode = "mode";
    public static final String CmdOutput = "out";

    /**
     * Interprets the arguments passed
     * @param args The arguments passed from command line.
     * @throws SyntaxError If no source file is provided for compilation.
     */
    public static void interpretArgs(String[] args) throws SyntaxError
    {
        // Prepare
        int argNumber = 0;
        sourceFile = null;
        compilationMode = Mode.Beginner;
        outputPlatform = Compiler.OutputPlatform.Bsh;
        Position pos = new Position( "top level", 0, 0 );

        for(; argNumber < args.length; ++argNumber) {
            String arg = args[ argNumber ];

            // Is it the source code file or a switch?
            if ( arg.startsWith( "-" ) ) {

                while ( arg.startsWith( "-" ) ) {
                    arg = arg.substring( 1 );
                }

                int posColon = arg.indexOf( ':' );
                if ( posColon > -1 ) {
                    String cmd = arg.substring( 0, posColon );
                    String value = arg.substring( posColon + 1 ).toUpperCase();

                    if ( cmd.equalsIgnoreCase( CmdLanguage ) ) {
                        try {
                            language = Language.valueOf( value );
                        } catch(IllegalArgumentException exc) {
                            throw new SyntaxError( pos, I18n.get( I18n.Word.Expected )
                                    + ' ' + I18n.get( I18n.Sentence.LanguageCode )
                                    + ", " + I18n.get( I18n.Word.Not )
                                    + ": " + value
                            );
                        }
                    }
                    else
                    if ( cmd.equalsIgnoreCase( CmdOutput ) ) {
                        try {
                            outputPlatform = Compiler.OutputPlatform.parse( value );
                        } catch(IllegalArgumentException exc)
                        {
                            throw new SyntaxError( pos, I18n.get( I18n.Word.Expected )
                                    + ' ' + I18n.get( I18n.Word.Arguments )
                                    + ", " + I18n.get( I18n.Word.Not )
                                    + ": '" + value + '\''
                            );
                        }
                    }
                    else
                    if ( cmd.equalsIgnoreCase( CmdMode ) ) {
                        try {
                            compilationMode = Mode.parse( value  );
                        } catch(IllegalArgumentException exc)
                        {
                            throw new SyntaxError( pos, I18n.get( I18n.Word.Expected )
                                    + ' ' + I18n.get( I18n.Word.Mode )
                                    + ", " + I18n.get( I18n.Word.Not )
                                    + ": '" + value + '\''
                            );
                        }
                    }
                    else {
                        throw new SyntaxError( pos, I18n.get( I18n.Word.Expected )
                                            + ' ' + I18n.get( I18n.Sentence.LanguageCode )
                                            + ", " + I18n.get( I18n.Word.Mode )
                        );
                    }
                }
                else {
                    throw new SyntaxError( pos, I18n.get( I18n.Word.Expected )
                            + ' ' + I18n.get( I18n.Sentence.LanguageCode )
                            + ", " + I18n.get( I18n.Word.Mode )
                    );
                }
            } else {
                // It is the source file!
                if ( sourceFile == null ) {
                    sourceFile = new File( arg );
                } else {
                    throw new SyntaxError( pos, I18n.get( I18n.Word.Duplicated )
                            + ' ' + I18n.get( I18n.Word.Source )
                            + ": '" + arg + '\''
                    );
                }
            }
        }

        if ( sourceFile == null ) {
            throw new SyntaxError( pos, I18n.get( I18n.Word.Expected )
                    + ' ' + I18n.get( I18n.Word.Source )
            );
        }

        return;
    }

    /**
     * Triggers the execution from command line.
     * @param args the command line arguments
     */
    public static void main(String[] args)
    {
        // Prepare and present
        System.out.println( AppInfo.Name + " - jC Compiler " + AppInfo.Version );

        try {
            Compiler compiler = null;

            interpretArgs( args );
            System.out.print( '\t' + language.toString() );
            System.out.print( ", " + outputPlatform.toString() );
            System.out.println( ", " + compilationMode.toString() );
            System.out.println();

            if ( outputPlatform == Compiler.OutputPlatform.Bsh ) {
                compiler = new BshCompiler( sourceFile, compilationMode, language );
            } else {
                compiler = new JsCompiler( sourceFile, compilationMode, language );
            }

            compiler.compile();
            System.out.println( "Ok (" + sourceFile.getName() + ')' );
        }
        catch(CompilingError exc)
        {
            System.out.print( exc.getMessage() + '\n' );
        }
        catch(Exception exc)
        {
            System.out.print( "ERROR (internal): " + exc.getMessage() + '\n' );
        }
    }

    private static Mode compilationMode = Mode.Beginner;
    private static File sourceFile;
    private static Compiler.OutputPlatform outputPlatform;
    private static Language language = Language.EN;
}
