/*
 * HTML Document browsing
 */

package jcv.ui;

import jcv.core.AppInfo;
import jcv.core.L18n;
import jcv.core.L18n.Msg;

import javax.swing.*;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkEvent.EventType;
import javax.swing.event.HyperlinkListener;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.HashMap;

/**
 * Allows to browse various documentation (html) files
 * Acts like a wrapper around JTabbedPane, passed in the constructor
 * @author baltasarq
 */
public class DocBrowser {

    public class Display {
        public Component getTabComponent()
        {
            return this.component;
        }

        public JEditorPane getTabEditor()
        {
            return this.editor;
        }

        public Display(Component cmp, JEditorPane ed)
        {
            this.component = cmp;
            this.editor = ed;
        }

        private Component component;
        private JEditorPane editor;
    }

    protected DocBrowser(JTabbedPane tpPane) throws IOException
    {
        this.cntxtMenu = null;
        this.tabComponent = tpPane;
        this.tabComponent.setTabLayoutPolicy( JTabbedPane.SCROLL_TAB_LAYOUT );
        this.buildContextualMenu();
    }

    protected void buildContextualMenu() throws IOException
    {
        if ( this.cntxtMenu == null ) {
            this.cntxtMenu = new JPopupMenu();
            JMenuItem close = new JMenuItem( L18n.getMsg( Msg.OptClose ) );
            close.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    closeActivated( e );
                }
            });

            JMenuItem refresh = new JMenuItem( L18n.getMsg( Msg.OptRefresh ) );
            refresh.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    try {
                        refreshActivated( e );
                    } catch(IOException exc)
                    {
                        SwingUtil.showErrorMsg( null, AppInfo.Name, exc.getLocalizedMessage() );
                    }
                }
            });

            this.cntxtMenu.add( refresh );
            this.cntxtMenu.add( close );
        }
    }

    public static DocBrowser createDocBrowser(JTabbedPane pane) throws IOException
    {
        if ( browser == null ) {
            browser = new DocBrowser( pane );
        }

        return browser;
    }

    public void browse(File f) throws URISyntaxException, IOException
    {
        Display display = tabs.get( f );
        Component selected = null;

        if ( display == null ) {
            JEditorPane doc = new JEditorPane();
            doc.setVisible( true );
            doc.setEditable( false );
            doc.addHyperlinkListener(
                new HyperlinkListener() {
                    public void hyperlinkUpdate(HyperlinkEvent evt) {
                        if ( evt.getEventType() == EventType.ACTIVATED ) {
                            try {
                                browse( new File( evt.getURL().toURI().getPath() ) );
                            } catch(URISyntaxException | IOException exc)
                            {
                                SwingUtil.showErrorMsg( null, AppInfo.Name, exc.getLocalizedMessage() );
                            }
                        }
                    }
                });

            doc.setComponentPopupMenu( cntxtMenu );
            this.tabComponent.addTab(
                    new File( f.getPath() ).getName(),
                    new JScrollPane( doc )
            );

            selected = this.tabComponent.getComponentAt( this.tabComponent.getTabCount() -1 );
            this.tabs.put( f, new Display( selected, doc ) );
            this.files.put( selected, f );
            this.showDoc( selected );
            this.tabComponent.setSelectedComponent( selected );
        } else {
            this.tabComponent.setSelectedComponent( display.getTabComponent() );
        }

        return;
    }

    /// Retrieves the URI for a tab and calls showDoc(URI). @see showDoc
    public void showDoc(Component currentTab) throws IOException
    {       
        File f = this.files.get( currentTab );
        assert( f != null );

        this.showDoc( f );
    }

    /// Shows/refreshes the document of a given URI
    public void showDoc(File f) throws IOException
    {
        JEditorPane editor = this.tabs.get( f ).getTabEditor();
        editor.setDocument( editor.getEditorKit().createDefaultDocument() );
        editor.setPage( f.toURI().toURL() );
    }

    /// Refresh context menu option activated
    private void refreshActivated(ActionEvent e) throws IOException
    {
        this.refresh();
    }

    /// Close context menu option activated
    private void closeActivated(ActionEvent e) {
        // Close triggered
       this.close();
    }

    /// Closes current tab
    public void close()
    {
        this.close( this.tabComponent.getSelectedComponent() );
    }

    /// Closes a given tab
    public void close(Component currentTab)
    {
        if ( currentTab != null ) {
            // Get affected URI
            File f = files.get( currentTab );

            assert( f != null );

            // Remove from dictionaries
            this.tabs.remove( f );
            this.files.remove( currentTab );

            // Close tab
            this.tabComponent.remove( currentTab );
        }
    }

    /// Refreshes current tab
    public void refresh() throws IOException
    {
        this.refresh( this.tabComponent.getSelectedComponent() );
    }

    /// Refreshes a given tab
    public void refresh(Component currentTab) throws IOException
    {
        this.showDoc( currentTab );
    }

    /// Refreshes all tabs
    public void refreshAll()  throws IOException
    {
        for(Component tab : this.files.keySet()) {
            this.refresh( tab );
        }
    }

    public static DocBrowser browser = null;

    private JTabbedPane tabComponent;
    private HashMap<File, Display> tabs = new HashMap<File, Display>();
    private HashMap<Component, File> files = new HashMap<Component, File>();
    private JPopupMenu cntxtMenu;
}
