/*
 * SourceEditor derives from JTextPane in order to offer
 * a simple way to have an editor with keyword (not syntax) highlighting.
 */

package jcv.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import javax.swing.BorderFactory;
import javax.swing.JTextPane;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultStyledDocument;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.StyledDocument;

/**
 *
 * @author baltasarq
 */
public class SourceEditor extends JTextPane {
    private String[] reservedWords = null;
    private String[] remarkWords = null;
    public static final String EtqNormalStyle = "NORMAL";
    public static final String EtqKeywordStyle = "KEYWORD";
    public static final String EtqRemarkStyle = "REMARK";
    public static final String WhiteSpaces = " \t";
    public static final boolean Token = false;
    public static final boolean Keyword = true;

    private Style normalStyle = null;
    private Style keywordStyle = null;
    private Style remarkStyle = null;

    public SourceEditor(
            Color fgColor, Color bgColor, Color caretColor,
            Color hiColor, Color reColor,
            int fontSize,
            String[] rwords, String[] mwords)
    {
        // Set EditorPane
        setBorder( BorderFactory.createEmptyBorder( 5, 5, 5, 5 ) );
        StyleContext sc = new StyleContext();
        final DefaultStyledDocument doc = new DefaultStyledDocument( sc );
        setDocument( doc );
        setDoubleBuffered( true );

        // Colors and basic font
        setForeground( fgColor );
        setBackground( bgColor );
        setFont( new Font( "monospaced", Font.PLAIN, fontSize ) );
        setCaretColor( caretColor );
        reservedWords = rwords;
        remarkWords = mwords;

        // Styles
        final Style defaultStyle = sc.getStyle( StyleContext.DEFAULT_STYLE );
        normalStyle = sc.addStyle( EtqNormalStyle, defaultStyle );
        StyleConstants.setForeground( normalStyle, fgColor );
        StyleConstants.setFontSize( normalStyle, fontSize );
        StyleConstants.setFontFamily( normalStyle, "monospaced" );

        keywordStyle = sc.addStyle( EtqKeywordStyle, defaultStyle );
        StyleConstants.setForeground( keywordStyle, hiColor );
        StyleConstants.setFontSize( keywordStyle, fontSize );
        StyleConstants.setBold( keywordStyle, true );

        remarkStyle = sc.addStyle( EtqRemarkStyle, defaultStyle );
        StyleConstants.setForeground( remarkStyle, reColor );
        StyleConstants.setBold( remarkStyle, true );

        doc.setLogicalStyle( 0, normalStyle );

        this.addKeyListener(
                new KeyListener() {
                     public void keyReleased(KeyEvent evt)
                        { onKeyReleased( evt ); }
                     public void keyPressed(KeyEvent evt)
                        { onKeyPressed( evt ); }
                     public void keyTyped(KeyEvent evt)
                        {}
                }
        );

    }

    @Override
    public boolean getScrollableTracksViewportWidth()
    {
        // should not allow text to be wrapped
        return false;
    }

    @Override
    public void setSize(Dimension d)
    {
        Dimension pSize = getParent().getSize();

        if ( d.width < pSize.width ) {
            // don't let the Textpane get sized differently than its parent
            super.setSize( pSize.width, d.height );
        } else {
            super.setSize( d );
        }
    }

    public void setFontSize(int fontSize)
    {
        setFont( new Font( "monospaced", Font.PLAIN, fontSize ) );
        
        StyleConstants.setFontSize( normalStyle, fontSize );
        StyleConstants.setFontSize( keywordStyle, fontSize );
        StyleConstants.setFontSize( remarkStyle, fontSize );

        highlightAllSource();
    }

    public void setFile(File file) throws IOException
    {
            if ( file == null ) {
                setText( "" );
            } else {
                BufferedInputStream in = new BufferedInputStream( new FileInputStream( file ) );
                read( in, null );
                in.close();
                highlightAllSource();
            }
    }

    public void setReservedWords(String[] rwords)
    {
        reservedWords = rwords;
    }

    public void removeChar(int pos)
    {
         // Control positions
        if ( pos < 0 ) {
            pos = 0;
        }

        if ( pos >= getText().length() ) {
            pos = getText().length();
        }

        // Insert text
        try {
            this.getDocument().remove( pos, 1 );
        } catch (BadLocationException ex) {
            // Error
        }
    }

    public void insert(int pos, String txt)
    {
        // Control positions
        if ( pos < 0 ) {
            pos = 0;
        }

        if ( pos >= getText().length() ) {
            pos = getText().length();
        }

        // Insert text
        try {
            this.getDocument().insertString( pos, txt, null );
        } catch (BadLocationException ex) {
            // Error
        }
    }
    
    public int getLineOfOffset(int offset)
    {
        javax.swing.text.Element map = getDocument().getDefaultRootElement();

        if (offset < 0) {
            return 0;
        } else if (offset > getDocument().getLength()) {
            return map.getElementIndex( map.getElementCount() -1 );
        } else {
            return map.getElementIndex( offset );
        }
    }

    public int getLineStartOffset(int row)
    {
        int offset = 0;
        javax.swing.text.Element map = getDocument().getDefaultRootElement();

        if ( row >= 0 ) {
            if ( row >= map.getElementCount() )
                    offset = map.getElement( map.getElementCount() -1 ).getStartOffset();
            else    offset = map.getElement( row ).getStartOffset();
        }

        return offset;
    }

    private boolean highlighting = false;

    public boolean isHighlighting()
    {
        return highlighting;
    }

    public boolean shouldHighlight(String text, int pos, int endPos)
    {
        boolean toret = true;

        --pos;
        ++endPos;

        // Test whether it should be highlighted or not
        if ( pos >= 0 ) {
           if ( Character.isDigit( text.charAt( pos ) )
             || Character.isLetter( text.charAt( pos ) ) )
           {
               toret = false;
           }
        }

        if ( endPos < text.length() ) {
            if ( Character.isDigit( text.charAt( endPos ) )
              || Character.isLetter( text.charAt( endPos ) ) )
            {
                toret = false;
            }
        }

        return toret;
    }

    public void highlightAllSource()
    {
       highlighting = true;
       setIgnoreRepaint( true );

       StyledDocument doc = getStyledDocument();

       if ( remarkStyle != null
         && keywordStyle != null
         && normalStyle != null )
       {
           // Set default colors
           removeStyles( 0, getText().length() );

           // Color reserved & remarked words
           if ( reservedWords != null ) {
            colorSource( reservedWords, keywordStyle, Keyword );
           }

           if ( remarkWords != null ) {
            colorSource( remarkWords, remarkStyle, Token );
           }
       }

       setIgnoreRepaint( false );
       highlighting = false;
    }

    private void onKeyPressed(java.awt.event.KeyEvent evt)
    {
        if ( this.getSelectedText() != null
          && !( this.getSelectedText().isEmpty() ) )
        {
            if ( evt.getKeyCode() == KeyEvent.VK_TAB ) {
                this.setIgnoreRepaint( true );

                int[] lines = this.getMarkedLines();
                int firstPos = lines[ 0 ];
                int firstLine = lines[ 2 ];
                int lastLine = lines[ 3 ];

                this.setSelectionEnd( firstPos );

                if ( evt.isShiftDown() ) {
                    this.indentLess( firstLine, lastLine );
                    evt.consume();
                }
                else {
                    this.indentMore( firstLine, lastLine );
                    evt.consume();
                }
                
                this.setIgnoreRepaint( false );
            }
        }
    }

    public int[] getMarkedLines()
    {
        int[] toret = new int[ 4 ];

        toret[ 0 ] = this.getSelectionStart();
        toret[ 1 ] = this.getSelectionEnd();
        toret[ 2 ] = this.getLineOfOffset( toret[ 0 ] );
        toret[ 3 ] = this.getLineOfOffset( toret[ 1 ] );

        if ( toret[ 1 ] > this.getLineStartOffset( toret[ 3 ] ) ) {
                    ++( toret[ 3 ] );
        }

        return toret;
    }

    public void indentMore(int firstLine, int lastLine)
    {
        for(int i = firstLine; i < lastLine; ++i) {
            int insertPos = this.getLineStartOffset( i );
            this.insert( insertPos, "\t" );
        }

        this.setSelectionStart( this.getLineStartOffset( firstLine ) );
        this.setSelectionEnd( this.getLineStartOffset( lastLine ) );
    }

    public void indentLess(int firstLine, int lastLine)
    {
        for(int i = firstLine; i < lastLine; ++i) {
            int removePos = this.getLineStartOffset( i );
            char ch = this.getText().charAt( removePos );

            if ( WhiteSpaces.indexOf( ch ) >= 0 ) {
                this.removeChar( removePos );
            }
        }

        this.setSelectionStart( this.getLineStartOffset( firstLine ) );
        this.setSelectionEnd( this.getLineStartOffset( lastLine ) );
    }

    private void onKeyReleased(java.awt.event.KeyEvent evt) {
        if ( !evt.isActionKey() ) {
            updateHighlighting();
        }

        if ( evt.getKeyCode() == KeyEvent.VK_ENTER ) {
            setIgnoreRepaint( true );            

            // Add spaces if needed
            String spaces = "";
            int currentPos = getCaretPosition();
            int line = getLineOfOffset( currentPos );
            String sourceCode = getText();

            if ( line > 0 ) {
                int pos = getLineStartOffset( line - 1 );
                char ch = sourceCode.charAt( pos );

                while( WhiteSpaces.indexOf( ch ) > -1 ) {
                    if ( ch == '\t' )
                            spaces += '\t';
                    else    spaces += ' ';

                    ++pos;
                    ch = sourceCode.charAt( pos );
                }

                insert( currentPos, spaces );
                setCaretPosition( currentPos + spaces.length() );
            }

            setIgnoreRepaint( false );
        }
    }

    private void colorSource(final String[] reservedWords,
                             final Style keywordStyle, final boolean isKeyword)
    {
        new Thread(
            new Runnable() {
                public void run()
                {
                    doColorSource( reservedWords, keywordStyle, isKeyword );
                }
        }).start();
    }

    private void doColorSource(String[] reservedWords, Style keywordStyle, boolean isKeyword)
    {
       int pos = 0;
       int endPos = 0;
       String text = getText();

       for(String pattern : reservedWords ) {
               // Search for pattern
               while ( ( pos = text.indexOf( pattern, pos ) ) >= 0 ) {
                    endPos = pos + pattern.length() -1;


                    if (  !isKeyword
                       || shouldHighlight( text, pos, endPos ) )
                    {
                        // Highligh text
                        colorToken( pos, endPos - pos +1, keywordStyle );
                    }

                    pos = endPos +1;
               }
        }
    }

    private void colorToken(int pos, int len, Style style)
    {
        getStyledDocument().setCharacterAttributes( pos, len, style, true );
    }

    private void removeStyles(int pos, int len)
    {
        getStyledDocument().setCharacterAttributes( pos, len, normalStyle, true );
    }

    private void updateHighlighting()
    {
        int pos = getCaretPosition();
        int lineNumber = getLineOfOffset( pos );
        int posBeg = getLineStartOffset( lineNumber );
        int posEnd = getLineStartOffset( lineNumber +1 ) -1;
        String text = getText();

        removeStyles( posBeg, ( posEnd - posBeg ) + 1 );
        colorSourcePart( posBeg, posEnd, remarkWords, remarkStyle, Token );
        colorSourcePart( posBeg, posEnd, reservedWords, keywordStyle, Keyword );
    }

    private void colorSourcePart(
            int posBeg, int posEnd,
            String[] reservedWords, Style keywordStyle, boolean isKeyword)
    {
       int pos = posBeg;
       int endPos = 0;
       String text = getText();

       for(String pattern : reservedWords ) {
               // Search for pattern
               while ( ( pos = text.indexOf( pattern, pos ) ) >= 0 ) {
                    endPos = pos + pattern.length() -1;

                    if ( pos >= posEnd ) {
                        break;
                    }
                    else
                    if ( !isKeyword
                      || shouldHighlight( text, pos, endPos ) )
                    {
                        // Highligh text
                        colorToken( pos, endPos - pos +1, keywordStyle );
                    }

                    pos = endPos +1;
               }
        }
    }

    public boolean isToken(String[] tokens, String token)
    {
        for(String tok : tokens) {
            if ( tok.compareTo( token ) == 0 ) {
                return true;
            }
        }

        return false;
    }

    public static boolean isIdentifierPart(char ch)
    {
        return ( Character.isDigit( ch ) || Character.isLetter( ch ) || ch == '_' );
    }

    @Override
    public void setText(String txt)
    {
        super.setText( txt );
        highlightAllSource();
    }
}
