// SwingUtil.java

package jcv.ui;

import javax.swing.*;
import javax.swing.filechooser.FileNameExtensionFilter;
import java.io.File;

/**
 * Utility functions for Swing
 * User: baltasarq
 * Date: 5/21/13
 * Time: 2:13 PM
 */

public class SwingUtil {
    /**
     * Shows an error message
     * @param window the parent window
     * @param name the title of the message window
     * @param txt the message itself
     */
    public static void showErrorMsg(JFrame window, String name, String txt)
    {
        showMsg( window, name, txt, JOptionPane.ERROR_MESSAGE );
    }

    /**
     * Shows an information message
     * @param window the parent window
     * @param name the title of the message window
     * @param txt the message itself
     */
    public static void showInfoMsg(JFrame window, String name, String txt)
    {
        showMsg( window, name, txt, JOptionPane.INFORMATION_MESSAGE );
    }

    private static void showMsg(JFrame window, String appName, String txt, int msgType)
    {
        JOptionPane.showMessageDialog( window, txt, appName, msgType );
    }

    /**
     * Shows a save dialog
     * @param window the parent window
     * @param title the title of the save dialog
     * @param ext the extension filter (i.e. "html")
     * @param extExpl the explanation for the filter (i.e. "webpage")
     * @param f The last or current path
     * @return the file selected, null if cancelled
     */
    public static File showSaveFileDlg(JFrame window, String title, String ext, String extExpl, File f)
    {
        FileNameExtensionFilter[] filters = { new FileNameExtensionFilter( extExpl, ext ) };

        return showSaveFileDlg( window, title, filters, f );
    }

    /**
     * Shows a save dialog
     * @param window the parent window
     * @param title the title of the save dialog
     * @param ext a vector with the extension filters (i.e. "html")
     * @param extExpl a vector with the explanations for the filters (i.e. "webpage")
     * @param f The last or current path
     * @return the file selected, null if cancelled
     */
    public static File showSaveFileDlg(JFrame window, String title, String[] ext, String[] extExpl, File f)
    {
        FileNameExtensionFilter[] filters = new FileNameExtensionFilter[ ext.length ];

        for(int i = 0; i < ext.length; ++i) {
            filters[ i ] = new FileNameExtensionFilter( extExpl[ i ], ext[ i ] );
        }

        return showSaveFileDlg( window, title, filters, f );
    }

    private static File showSaveFileDlg(JFrame window, String title, FileNameExtensionFilter[] filters, File f)
    {
        // Create save dialog
        File toret = null;
        JFileChooser fileSelector = new JFileChooser();
        fileSelector.setDialogType( JFileChooser.SAVE_DIALOG );
        fileSelector.setDialogTitle( title );

        // Add filters
        for( FileNameExtensionFilter filter : filters) {
            fileSelector.addChoosableFileFilter( filter );
        }
        fileSelector.setFileFilter( filters[ 0 ] );

        // Locate the last directory, if any
        if ( f != null ) {
            fileSelector.setSelectedFile( f );
        }

        // run
        if ( fileSelector.showSaveDialog( window ) == JFileChooser.APPROVE_OPTION ) {
            toret = fileSelector.getSelectedFile();
        }

        return toret;
    }

    /**
     * Shows am open dialog
     * @param window the parent window
     * @param title the title of the save dialog
     * @param ext the extension filter (i.e. "html")
     * @param extExpl the explanation for the filter (i.e. "webpage")
     * @param f The last or current path
     * @return the file selected, null if cancelled
     */
    public static File showOpenFileDlg(JFrame window, String title, String[] ext, String[] extExpl, File f)
    {
        FileNameExtensionFilter[] filters = new FileNameExtensionFilter[ ext.length ];

        for(int i = 0; i < ext.length; ++i) {
            filters[ i ] = new FileNameExtensionFilter( extExpl[ i ], ext[ i ] );
        }

        return showOpenFileDlg( window, title, filters, f );
    }

    /**
     * Shows an open dialog
     * @param window the parent window
     * @param title the title of the save dialog
     * @param ext a vector with the extension filters (i.e. "html")
     * @param extExpl a vector with the explanations for the filters (i.e. "webpage")
     * @param f The last or current path
     * @return the file selected, null if cancelled
     */
    public static File showOpenFileDlg(JFrame window, String title, String ext, String extExpl, File f)
    {
        FileNameExtensionFilter[] filters = {
                new FileNameExtensionFilter( extExpl, ext )
        };

        return showOpenFileDlg( window, title, filters, f );
    }

    private static File showOpenFileDlg(JFrame window, String title, FileNameExtensionFilter[] filters, File f)
    {
        // Create open dialog
        File toret = null;
        JFileChooser fileSelector = new JFileChooser();
        fileSelector.setDialogType( JFileChooser.OPEN_DIALOG );
        fileSelector.setDialogTitle( title );

        // Add filters
        for( FileNameExtensionFilter filter : filters) {
            fileSelector.addChoosableFileFilter( filter );
        }
        fileSelector.setFileFilter( filters[ 0 ] );

        // Locate the last directory, if any
        if ( f != null ) {
            if ( f.isDirectory() )
                fileSelector.setCurrentDirectory( f );
            else    fileSelector.setSelectedFile( f );
        }

        // run
        if ( fileSelector.showOpenDialog( window ) == JFileChooser.APPROVE_OPTION ) {
            toret = fileSelector.getSelectedFile();
        }

        return toret;
    }

    /**
     * @param window the parent window
     * @param title the title of the directory dialog
     * @param f the last path used
     * @return the path to the file, or null if cancelled
     */
    public static File showOpenDirDlg(JFrame window, String title, File f)
    {
        // Prepare dialog
        JFileChooser fileSelector = new JFileChooser();
        fileSelector.setDialogTitle( title );
        fileSelector.setDialogType( JFileChooser.OPEN_DIALOG );

        return selectDirectory( window, title, fileSelector, f );
    }

    /**
     * @param window the parent window
     * @param title the title of the directory dialog
     * @param f the last path used
     * @return the path to the file, or null if cancelled
     */
    public static File showSaveDirDlg(JFrame window, String title, File f)
    {
        // Prepare dialog
        JFileChooser fileSelector = new JFileChooser();
        fileSelector.setDialogTitle( title );
        fileSelector.setDialogType( JFileChooser.SAVE_DIALOG );

        return selectDirectory( window, title, fileSelector, f );
    }

    private static File selectDirectory(JFrame window, String title, JFileChooser fileSelector, File f)
    {
        int selectorValue = JFileChooser.CANCEL_OPTION;
        File toret = null;

        // Locate the last directory, if any
        if ( f != null ) {
            if ( f.isDirectory() ) {
                fileSelector.setCurrentDirectory( f );
            } else {
                fileSelector.setCurrentDirectory( f.getParentFile() );
            }
        }

        // run
        fileSelector.setFileSelectionMode( JFileChooser.DIRECTORIES_ONLY );

        if ( fileSelector.getDialogType() == JFileChooser.OPEN_DIALOG ) {
            selectorValue =  fileSelector.showOpenDialog( window );
        } else {
            selectorValue =  fileSelector.showSaveDialog( window );
        }

        if ( selectorValue == JFileChooser.APPROVE_OPTION )
        {
            toret = fileSelector.getSelectedFile();
        }

        return toret;
    }
}
